/*
** Module   : Persistence.java
** Abstract : Provides persistence services using Hibernate ORM
**
** Copyright (c) 2004-2009, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description---------------------
** 001 ECF 20050830   @22419 Created initial version. Provides bootstrap
**                           services to configure a Hibernate session
**                           factory per database.
** 002 ECF 20050902   @22480 Added essential query services; refactored.
**                           This includes scrollable cursor queries,
**                           individual record loads into a re-used data
**                           model object, and enforced, unique queries.
**                           Refactored all bootstrap and database
**                           initialization logic into external class
**                           (DatabaseManager).
** 003 ECF 20051104   @23254 Changes to support locking and transactions.
**                           Added methods to lock records and query lock
**                           status, and to begin, commit, and roll back
**                           transactions.
** 004 ECF 20051109   @23382 Changed use paradigm from all static methods
**                           to getting an instance by factory. Previously
**                           all methods were static and required a
**                           physical database name to be specified. Now
**                           one instance is created and cached per
**                           physical database. Client code retrieves an
**                           instance via a factory method and operates on
**                           that instance. Instances are immutable and
**                           shared across multiple contexts. Added method
**                           to retrieve the next unique primary key using
**                           a database sequence.
** 005 ECF 20051215   @23751 Added support for ambiguous check. Also made
**                           primary key (phase 1) searches read-only.
** 006 ECF 20060120   @24044 Integrated with P2J server. Replaced
**                           ThreadLocal with ContextLocal.
** 007 ECF 20060130   @24354 Added temp table support. Required a new
**                           list() method, disabling the lock manager for
**                           temp table cases, and enabling arbitrary SQL
**                           statements to be executed (for temp table
**                           creation and dropping).
** 008 ECF 20060218   @24720 Created uniqueResultViolation method. This
**                           processing was needed from multiple places,
**                           including outside this class.
** 009 ECF 20060306   @25016 Added loadReadOnly method. Supports internal
**                           queries used for persistence framework
**                           housekeeping, including resolving foreign
**                           records using legacy keys (for natural join
**                           reverse resolution).
** 010 ECF 20060329   @25255 Fixed executeSQL method. Was passing the SQL
**                           string to PreparedStatement's execute method.
** 011 ECF 20060602   @26943 Optimization. HQL queries now will always
**                           attempt to load the full record, rather than
**                           just the primary key ID. This avoids the
**                           overhead of a separate load step.
** 012 ECF 20060608   @27001 Changed list() method signature and updated
**                           implementation. Number of results can now be
**                           limited and results can be set read-only.
** 013 ECF 20060610   @27098 Consolidated lock processing. Internal lock
**                           processing now uses public lock() method.
** 014 ECF 20060621   @27505 Changed session management and transaction
**                           implementation. Sessions are now always
**                           created upon demand, and detached records
**                           from a previous session are immediately
**                           associated with the new session. The begin
**                           transaction API is more lenient, reporting if
**                           a new transaction could not be created,
**                           rather than throwing an exception.
** 015 ECF 20060627   @27590 Improved temp table support. Always flush
**                           Hibernate session before executing JDBC
**                           statements directly. Ensure save and delete
**                           operations are executed within transactions.
** 016 ECF 20060712   @28050 Added trimTextArgs. Trims trailing whitespace
**                           from any query substitution parameter which
**                           is of type com.goldencode.p2j.util.character.
** 017 ECF 20060719   @28100 Fixed memory leak. Implemented cleanup()
**                           method for Hibernate Session's ContextLocal
**                           variable to ensure Session is explicitly
**                           closed when context ends. Although the P2J
**                           server would not pin the Session's memory
**                           directly, there was still a reference within
**                           Hibernate, such that the Session data could
**                           not be garbage collected after the client
**                           context ended.
** 018 ECF 20060719   @28101 Enable result set 'paging'. The list() method
**                           now accepts a startOffset parameter.
** 019 ECF 20060720   @28115 Integrated ChangeBroker. Enables optimization
**                           of avoiding unnecessary auto-flushing of the
**                           active Hibernate session. This dramatically
**                           improves the performance of RandomAccessQuery
**                           when the session contains a large number of
**                           entities.
** 020 ECF 20060725   @28187 Consolidated context local variables within
**                           the Context inner class. Also added a
**                           mechanism to prevent sessions from being
**                           closed while external entities are dependent
**                           upon them.
** 021 ECF 20060728   @28269 Made queries cacheable; improved reliability
**                           of locking in dynamic retrieval mode.
** 022 ECF 20060828   @28930 Better error handling. In some cases, session
**                           was not being closed after catching an
**                           unrecoverable exception.
** 023 ECF 20060830   @29005 Fixed getQuery(). In case where we were not
**                           explicitly paging results, the query would
**                           return an empty result set, due to incorrect
**                           maxResults setting.
** 024 ECF 20060901   @29136 Fixed memory leak. decrementSessionsUsers()
**                           was actually incrementing the session user
**                           count, due to a typo.
** 025 ECF 20060903   @29203 Fixed Hibernate assertion failure. Ensure
**                           session is flushed before an object is
**                           evicted.
** 026 ECF 20060904   @29219 Better error reporting. scroll() reports HQL
**                           statement and context info on query error.
** 027 GES 20060905   @29230 Updated trimTextArgs() to truncate any 
**                           character vars that have embedded null bytes.
**                           This ensures that null bytes can never be
**                           passed into queries as substitution
**                           arguments.
** 028 ECF 20060911   @29460 Fixed query defect. Restore default query
**                           settings when checking out a cached query.
** 029 ECF 20060919   @29697 Made mods required by changes to ChangeBroker
**                           calls. Also added trace debug output during
**                           record eviction.
** 030 ECF 20060929   @30035 Fixed out of range query substitution date
**                           parameters. We now 'clip' those dates to the
**                           minimum and maximum date values supported by
**                           the backing database and log a debug message.
**                           Renamed trimTextArgs() method to
**                           preprocessQueryParameters() to reflect
**                           broader purpose.
** 031 ECF 20061010   @30309 Made changes to accommodate auto-commit temp
**                           tables. These were primarily side effects of
**                           changes made to TemporaryBuffer and
**                           RecordBuffer. A new variant of the static
**                           factory method to retrieve instances of this
**                           class was added. The new version of get()
**                           will create a secondary instance of this
**                           class, backed by the same database as the
**                           primary instance. This allows auto-commit and
**                           non-auto-commit temp tables to share the same
**                           physical backend, but maintain separate
**                           connections and Hibernate sessions.
** 032 ECF 20061012   @30390 Fixed nonthreadsafe access error. Flush
**                           session to ensure consistency, when a request
**                           to close the session cannot be honored
**                           immediately (because a query is pinning it
**                           open).
** 033 ECF 20061014   @30527 Integrated SessionListener architecture.
**                           Objects which register as SessionListeners
**                           are notified when a Hibernate session is
**                           about to close. Deprecated session user count
**                           mechanism which would pin open a session as
**                           long as a reference count of dependents was
**                           non-zero.
** 034 ECF 20061110   @31094 Improved logging. Errors reassociating
**                           detached records are logged in detail to
**                           prevent these errors from being masked in
**                           silent error mode.
** 035 ECF 20061113   @31140 Added new variant of load(). Needed for cases
**                           where a RecordBuffer instance is not readily
**                           available.
** 036 ECF 20061116   @31236 Minor adjustments to exception handling.
**                           Ensure session gets closed when a database
**                           error occurs.
** 037 ECF 20061120   @31325 Changed session closing behavior. Do not
**                           close temp table sessions if the temp table
**                           count exceeds zero (with no database errors).
** 038 ECF 20061212   @31667 Fixed context cleanup. Inner class Context's
**                           cleanup was not properly connected to end of
**                           context cleanup, so it was never being
**                           invoked.
** 039 ECF 20070111   @31779 Added flushBatch method. Flushes a batch of
**                           records to the database and clears the active
**                           session.
** 040 ECF 20070130   @32039 Various cleanup. Ensure session is flushed if
**                           necessary after attaching a record. Ensure
**                           session is closed for all Hibernate exception
**                           processing. Ensure session is closed if a
**                           SessionListener throws an exception during
**                           close notification. Deprecated method which
**                           creates secondary Persistence instance for a
**                           single database.
** 041 ECF 20070212   @32155 Integrated DirtyReadManager. This allows DMO
**                           validation to check uncommitted updates
**                           across client contexts. Also added the
**                           ability to temporarily disable flushing,
**                           which was necessary to correct a flaw in
**                           validation processing.
** 042 ECF 20070228   @32246 Modified HQL-based load() method. It can now
**                           handle either an ID-only projection query or
**                           a query which fetches and returns a DMO
**                           directly. In the latter case, the additional
**                           lock() and get() steps are omitted.
** 043 ECF 20070326   @32611 Added bulk delete method. Deletes all records
**                           which meet the specified criteria for a given
**                           DMO type.
** 044 ECF 20070329   @32688 Fixed attach/detach. Flush check was not
**                           correct.
** 045 ECF 20070412   @32983 Retrofitted locking to use RecordLockContext.
**                           An instance of this class is created for each
**                           context, one per database. Simplified context
**                           cleanup.
** 046 ECF 20070416   @33025 Integrated user interrupt handling.
** 047 ECF 20070508   @33458 Added queryIndexData(). Uses database
**                           metadata to inspect indices on a given table.
** 048 ECF 20070514   @33485 Fixed queryIndexData(). In some cases, root
**                           column name must be extracted from string
**                           queried from index metadata in a dialect
**                           specific way.
** 049 ECF 20070516   @33669 Fixed abend caused by an orphaned transaction
**                           at session close. When a session is closed
**                           due to an error, roll back any open
**                           transaction before closing the session. Also
**                           made error logging less noisy at this layer,
**                           since a severe error usually will be logged
**                           in error handling provided by users of the
**                           persistence framework. Details are still
**                           logged at FINE logging level.
** 050 ECF 20070519   @33682 Fixed load() method (variant which loads by
**                           primary key). Ensure session is flushed if
**                           necessary and use Session.get() instead of
**                           Session.load(), in case record was deleted.
** 051 ECF 20070602   @33926 Integrated new scope options of
**                           SessionListener. GLOBAL_SCOPE registers a
**                           session listener cleanup routine with the
**                           transaction manager global scope. NO_SCOPE
**                           does not register a cleanup method at all.
** 052 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 053 ECF 20070703   @34360 Removed IllegalStateException from flush().
**                           Now throws recoverable PersistenceException
**                           instead when no session is available.
** 054 ECF 20070706   @34401 Implemented SessionListener deregistration
**                           notification.
** 055 ECF 20070708   @34414 Optimization. Cache ChangeBroker in context
**                           local state, rather than retrieving it every
**                           time flushing is checked.
** 056 ECF 20070709   @34419 Optimization. Consolidated access to Context
**                           instance to prevent multiple ContextLocal
**                           lookups within the same methods. Changed some
**                           signatures and added/removed several private
**                           methods to accomodate this.
** 057 ECF 20070720   @34620 Modified implementation of attach(). It now
**                           associates a detached DMO with the current
**                           Hibernate session.
** 058 ECF 20070812   @34857 Changed error message on non-unique result
**                           exception. Report buffer name rather than
**                           DMO name.
** 059 ECF 20070817   @34876 Fixed all variants of load() to prevent stale
**                           data retrieval. There were cases where a DMO
**                           instance in the Hibernate session cache would
**                           become stale based on an update in another
**                           user context. We now force a refresh from the
**                           database when reading a record with a lock.
** 060 ECF 20070901   @35302 Changed executeSQL() method. Connection is
**                           now set to autocommit=true before the SQL
**                           statement is executed. Added new method
**                           executeSQLQuery() to allow execution of
**                           arbitrary SQL queries with no change to
**                           connection's autocommit setting. Replaced use
**                           of database name string with database info
**                           object. Added trace logging of execution time
**                           of queries and other database accesses.
** 061 ECF 20071002   @35331 Implemented dialect-specific callback hooks.
**                           Registered P2JDialect is given a chance to
**                           perform post-transaction processing with
**                           dialect-specific data stored in the current
**                           client context.
** 062 ECF 20071008   @35399 Added back convenience method get(String).
** 063 ECF 20071008   @35445 Fixed regression introduced by #060 (@35302).
**                           Incorrect LockManager and DirtyReadManager
**                           instances were being created for the temp
**                           table database.
** 064 ECF 20071018   @35575 Refactored to enable creation by factory.
**                           Moved static factory methods to new class
**                           PersistenceFactory. Also, lock manager and
**                           dirty read manager classes are now read from
**                           the directory instead of hard-coded.
** 065 ECF 20071221   @36545 Fixed load() method implementations. We now
**                           catch UnresolvableObjectException, which is a
**                           recoverable condition which can occur when
**                           invoking Session.refresh(). This will be
**                           thrown by Hibernate when we have a DMO in the
**                           first level cache, which has been deleted
**                           from the database.
** 066 SVL 20071231   @36626 Identity manager is initialized to support
**                           primary key generation on a per-table basis.
** 067 ECF 20080102   @36644 Removed deprecated nextPrimaryKey() method.
**                           Replacement is nextPrimaryKey(String) added
**                           in #066 (@36626).
** 068 ECF 20080108   @36733 Fixed bulk delete. Removed optization of
**                           query flush, which was causing a
**                           StaleStateException in cases of deleted DMOs
**                           with collection associations.
** 069 SVL 20080111   @36772 getIdentityManager() function has been added.
** 070 ECF 20080129   @37004 Added debug logging. Log DMO persist event.
** 071 ECF 20080312   @37480 Integrated support for embedded H2 database.
**                           Also changed return type of list() method to
**                           be less specific.
** 072 ECF 20080403   @38136 Use entity name instead of DMO class to test
**                           if flush is required. Added load() variant
**                           which accepts entity name instead of DMO
**                           class.
** 073 SVL 20080421   @38097 Added foreignKeysEnabled property.
** 074 CA  20080430   @38160 Added a static method retrieveIndexData which
**                           contains the queryIndexData's code related to
**                           building the index list from a result set.
** 075 ECF 20080606   @38642 Removed ValidationContext references.
** 076 ECF 20080611   @38697 Major refactoring. Index metadata query
**                           methods were rationalized to be accessible in
**                           more situations. Added methods to temporarily
**                           "borrow" a connection from a current session.
**                           Removed all references to the obsolete
**                           DirtyReadManager class hierarchy.
** 077 ECF 20080624   @38918 Optimized column type metadata methods used
**                           primarily at startup. This approach avoids
**                           many round trips to the database to query
**                           column metadata, but may not work with all
**                           JDBC drivers. In case not, the original,
**                           unoptimized approach is used.
** 078 CA  20080605   @38566 Renamed method delete to deleteOrUpdate.
** 079 ECF 20080806   @39324 Added rowid support. Changed default identity
**                           manager to SequenceIdentityManager, which
**                           generates database-wide unique IDs. Primary
**                           keys are now 64-bit integers.
** 080 CA  20080815   @39462 Support API change in DBUtils. Fixed NPE in
**                           getColumnDataType.
** 081 SVL 20080901   @39652 Identity pool is initialized together with
**                           identity manager.
** 082 ECF 20080923   @39936 Memory tuning. Do not cache queries used for
**                           deletes/updates.
** 083 ECF 20080924   @39946 Added load(RecordIdentifier). Loads a single
**                           record, given its unique identifier.
** 084 ECF 20081015   @40091 Refactored query substitution parameter
**                           preprocessing. Abstracted out a static method
**                           to process a single parameter, so it is
**                           accessible to other classes in the package.
** 085 ECF 20081021   @40208 Force refresh of DMOs in active buffers when
**                           getting a new Session after a rollback. This
**                           ensures active buffers have current versions
**                           of these records.
** 086 ECF 20081024   @40224 Fixed regression caused by #085 (@40208). If
**                           a record in a buffer was deleted during a
**                           rollback, we must handle this stale state
**                           when creating the new session and refreshing
**                           records left in active buffers.
** 087 ECF 20081106   @40354 Modified directory lookups for lock manager
**                           and identity manager. Changed path to allow
**                           for other, related properties to be set in
**                           the directory.
** 088 ECF 20081118   @40549 Minor change to error message text.
** 089 ECF 20081204   @40803 Added performance warning logging. Queries
**                           which take longer than certain thresholds
**                           (configured in the directory) trigger a
**                           warning log entry.
** 090 ECF 20090114   @41162 Fixed load() (HQL variant). Was not correctly
**                           unlocking when passed a lock type of NONE.
** 091 ECF 20090128   @41239 Improved Context.getSession(). When
**                           associating the current records of active
**                           buffers with a new session after a rollback
**                           of the previous session, we only refresh
**                           those we know to be stale, rather than all
**                           of them. This reduces the number of trips
**                           to the database. The rolledBack parameter to
**                           Context.closeSession() was removed as a
**                           result of this change.
** 092 ECF 20090204   @41257 Improved getQuery() performance. We now cache
**                           Query objects more aggressively, which avoids
**                           the overhead of Hibernate's initialization of
**                           these objects when possible.
** 093 ECF 20090212   @41295 Added bulk eviction method. Exposed Hibernate
**                           session via bind() method.
** 094 ECF 20090311   @41502 Added support for limit/max value inlining in
**                           a SELECT statement. We do this only in very
**                           limited cases: currently only for simple load
**                           queries.
** 095 ECF 20090601   @42576 API change. Renamed load(RecordIdentifier) to
**                           quickLoad(RecordIdentifier). Also made method
**                           public, as it is needed from dirty package.
** 096 ECF 20090619   @42771 Added scroll() variant which accepts ScrollMode.
**                           Existing variant now calls this variant with a
**                           default ScrollMode of SCROLL_INSENSITIVE.
** 097 CA  20090625   @42945 Transaction roll back is executed at database 
**                           level, so session needs to be flushed before the
**                           actual roll back.
** 098 ECF 20090702   @43036 Fixed session closing/cleanup logic. Removed
**                           deprecated code.
** 099 ECF 20090711   @43188 Implemented implicit transaction support. To
**                           prevent Hibernate from beginning open-ended
**                           transactions for query operations, we now open an
**                           implicit transaction for each such operation.
**                           These are closed either when an explicit
**                           transaction is opened, or after an idle timeout.
** 100 ECF 20090716   @43219 Fixed regression in #099 (@43188). In the event a
**                           session was marked as broken in the implicit
**                           transaction timer thread, this state was never
**                           being reset after a new session has been opened.
** 101 ECF 20090719   @43305 Added safety code to implicit commit task. It is
**                           more resilient to unexpected errors (such as
**                           OOME), to avoid taking down the timer thread.
** 102 ECF 20090722   @43325 Disabled implicit transaction commit thread. It
**                           is the prime suspect in a "possible nonthreadsafe
**                           access to session" error. May be re-enabled at a
**                           later time.
** 103 ECF 20090723   @43339 Made preprocessQueryParameter() public. This
**                           static method is needed by the dirty subpackage.
** 104 CA  20090724   @43381 Rolled back H091 (#41239) - removed the
**                           buffer.isStale() dependency in
**                           Context.getSession().
** 105 ECF 20090723   @43418 Rewrote implicit transaction support. Removed
**                           secondary thread and push/pop calls from low
**                           level query services. These must be implemented
**                           by callers instead to enable coarser, implicit
**                           transaction granularity (i.e., fewer transactions
**                           when possible). If a session is being used only
**                           for an implicit transaction, it is made read-only
**                           to prevent unnecessary session flushing.
** 106 ECF 20090810   @43587 Fixed SessionListener notifications. Modified
**                           notification mechanism to allow more event types.
**                           Notify listeners when a transaction is about to
**                           be committed, in addition to session closing
**                           events.
** 107 ECF 20090811   @43599 Fixed regressions in #105 (@43418). Implicit
**                           transaction use count was not being properly set
**                           and used in some places. Fixed endTransaction(),
**                           which was nulling out active transaction too
**                           early.
** 108 CA  20090901   @43818 In initializeInstance, catch any StopCondition's
**                           and re-throw them (do not pack it in a generic
**                           runtime exception).
** 109 ECF 20090925   @44029 Temporarily disabled identity pool. There is a
**                           defect in the implementation of IdentityPool that
**                           can prevent primary keys from being distributed.
**                           When that is fixed, this should be re-enabled.
**                           Affected code was commented out from
**                           createIdentityManager().
** 110 CA  20091001   @44062 Modified executeSQLQuery() to leave the statement
**                           open; else, it will not be able to iterate 
**                           through the returned result set.
** 111 ECF 20091002   @44082 Re-enabled identity pool. Also made additional
**                           change to executeSQLQuery() to require the caller
**                           to push/pop the implicit transaction counter. We
**                           cannot do this in the method, as this could cause
**                           the transaction to be closed before the caller
**                           could process results.
*/

package com.goldencode.p2j.persist;

import java.lang.InstantiationException;
import java.lang.ref.*;
import java.io.*;
import java.sql.*;
import java.util.*;
import java.util.Date;
import org.apache.commons.logging.*;
import org.hibernate.*;
import org.hibernate.Session;
import org.hibernate.dialect.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.embed.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.hql.*;
import com.goldencode.p2j.persist.id.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

/**
 * The central point for fundamental persistence services in the P2J
 * environment.  This implementation is backed by the Hibernate ORM framework.
 * This class provides basic CRUD (create, read, update, delete) services,
 * as well as methods to begin/commit/rollback database-level transactions.
 * <p>
 * An instance of this class is retrieved from the factory method {@link
 * #get}.  One instance of this class is associated with a single, physical
 * database, per user context.
 * <p>
 * An instance of a concrete {@link LockManager} implementation is created
 * for each each instance of this class.  The implementation to use is
 * queried from the directory service, and defaults to the {@link
 * InMemoryLockManager} if no other implementation is specified in the
 * directory.
 * <p>
 * Transactions are explicitly managed with {@link #beginTransaction}, {@link
 * #commit} and {@link #rollback}.  Only one transaction per client context
 * may be active at a time.  These are database-level transactions, which may
 * or may not coincide with application-level transactions.
 * <p>
 * Most methods which perform some sort of query will open an implicit
 * transaction before the query is executed, but only if no transaction
 * already is active.  Implicit transactions are closed in one of two ways:
 * <ul>
 *   <li>a new explicit transaction is begun (in this case the implicit
 *       transaction is first committed);
 *   <li>a timeout period elapses with no additional query activity within the
 *       implicit transaction.
 * </ul>
 * In the latter case, the transaction is closed on a different thread than
 * that on which it was opened.  This timeout mechanism is used to avoid a
 * very noticeable performance penalty which is incurred when performing many
 * consecutive queries, if a transaction is committed immediately after each
 * one.
 * <p>
 * The exception to this rule are the variants of the <code>scroll</code>
 * method.  Because of the nature of <code>ScrollableResults</code>, a commit
 * should not take place until after calling code is finished using this
 * object.  This is necessary because fetching results may require additional
 * trips to the database within the transaction.  In this case, the
 * <code>scroll</code> methods themselves do not trigger implicit
 * transactions.  Rather, it is up to calling code (see {@link
 * ScrollingResults}) to bracket its use of the result set with calls to
 * {@link #pushImplicitTransaction()} and {@link #popImplicitTransaction()}.
 * <p>
 * The implicit transaction mechanism is intended only to support legacy use
 * cases.  <strong>External code which directly accesses query methods within
 * the <code>Persistence</code> class should always properly bracket its use
 * of these methods with explicit transactions</strong>.  Otherwise,
 * unpredictable behavior may result as implicit transactions are opened and
 * closed automatically.  Take special care to initialize any necessary
 * collection associations (for example, converted extent fields within DMOs
 * are implemented as lazily initialized, associated collections), before
 * ending a transaction.
 * <p>
 * Hibernate sessions are managed implicitly, transparently to application
 * code and even to other classes within the persistence framework.  No more
 * than one session per user context may be active at a time.  Sessions are
 * opened lazily as they are needed to fulfill database operations:  executing
 * a query, persisting a record, creating a temp table, etc.  They are closed
 * upon committing or rolling back a transaction, or if any exception is
 * caught during the processing of a database operation.  Upon opening a new
 * session, the current context's {@link BufferManager} instance is
 * interrogated for all active records.  These are the current records stored
 * in all open record buffers for the database matching the current
 * <code>Persistence</code> instance;  they represent detached records
 * previously associated with the last open session (if any).
 */
public class Persistence
{
   public static volatile int openTxCountT = 0;
   public static volatile int openTxCountP = 0;
   
   public static volatile int commitTxCountT = 0;
   public static volatile int commitTxCountP = 0;

   public static volatile int commitTxTime = 0;
   public static volatile int openTxTime = 0;
   
   public static volatile int pushTxCount = 0;
   
   public static volatile int popTxCount = 0;
   
   public static volatile int sessOpenCount = 0;
   public static volatile int sessCloseCount = 0;

   public static volatile int saveTxCount = 0;
   public static volatile int flushTxCount = 0;
   public static volatile int deleteTxCount = 0;
   public static volatile int deleteUTxCount = 0;

   public static volatile int saveTxTime = 0;
   public static volatile int flushTxTime = 0;
   public static volatile int deleteTxTime = 0;
   public static volatile int deleteUTxTime = 0;
   
   public static volatile int scrollCount = 0;
   public static volatile int load1Count = 0;
   public static volatile int load2Count = 0;
   public static volatile int listCount = 0;
   
   public static volatile int scrollTime = 0;
   public static volatile int load1Time = 0;
   public static volatile int load2Time = 0;
   public static volatile int listTime = 0;
   
   public static volatile int attachTime = 0;
   public static volatile int detachTime = 0;
   public static volatile int getQueryTime = 0;
   public static volatile int maybeFlushTime = 0;
   public static volatile int rollbackTime = 0;
   
   public static volatile int attachCount = 0;
   public static volatile int detachCount = 0;
   public static volatile int getQueryCount = 0;
   public static volatile int maybeFlushCount = 0;
   public static volatile int rollbackCount = 0;
   

   /** Well-known name of ID generator sequence */
   public static final String ID_GEN_SEQUENCE = "p2j_id_generator_sequence";
   
   /** Logger */
   private static final Log LOG = LogFactory.getLog(Persistence.class);
   
   /** Default lock manager implementation */
   private static final String DEFAULT_LOCK_MANAGER =
      "com.goldencode.p2j.persist.InMemoryLockManager";
   
   /** Default identity manager implementation */
   private static final String DEFAULT_IDENTITY_MANAGER =
      "com.goldencode.p2j.persist.id.SequenceIdentityManager";
   
   /** A transient mapping of JDBC connections to column data type maps */
   private static final Map<Connection, Map<ColumnKey, Integer>> columnTypes =
      new WeakHashMap<Connection, Map<ColumnKey, Integer>>(4);
   
   /** Flag indicating whether records are joined using foreign keys */
   private static boolean foreignKeysEnabled =
      Utils.getDirectoryNodeBoolean(null,
                                    "persistence/foreign-keys",
                                    false,
                                    false);
   
   /** Context local object to manage persistence variables */
   private final ContextLocal<Context> context = new ContextLocal<Context>()
   {
      protected Context initialValue() { return (new Context()); }
      protected void cleanup(Context value) { value.cleanup(); }
   };
   
   /** Database served by this object */
   private final Database database;
   
   /** Hibernate database dialect helper */
   private final Dialect dialect;
   
   /** Milliseconds a scrolling query may take before being reported */
   private final long scrollThreshold;
   
   /** Milliseconds a list query may take before being reported */
   private final long listThreshold;
   
   /** Milliseconds a load query may take before being reported */
   private final long loadThreshold;
   
   /** Milliseconds a load-by-id query may take before being reported */
   private final long loadByIDThreshold;
   
   /** Pessimistic lock manager */
   private LockManager lockManager;
   
   /** Identity manager */
   private IdentityManager identityManager;
   
   /** Sessions in use for special connection requests */
   private Map<Connection, Session> borrowedSessions =
      new HashMap<Connection, Session>();
   
   /**
    * Construct an instance for a particular, physical database, storing the
    * dialect class which Hibernate uses to create a portability layer.
    *
    * @param   database
    *          Database with which this instance is permanently associated.
    * @param   dialect
    *          Portability object used by Hibernate to issue commands to the
    *          database.
    */
   protected Persistence(Database database, Dialect dialect)
   {
      this.database = database;
      this.dialect = dialect;
      
      String dbName = database.getName();
      String spec = "database/%s/p2j/warn_threshold/enable";
      String path = String.format(spec, dbName);
      boolean warn = Utils.getDirectoryNodeBoolean(null, path, false, false);
      
      if (warn)
      {
         // Configure various warning thresholds to trigger reporting of slow
         // data access.
         spec = "database/%s/p2j/warn_threshold/data_access/%s";
         path = String.format(spec, dbName, "scroll_query");
         scrollThreshold = Utils.getDirectoryNodeInt(null, path, -1, false);
         path = String.format(spec, dbName, "list_query");
         listThreshold = Utils.getDirectoryNodeInt(null, path, -1, false);
         path = String.format(spec, dbName, "load_query");
         loadThreshold = Utils.getDirectoryNodeInt(null, path, -1, false);
         path = String.format(spec, dbName, "load_by_id_query");
         loadByIDThreshold = Utils.getDirectoryNodeInt(null, path, -1, false);
         
         if (LOG.isInfoEnabled())
         {
            StringBuilder buf = new StringBuilder("Warn thresholds for ");
            buf.append(database);
            buf.append(": scroll_query=");
            buf.append(scrollThreshold);
            buf.append("; list_query=");
            buf.append(listThreshold);
            buf.append(": load_query=");
            buf.append(loadThreshold);
            buf.append(": load_by_id_query=");
            buf.append(loadByIDThreshold);
            
            LOG.info(buf.toString());
         }
      }
      else
      {
         scrollThreshold = -1L;
         listThreshold = -1L;
         loadThreshold = -1L;
         loadByIDThreshold = -1L;
      }
   }
   
   /**
    * Get the instance of this class associated with the given, physical
    * database name, creating it first if necessary.  Newly created instances
    * are stored in a static cache.
    *
    * @param   name
    *          Name of physical database with which the instance is
    *          permanently associated.
    *          
    * @return  Persistence object associated with the given database.
    * 
    * @deprecated Use {@link PersistenceFactory#getInstance(String)} instead.
    */
   public static Persistence get(String name)
   {
      return PersistenceFactory.getInstance(name);
   }
   
   /**
    * Initialize various infrastructure components of the persistence
    * framework:
    * <ul>
    *   <li><code>BufferManager</code> which tracks open buffers and
    *       transactions.
    *   <li><code>ChangeBroker</code> which dispatches notifications about
    *       changes to DMO properties.
    *   <li><code>DatabaseManager</code> which configures the Hibernate
    *       environment for the databases to which the framework connects.
    * </ul>
    *
    * @throws  IllegalStateException
    *          if invoked more than once.
    * @throws  PersistenceException
    *          if any error is encountered initializing the persistence
    *          framework.
    *
    * @see     BufferManager#initialize
    * @see     ChangeBroker#initialize
    * @see     DatabaseManager#initialize
    */
   public static void initialize()
   throws PersistenceException
   {
      DataTypeHelper.initialize();
      BufferManager.initialize();
      ChangeBroker.initialize();
      DatabaseManager.initialize();
   }
   
   /**
    * Perform special handling on a query substitution parameter to avoid
    * errors during query execution.  This entails:
    * <ul>
    *   <li>Trimming all trailing spaces, tabs, line feeds, and carriage
    *       returns from any parameter of type <code>character</code>.  Any
    *       parameter affected is replaced with a new <code>character</code>
    *       instance at the same index position within <code>args</code>.
    *   <li>Clipping out of bounds date values to the range supported by the
    *       backing database's implementation of a date data type.  Any
    *       parameter affected is replaced with a new <code>date</code>
    *       instance at the same index position within <code>args</code>.
    * </ul>
    * <p>
    * Note:  this method has special processing for <code>null</code> bytes
    * embedded in <code>character</code> variables.  The text of the variable
    * is truncated starting at the first embedded <code>null</code> byte.
    * This means that the resulting substitution parameters are guaranteed
    * to never include <code>null</code> bytes.
    *
    * @param   arg
    *          A single substitution parameter.
    * @param   dialect
    *          Database dialect.
    * 
    * @return  The given parameter, processed as described above.
    */
   public static Object preprocessQueryParameter(Object arg,
                                                 P2JDialect dialect)
   {
      Object processed = arg;
      
      // Trim text arguments.
      if (arg instanceof character)
      {
         character c = (character) arg;
         
         if (!c.isUnknown())
         {
            // this obtains a clone of the character variable,
            // truncating it's text at any null byte that may be
            // embedded
            c = c.safeDuplicate();
            
            // this is safe because it doesn't alter the original
            // reference's state but rather replaces that reference
            processed = character.rightTrim(c, " \t\n\r");
         }
      }
      
      // "Clip" dates which are out of range.  This may result in
      // incorrect query behavior (hence the logged warning), but will
      // avoid catastrophic failure.
      else if (arg instanceof date)
      {
         date d = (date) arg;
         Date when = d.dateValue();
         
         Date low = dialect.getLowDate();
         if (low != null && when.before(low))
         {
            processed = new date(low);
            if (LOG.isDebugEnabled())
            {
               StringBuilder buf = new StringBuilder();
               buf.append("Query substitution date parameter ");
               buf.append(d.toStringMessage());
               buf.append(" was reset to the earliest date (");
               buf.append(low);
               buf.append(") supported by the backing database");
               LOG.debug(buf.toString());
            }
         }
         
         Date high = dialect.getHighDate();
         if (high != null && when.after(high))
         {
            processed = new date(high);
            if (LOG.isDebugEnabled())
            {
               StringBuilder buf = new StringBuilder();
               buf.append("Query substitution date parameter ");
               buf.append(d.toStringMessage());
               buf.append(" was reset to the latest date (");
               buf.append(high);
               buf.append(") supported by the backing database");
               LOG.debug(buf.toString());
            }
         }
      }
      
      return processed;
   }
   
   /**
    * Gather a list of <code>P2JIndex</code> objects which describe the
    * indices on the given table.  This method uses the database metadata,
    * if any, exposed by the database's JDBC driver.
    *
    * @param   database
    *          Database which contains <code>table</code>.
    * @param   table
    *          Name of table to inspect.
    * @param   unique
    *          <code>true</code> to gather information only on unique
    *          indices;  <code>false</code> to gather information on all
    *          indices on the table.
    *
    * @return  A list of <code>P2JIndex</code> objects, one per index found,
    *          or <code>null</code> if the JDBC metadata query throws
    *          <code>SQLException</code>, indicating a database error, or
    *          that the JDBC driver does not support the requested feature.
    *          An empty list is returned if the driver supported the query,
    *          but no indices were found.
    *
    * @throws  PersistenceException
    *          if there is an error getting a Hibernate session or determining
    *          an indexed column's data type.
    * @throws  UnsupportedOperationException
    *          if the JDBC driver cannot query database metadata.
    */
   static List<P2JIndex> queryIndexData(Database database,
                                        String table,
                                        boolean unique)
   throws PersistenceException
   {
      Connection conn = null;
      
      try
      {
         conn = DatabaseManager.checkOutConnection(database);
         DatabaseMetaData meta = conn.getMetaData();
         ResultSet rs = meta.getIndexInfo(null, null, table, unique, false);
         
         Map<String, P2JIndex> indices = retrieveIndexData(rs,
                                                           database,
                                                           table);
         
         return (new ArrayList<P2JIndex>(indices.values()));
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);

         throw new UnsupportedOperationException(
            "Error querying index metadata for " + database,
            exc);
      }
      finally
      {
         if (conn != null)
         {
            DatabaseManager.checkInConnection(conn, database);
         }
      }
   }
   
   /**
    * Given a the result set of an index metadata query (for a certain table),
    * this method is responsible for creating a map which will hold, for each
    * index, a {@link P2JIndex} object.
    * 
    * @param   rs
    *          The {@link ResultSet} instance used to retrieve the index 
    *          metadata.
    * @param   database
    *          Database which contains <code>table</code>.
    * @param   table
    *          The table for which the indexes are retrieved.
    * 
    * @return  A map with {@link P2JIndex} instances for each existing table
    *          index.
    * 
    * @throws  PersistenceException
    *          if any database error is encountered.
    */
   static Map<String, P2JIndex> retrieveIndexData(ResultSet rs,
                                                  Database database,
                                                  String table)
   throws PersistenceException
   {
      Map<String, P2JIndex> indices = new LinkedHashMap<String, P2JIndex>();
      
      try
      {
         while (rs.next())
         {
            if (rs.getShort("TYPE") == DatabaseMetaData.tableIndexStatistic)
            {
               continue;
            }
            
            String indexName = rs.getString("INDEX_NAME");
            if (indexName == null)
            {
               continue;
            }
            
            P2JIndex index = indices.get(indexName);
            if (index == null)
            {
               boolean nonUniq = rs.getBoolean("NON_UNIQUE");
               index = new P2JIndex(table, indexName, !nonUniq, false, false);
               indices.put(indexName, index);
            }
            
            String colName = rs.getString("COLUMN_NAME");
            
            // Determine case insensitivity in a dialect-specific way.
            Dialect dialect = DatabaseManager.getDialect(database);
            P2JDialect p2jDialect = (P2JDialect) dialect;
            boolean ignoreCase = p2jDialect.isCaseInsensitiveColumn(colName);
            
            // Extract root column name in a dialect-specific way.
            colName = p2jDialect.extractColumnName(colName);
            
            // Determine the data type of the indexed column.
            int dataType = getColumnDataType(database, table, colName);
            boolean isChar = (dataType == Types.VARCHAR);
            
            String sort = rs.getString("ASC_OR_DESC");
            boolean descending = "D".equals(sort);
            
            index.addComponent(colName,
                               descending,
                               isChar,
                               isChar && ignoreCase,
                               false);
         }
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);

         throw new PersistenceException(exc);
      }
      
      return indices;
   }
   
   /**
    * Detemine whether records are joined using foreign keys
    *
    * @return  <code>true</code> if records are joined using foreign keys.
    *          <code>false</code> if records are joined using legacy key
    *          fields.
    */
   static boolean isForeignKeysEnabled()
   {
      return foreignKeysEnabled;
   }
   
   /**
    * Use database metadata to determine the data type of a database column.
    * 
    * @param   database
    *          Target database.
    * @param   table
    *          Target table.
    * @param   column
    *          Target column.
    * 
    * @return  Data type of the given column, as one of the
    *          <code>java.sql.Types</code> constants.
    * 
    * @throws  PersistenceException
    *          if there is any database error, or if the metadata query fails
    *          to produce a result for the given column.
    */
   private static int getColumnDataType(Database database,
                                        String table,
                                        String column)
   throws PersistenceException
   {
      Connection conn = null;
      
      try
      {
         conn = DatabaseManager.checkOutConnection(database);
         DatabaseMetaData meta = null;
         
         if (DatabaseManager.isInitializing())
         {
            // Initialization optimization:
            // Try to gather up and cache all public schema column data types
            // in a single metadata query, rather than with many individual
            // queries.  The cache only lasts as long as the connection, which
            // is assumed to be shared across many of these data type
            // requests.
            Map<ColumnKey, Integer> types = columnTypes.get(conn);
            if (types == null)
            {
               types = new HashMap<ColumnKey, Integer>();
               columnTypes.put(conn, types);
            }
            
            ColumnKey key = new ColumnKey(table, column);
            Integer dataType = types.get(key);
            
            if (dataType == null)
            {
               meta = conn.getMetaData();
               // TODO:  is "public" schema portable?
               ResultSet rs = meta.getColumns(null, "public", "", "");
               
               while (rs.next())
               {
                  String nextTab = rs.getString("TABLE_NAME");
                  String nextCol = rs.getString("COLUMN_NAME");
                  int nextType = rs.getInt("DATA_TYPE");
                  ColumnKey nextKey = new ColumnKey(nextTab, nextCol);
                  if (key.equals(nextKey))
                  {
                     dataType = nextType;
                  }
                  types.put(nextKey, nextType);
               }
            }
            
            if (dataType != null)
            {
               return dataType;
            }
         }
         
         // Either we are not in the persistence initialization phase or the
         // all-in-one strategy failed.  The latter could happen if a JDBC
         // driver does not honor an empty string for the column name pattern
         // in the meta.getColumns(...) call above.  Try an individual
         // metadata query for the exact table and column name instead.  This
         // is less efficient, but may be necessary with some drivers.
         // TODO:  is "public" schema portable?
         if (meta == null)
         {
            meta = conn.getMetaData();
         }

         ResultSet rs = meta.getColumns(null, "public", table, column);
         if (rs.next())
         {
            return rs.getInt("DATA_TYPE");
         }
         
         throw new PersistenceException(
            "Unable to determine data type for column '" +
            table +
            ":" +
            column +
            "'");
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);

         throw new PersistenceException(exc);
      }
      finally
      {
         if (conn != null)
         {
            DatabaseManager.checkInConnection(conn, database);
         }
      }
   }
   
   /**
    * Get the database dialect helper object associated with the database
    * backing this object.
    *
    * @return  Database dialect helper.
    */
   public Dialect getDialect()
   {
      return dialect;
   }
   
   /**
    * Get the database information object associated with this persistence
    * service object.
    *
    * @return  Database information object.
    */
   public Database getDatabase()
   {
      return database;
   }
   
   /**
    * Get the lock manager associated with this <code>Persistence</code>
    * instance.  For temp tables, this will be <code>null</code>, since temp
    * table data is naturally compartmentalized and record locking is
    * meaningless in this case.  For permanent databases, this method will
    * return the lock manager created when this class was instantiated.
    *
    * @return  Lock manager instance associated with this persistence service
    *          object.
    */
   public LockManager getLockManager()
   {
      return lockManager;
   }
   
   /**
    * Get the identity manager associated with this <code>Persistence</code>
    * instance.  For temp tables, this will be <code>null</code>, since for
    * temp tables unique primary keys generated using {@link
    * TemporaryBuffer#nextPrimaryKey()}. For permanent databases, this
    * method will return the identity manager created when this class was
    * instantiated.
    *
    * @return  Identity manager instance associated with this persistence
    *          service object.
    */
   public IdentityManager getIdentityManager()
   {
      return identityManager;
   }
   
   /**
    * Execute an HQL query and get a scrollable cursor on the result set.
    * Records represented by the result set are not locked by this method.
    * This method creates a new, context local Hibernate session if necessary.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    *
    * @return  An object which provides access to a scrollable cursor on the
    *          result set.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public ScrollableResults scroll(String hql, Object[] values, Type[] types)
   throws PersistenceException
   {
      return scroll(hql, values, types, 0, 0);
   }
   
   /**
    * Execute an HQL query and get a scrollable cursor on the result set.
    * Records represented by the result set are not locked by this method.
    * This method creates a new, context local Hibernate session if necessary.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    *
    * @return  An object which provides access to a scrollable cursor on the
    *          result set.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public ScrollableResults scroll(String hql,
                                   Object[] values,
                                   Type[] types,
                                   int maxResults,
                                   int startOffset)
   throws PersistenceException
   {
      return scroll(hql,
                    values,
                    types,
                    maxResults,
                    startOffset,
                    ScrollMode.SCROLL_INSENSITIVE);
   }
   
   /**
    * Execute an HQL query and get a scrollable cursor on the result set.
    * Records represented by the result set are not locked by this method.
    * This method creates a new, context local Hibernate session if necessary.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   scrollMode
    *          Scroll mode which should apply to the result set.
    *
    * @return  An object which provides access to a scrollable cursor on the
    *          result set.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public ScrollableResults scroll(String hql,
                                   Object[] values,
                                   Type[] types,
                                   int maxResults,
                                   int startOffset,
                                   ScrollMode scrollMode)
   throws PersistenceException
   {
      long t1 = System.currentTimeMillis();
      
      boolean trace = LOG.isTraceEnabled();
      boolean warn = (trace || LOG.isWarnEnabled());
      long start = (warn ? System.currentTimeMillis() : 0L);

      Context local = context.get();
      
      try
      {
         Query query = getQuery(local,
                                hql,
                                maxResults,
                                startOffset,
                                values,
                                types,
                                false);
         if (!local.isFlushRequired(null))
         {
            query.setFlushMode(FlushMode.NEVER);
         }
         
         return query.scroll(scrollMode);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = message("error executing query [" + hql + "]");
         
         if (LOG.isWarnEnabled())
         {
            LOG.warn(msg);
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Details:", exc);
            }
         }
         
         throw new PersistenceException(msg, exc);
      }
      finally
      {
         if (warn && scrollThreshold >= 0L)
         {
            long elapsed = System.currentTimeMillis() - start;
            boolean slow = (elapsed > scrollThreshold);
            if (trace || slow)
            {
               StringBuilder buf = new StringBuilder("TIMING:SCROLL [");
               buf.append(elapsed);
               buf.append("] {");
               buf.append(startOffset);
               buf.append(":");
               buf.append(maxResults);
               buf.append("} ");
               buf.append(hql);
               String msg = message(buf.toString());
               
               if (slow)
               {
                  LOG.warn(msg);
               }
               else
               {
                  LOG.trace(msg);
               }
            }
         }
         
         long t2 = System.currentTimeMillis();
         scrollTime += (t2 - t1);
         ++scrollCount;
      }
   }
   
   /**
    * Execute an HQL query and return the results as a list.  If the query
    * returns multiple results per row, each element in the list will be an
    * array of <code>Object</code>s.
    * <p>
    * No locking is attempted.
    *
    * @param   hql
    *          HQL query statement passed to Hibernate.  This method makes no
    *          assumptions as to the types of object(s) returned.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   maxResults
    *          The maximum number of elements to be returned in the list.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   readOnly
    *          <code>true</code> to execute the query in read-only mode, else
    *          <code>false</code>.
    *
    * @return  A list of results, or <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query.
    */
   public List list(String hql,
                    Object[] values,
                    Type[] types,
                    int maxResults,
                    int startOffset,
                    boolean readOnly)
   throws PersistenceException
   {
      long t1 = System.currentTimeMillis();
      
      boolean trace = LOG.isTraceEnabled();
      boolean warn = (trace || LOG.isWarnEnabled());
      long start = (warn ? System.currentTimeMillis() : 0L);

      Context local = context.get();
      
      try
      {
         Query query = getQuery(local,
                                hql,
                                maxResults,
                                startOffset,
                                values,
                                types,
                                false);
         if (readOnly)
         {
            query.setReadOnly(readOnly);
         }
         if (!local.isFlushRequired(null))
         {
            query.setFlushMode(FlushMode.NEVER);
         }
         
         List results = query.list();
         
         return (results == null || results.isEmpty() ? null : results);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = message("error executing query [" + hql + "]");
         
         if (LOG.isWarnEnabled())
         {
            LOG.warn(msg);
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Details:", exc);
            }
         }
         
         throw new PersistenceException(msg, exc);
      }
      finally
      {
         if (warn && listThreshold >= 0L)
         {
            long elapsed = System.currentTimeMillis() - start;
            boolean slow = (elapsed > listThreshold);
            if (trace || slow)
            {
               StringBuilder buf = new StringBuilder("TIMING:LIST [");
               buf.append(elapsed);
               buf.append("] {");
               buf.append(startOffset);
               buf.append(":");
               buf.append(maxResults);
               buf.append("} ");
               buf.append(hql);
               String msg = message(buf.toString());
               
               if (slow)
               {
                  LOG.warn(msg);
               }
               else
               {
                  LOG.trace(msg);
               }
            }
         }
         
         long t2 = System.currentTimeMillis();
         listTime += (t2 - t1);
         ++listCount;
      }
   }
   
   /**
    * Execute an HQL query and load a single record from the database or from
    * the session-level cache.  The record will be locked with the specified
    * lock type upon return from this method.  Note that the lock refers to a
    * runtime-level lock, not a database-level lock.  The record is
    * <i>not</i> automatically loaded into the record buffer provided.  The
    * <code>buffer</code> argument is used only to extract information needed
    * to perform the query.
    * <p>
    * <b>Note</b>:  if a non-projection query (i.e., "from Foo...") is passed
    * to this method, we assume the caller does not expect to acquire a lock.
    * This optimized path is intended to support no-lock and temp-table
    * queries where a lock is meaningless.  If a lock is required, use an
    * ID-only projection query statement instead (i.e., "select foo.id from
    * Foo as foo...").  In this case, the id will be retrieved first, then a
    * lock acquired, then the DMO will be fetched in a separate step.  This
    * is less efficient, but necessary to ensure other sessions modifications
    * made during lock acquisition are properly reflected in the object
    * returned.
    *
    * @param   buffer
    *          Record buffer which holds information about the DMO class type
    *          and name of the table being queried.
    * @param   hql
    *          HQL query statement passed to Hibernate. This may be either an
    *          ID-only projection query or a full query (see above).
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   unique
    *          <code>true</code> if no more than one result is expected from
    *          this query;  <code>false</code> if the query could result in
    *          more than one record.
    * @param   updateLock
    *          <code>true</code> to update the current lock state;
    *          <code>false</code> to leave the current lock state unchanged.
    *
    * @return  A single DMO instance representing the query result, or
    *          <code>null</code> if no result was found.
    *
    * @throws  LockUnavailableException
    *          if a record was found but its lock could not be acquired.
    * @throws  PersistenceException
    *          if there was an error executing the query or if a unique query
    *          (i.e., <code>unique == true</code>) returned multiple results,
    */
   public Persistable load(RecordBuffer buffer,
                           String hql,
                           Object[] values,
                           Type[] types,
                           LockType lockType,
                           boolean unique,
                           boolean updateLock)
   throws LockUnavailableException,
          PersistenceException
   {
      long t1 = System.currentTimeMillis();
      
      boolean trace = LOG.isTraceEnabled();
      boolean warn = (trace || LOG.isWarnEnabled());
      long start = (warn ? System.currentTimeMillis() : 0L);

      Context local = context.get();
      
      boolean inlineLimit = (values == null || values.length == 0);
      Serializable id = null;
      Persistable result = null;
      String table = buffer.getTable();
      LockType oldLock = null;
      RecordLockContext lockContext = local.getRecordLockContext();
      P2JDialect p2jDialect = (P2JDialect) dialect;
      
      try
      {
         // The inlineLimit flag is a crude test to detect a "simple" query:
         // no substitution parameters.  We want to avoid inlining the limit
         // on more complex statements, as the intended performance benefit
         // may backfire in these cases.
         p2jDialect.inlineLimit(inlineLimit);
         
         // We have to give queries which are supposed to return a unique
         // result enough rope to hang themselves, so set max results to 2
         // for these.  All other queries are limited to 1 result.
         Query query = getQuery(local,
                                hql,
                                unique ? 2 : 1,
                                0,
                                values,
                                types,
                                false);
         
         // Disable flushing if possible, for better performance.
         Class<?> dmoClass = buffer.getDMOImplementationClass();
         if (!local.isFlushRequired(dmoClass.getName()))
         {
            query.setFlushMode(FlushMode.NEVER);
         }
         
         // Reset ambiguous flag in buffer.
         buffer.setAmbiguous(false);
         
         // Find the record.
         Object o = query.uniqueResult();
         
         // Reset limit inlining if necessary.
         if (inlineLimit)
         {
            p2jDialect.inlineLimit(false);
            inlineLimit = false;
         }
         
         if (o == null)
         {
            // No hit.
            return null;
         }
         
         if (o instanceof Persistable)
         {
            // This is the fast path where we assume we retrieved the DMO
            // directly instead of the ID because we don't have to lock
            // (we might need to unlock, however).
            // This is the case for temp-tables and no-lock queries.
            result = (Persistable) o;
            
            // Lock record if required, remembering old lock type in case we
            // need to roll back on error.
            if (updateLock)
            {
               id = result.getId();
               RecordIdentifier ident = new RecordIdentifier(table, id);
               oldLock = queryLock(ident);
               if (oldLock.compareTo(lockType) != 0)
               {
                  lockContext.lock(ident, lockType, buffer);
               }
            }
         }
         else
         {
            // If we got here, we're dealing with a primary key query.
            id = (Serializable) o;
            RecordIdentifier ident = new RecordIdentifier(table, id);
            
            // Lock record if required, remembering old lock type in case we
            // need to roll back on error.
            if (updateLock)
            {
               oldLock = queryLock(ident);
               lockContext.lock(ident, lockType);
            }
            else
            {
               lock(lockType, ident, false);
            }
            
            // We use get() instead of load() here because we don't want to
            // return a proxy and have to take another hit retrieving the data
            // almost immediately after this method returns, which is the common
            // case for dynamic record retrieval.
            Session session = local.getSession();
            result = (Persistable) session.get(dmoClass, id);
            
            // Ensure that if we are acquiring a lock, we have the latest
            // version of the record from the database.  The Hibernate API
            // advertises that Session.get() with a LockMode of READ should
            // bypass both levels of cache, but debugging indicates that the
            // record is always taken from the session cache if present.  So,
            // we brute force a refresh instead.
            if (result != null &&
                updateLock &&
                oldLock == LockType.NONE &&
                lockType != LockType.NONE)
            {
               session.refresh(result);
            }
         }
      }
      catch (NonUniqueResultException exc)
      {
         // This is recoverable, no need to close session.
         uniqueResultViolation(buffer, exc);
      }
      catch (UnresolvableObjectException exc)
      {
         // Catching this exception indicates we were holding a stale copy of
         // a record that was deleted in the meantime.  This can happen in
         // bulk delete cases, or when a preselect query is holding an old
         // reference.  This is not a fatal error, it just means that the
         // refresh failed and the result is null.
         result = null;
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = message("error loading record");
         
         if (warn)
         {
            LOG.warn(msg + System.getProperty("line.separator") + hql);
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Details:", exc);
            }
         }
         
         throw new PersistenceException(msg, exc);
      }
      finally
      {
         // Reset limit inlining if (still) necessary.
         if (inlineLimit)
         {
            p2jDialect.inlineLimit(false);
         }
         
         if (updateLock      &&
             id != null      &&
             result == null  &&
             oldLock != null &&
             oldLock.compareTo(lockType) != 0)
         {
            // Reset lock to its previous state.  This is required if another
            // thread deleted the record after we got its ID but before we
            // retrieved its data.
            try
            {
               RecordIdentifier ident = new RecordIdentifier(table, id);
               lockContext.lock(ident, oldLock.toNoWaitVariant());
            }
            catch (LockUnavailableException exc)
            {
               // OK to ignore.
            }
         }
         
         if (warn && loadThreshold >= 0L)
         {
            long elapsed = System.currentTimeMillis() - start;
            boolean slow = (elapsed > loadThreshold);
            if (trace || slow)
            {
               StringBuilder buf = new StringBuilder("TIMING:LOAD ");
               buf.append(result != null ? "+" : "-");
               buf.append("[");
               buf.append(elapsed);
               buf.append("] ");
               buf.append(hql);
               String msg = message(buf.toString());
               
               if (slow)
               {
                  LOG.warn(msg);
               }
               else
               {
                  LOG.trace(msg);
               }
            }
         }
         
         long t2 = System.currentTimeMillis();
         load2Time += (t2 - t1);
         ++load2Count;
      }
      
      return result;
   }
   
   /**
    * Load a single record from the database or from the session-level cache.
    * The record will be locked with the specified lock type upon return from
    * this method.  Note that the lock refers to a runtime-level lock, not a
    * database-level lock.  The record is <i>not</i> automatically loaded into
    * the record buffer provided.  The <code>buffer</code> argument is used
    * only to extract information needed to perform the locking and loading.
    *
    * @param   buffer
    *          Record buffer which holds information about the DMO class type
    *          and name of the table being queried.
    * @param   id
    *          Primary key of the record to be loaded into the buffer.  May
    *          not be <code>null</code>.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   updateLock
    *          <code>true</code> to update the current lock state;
    *          <code>false</code> to leave the current lock state unchanged.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  LockUnavailableException
    *          if a record was found but its lock could not be acquired.
    * @throws  PersistenceException
    *          if there was an error loading the DMO from the Hibernate
    *          session and/or database.
    */
   public Persistable load(RecordBuffer buffer,
                           Serializable id,
                           LockType lockType,
                           boolean updateLock)
   throws LockUnavailableException,
          PersistenceException
   {
      String table = buffer.getTable();
      String entity = buffer.getEntityName();
      
      return load(table, entity, id, lockType, updateLock);
   }
   
   /**
    * Load a single record from the database or from the session-level cache.
    * The record will be locked with the specified lock type upon return from
    * this method.  Note that the lock refers to a runtime-level lock, not a
    * database-level lock.
    *
    * @param   table
    *          Table from which record is to be read.
    * @param   dmoClass
    *          DMO implementation class of record being read.
    * @param   id
    *          Primary key of the record to be loaded into the buffer.  May
    *          not be <code>null</code>.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   updateLock
    *          <code>true</code> to update the current lock state;
    *          <code>false</code> to leave the current lock state unchanged.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  LockUnavailableException
    *          if a record was found but its lock could not be acquired.
    * @throws  PersistenceException
    *          if there was an error loading the DMO from the Hibernate
    *          session and/or database.
    */
   public Persistable load(String table,
                           Class<?> dmoClass,
                           Serializable id,
                           LockType lockType,
                           boolean updateLock)
   throws LockUnavailableException,
          PersistenceException
   {
      return load(table, dmoClass.getName(), id, lockType, updateLock);
   }
   
   /**
    * Load a single record from the database or from the session-level cache.
    * The record will be locked with the specified lock type upon return from
    * this method.  Note that the lock refers to a runtime-level lock, not a
    * database-level lock.
    *
    * @param   table
    *          Table from which record is to be read.
    * @param   entityName
    *          Name of DMO implementation class of record being read.
    * @param   id
    *          Primary key of the record to be loaded into the buffer.  May
    *          not be <code>null</code>.
    * @param   lockType
    *          Type of lock the lock manager should register on the resulting
    *          record for the current user context.
    * @param   updateLock
    *          <code>true</code> to update the current lock state;
    *          <code>false</code> to leave the current lock state unchanged.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  LockUnavailableException
    *          if a record was found but its lock could not be acquired.
    * @throws  PersistenceException
    *          if there was an error loading the DMO from the Hibernate
    *          session and/or database.
    */
   public Persistable load(String table,
                           String entityName,
                           Serializable id,
                           LockType lockType,
                           boolean updateLock)
   throws LockUnavailableException,
          PersistenceException
   {
      long t1 = System.currentTimeMillis();
      
      boolean trace = LOG.isTraceEnabled();
      boolean warn = (trace || LOG.isWarnEnabled());
      long start = (warn ? System.currentTimeMillis() : 0L);
      
      Persistable result = null;
      Context local = context.get();
      LockType oldLock = null;
      RecordIdentifier ident = new RecordIdentifier(table, id);
      RecordLockContext lockContext = local.getRecordLockContext();
      
      try
      {
         // Lock record if required, remembering old lock type in case we
         // need to roll back on error.
         if (updateLock)
         {
            oldLock = queryLock(ident);
            lockContext.lock(ident, lockType);
         }
         else
         {
            lock(lockType, ident, false);
         }
         
         Session session = local.getSession();
         
         if (local.isFlushRequired(entityName))
         {
            session.flush();
         }
         
         // Use get() instead of load() here, to ensure that if the record
         // has been deleted, we get a null back, rather than an invalid
         // proxy.
         result = (Persistable) session.get(entityName, id);
         
         // Ensure that if we are acquiring a lock, we have the latest
         // version of the record from the database.  The Hibernate API
         // advertises that Session.get() with a LockMode of READ should
         // bypass both levels of cache, but debugging indicates that the
         // record is always taken from the session cache if present.  So,
         // we brute force a refresh instead.
         if (result != null &&
             updateLock &&
             oldLock == LockType.NONE &&
             lockType != LockType.NONE)
         {
            session.refresh(result);
         }
         
         return result;
      }
      catch (UnresolvableObjectException exc)
      {
         // Catching this exception indicates we were holding a stale copy of
         // a record that was deleted in the meantime.  This can happen in
         // bulk delete cases, or when a preselect query is holding an old
         // reference.  This is not a fatal error, it just means that the
         // refresh failed and the result is null.
         return null;
      }
      catch (HibernateException exc)
      {
         try
         {
            local.closeSession(true);
         }
         finally
         {
            if (updateLock      &&
                oldLock != null &&
                oldLock.compareTo(lockType) != 0)
            {
               // Reset lock to its previous state.  This is required if
               // another thread deleted the record after we got its ID but
               // before we retrieved its data.
               try
               {
                  lockContext.lock(ident, oldLock.toNoWaitVariant());
               }
               catch (LockUnavailableException luExc)
               {
                  // OK to ignore.
               }
            }
            
            DBUtils.handleException(database, exc);
         }
         
         String msg = message("error loading record " + table + ":" + id);
         
         if (LOG.isWarnEnabled())
         {
            LOG.warn(msg);
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Details:", exc);
            }
         }
         
         throw new PersistenceException(msg, exc);
      }
      finally
      {
         if (warn && loadByIDThreshold >= 0L)
         {
            long elapsed = System.currentTimeMillis() - start;
            boolean slow = (elapsed > loadByIDThreshold);
            if (trace || slow)
            {
               StringBuilder buf = new StringBuilder("TIMING:LOAD_BY_ID ");
               buf.append(result != null ? "+" : "-");
               buf.append("[");
               buf.append(elapsed);
               buf.append("] ");
               buf.append(table);
               buf.append(":");
               buf.append(id);
               String msg = message(buf.toString());
               
               if (slow)
               {
                  LOG.warn(msg);
               }
               else
               {
                  LOG.trace(msg);
               }
            }
         }
         long t2 = System.currentTimeMillis();
         load1Time += (t2 - t1);
         ++load1Count;
      }
   }
   
   /**
    * Load a single record from the database or from the session-level cache.
    * No locking is attempted, the assumption being that any required lock
    * previously was acquired.
    *
    * @param   ident
    *          Record identifier of the record to be loaded.  Consists of the
    *          primary key and the fully qualified DMO entity name.
    *
    * @return  A single DMO instance loaded from the database, or
    *          <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error creating a Hibernate session or getting
    *          the requested record.
    */
   public Persistable quickLoad(RecordIdentifier ident)
   throws PersistenceException
   {
      Context local = context.get();
      String entity = ident.getTable();
      Serializable id = ident.getRecordID();
      Session session = local.getSession();
      
      try
      {
         if (local.isFlushRequired(entity))
         {
            session.flush();
         }
         
         // Use get() instead of load() here, to ensure that if the record
         // has been deleted, we get a null back, rather than an invalid
         // proxy.
         Persistable result = (Persistable) session.get(entity, id);
         
         return result;
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = message("error executing quick load [" +
                              entity +
                              ":" +
                              id +
                              "]");
         
         if (LOG.isWarnEnabled())
         {
            LOG.warn(msg);
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Details:", exc);
            }
         }
         
         throw new PersistenceException(msg, exc);
      }
   }
   
   /**
    * Execute an arbitrary SQL statement which does not produce a result set.
    * This method relies on Hibernate to obtain the JDBC connection, but
    * otherwise bypasses Hibernate and executes the SQL directly, using JDBC.
    * The Hibernate session is flushed before this statement is executed, to
    * ensure any pending changes are synchronized with the database first.
    *
    * @param   sql
    *          SQL statement text.
    *
    * @throws  PersistenceException
    *          if any error occurs executing the statement.
    */
   public void executeSQL(String sql)
   throws PersistenceException
   {
      long start = 0L;
      boolean trace = LOG.isTraceEnabled();
      if (trace)
      {
         start = System.currentTimeMillis();
      }

      Context local = context.get();
      Statement stmt = null;
      Connection conn = null;
      
      try
      {
         Session session = local.getSession();
         
         local.pushImplicitTransaction();
         
         if (local.isFlushRequired(null))
         {
            session.flush();
         }
         conn = session.connection();
         stmt = conn.createStatement();
         boolean rc = stmt.execute(sql);
         if (LOG.isDebugEnabled())
         {
            StringBuilder buf = new StringBuilder("[");
            buf.append(sql);
            buf.append("] rc = ");
            buf.append(rc);
            LOG.debug(buf.toString());
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            "Error executing SQL statement:  " + sql, exc);
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            "Error executing SQL statement:  " + sql, exc);
      }
      finally
      {
         try
         {
            if (stmt != null && !stmt.isClosed())
            {
               stmt.close();
            }
         }
         catch (SQLException exc)
         {
            DBUtils.handleException(database, exc);
            
            if (LOG.isWarnEnabled())
            {
               LOG.warn(message("error closing statement"), exc);
            }
         }
         finally
         {
            local.popImplicitTransaction();
         }
         
         if (trace)
         {
            long elapsed = System.currentTimeMillis() - start;
            StringBuilder buf = new StringBuilder("TIMING:SQL [");
            buf.append(elapsed);
            buf.append("] ");
            buf.append(sql);
            
            LOG.trace(message(buf.toString()));
         }
      }
   }
   
   /**
    * Execute an arbitrary SQL query.  This method relies on Hibernate to
    * obtain the JDBC connection, but otherwise bypasses Hibernate and
    * executes the SQL directly, using JDBC.  The Hibernate session is flushed
    * before this statement is executed, to ensure any pending changes are
    * synchronized with the database first.
    * <p>
    * It is the responsability of the calling code to close the returned
    * result set.  In addition, if this method is called outside of a database
    * transaction, the caller must invoke {@link #pushImplicitTransaction()}
    * before calling this method and {@link #popImplicitTransaction()} after
    * fully processing the result set returned by this method.  This will
    * ensure a transaction is opened and properly closed if necessary.
    *
    * @param   sql
    *          SQL query text.
    * 
    * @return  Result set produced by query.
    *
    * @throws  PersistenceException
    *          if any error occurs executing the statement.
    */
   public ResultSet executeSQLQuery(String sql)
   throws PersistenceException
   {
      long start = 0L;
      boolean trace = LOG.isTraceEnabled();
      if (trace)
      {
         start = System.currentTimeMillis();
      }

      Context local = context.get();
      Statement stmt = null;
      
      try
      {
         Session session = local.getSession();
         
         if (local.isFlushRequired(null))
         {
            session.flush();
         }
         Connection conn = session.connection();
         stmt = conn.createStatement();
         
         return stmt.executeQuery(sql);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            "Error executing SQL statement:  " + sql, exc);
      }
      catch (SQLException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            "Error executing SQL statement:  " + sql, exc);
      }
      finally
      {
         if (trace)
         {
            long elapsed = System.currentTimeMillis() - start;
            StringBuilder buf = new StringBuilder("TIMING:SQLQ [");
            buf.append(elapsed);
            buf.append("] ");
            buf.append(sql);
            
            LOG.trace(message(buf.toString()));
         }
      }
   }
   
   /**
    * Attempt to obtain the specified lock type on a particular record.  If
    * the lock cannot be obtained immediately, the current thread will block
    * until the lock becomes available and is obtained, unless the requested
    * lock type is a "no-wait" lock.  In the latter case, an exception is
    * raised, indicating that the lock is not available.  The current thread
    * continues once the lock is obtained.
    * <p>
    * A lock currently held can be released by specifying
    * <code>LockType.NONE</code> for the <code>lockType</code> parameter.
    * <p>
    * The normal return of this method indicates that the lock request (or
    * lack thereof) completed successfully, and that the current context now
    * holds a lock of the requested type.
    * <p>
    * NOTE:  legacy persistence framework code should access this method only
    * via a {@link RecordLockContext} <code>lock</code> method variant,
    * unless simply testing whether a lock is available (i.e., <code>update
    * == false</code>).  This ensures that record locks shared across record
    * buffers within a client context are managed properly.  Non-legacy code
    * may access this API directly.
    *
    * @param   lockType
    *          Type of lock to be obtained.  <code>LockType.NONE</code> is
    *          used to release an existing lock, and to continue with no
    *          lock.
    * @param   ident
    *          ID which uniquely identifies the record being locked.
    * @param   update
    *          <code>true</code> if the lock state for the current record
    *          should be updated;  else <code>false</code>.
    *
    * @throws  LockUnavailableException
    *          if a "no-wait" lock cannot be acquired immediately.
    *
    * @see     #getRecordLockContext
    */
   public void lock(LockType lockType, RecordIdentifier ident, boolean update)
   throws LockUnavailableException
   {
      if (lockManager != null)
      {
         lockManager.lock(lockType, ident, update);
      }
   }
   
   /**
    * Query the lock type currently held by the current context for the
    * specified database record.
    * <p>
    * This method will always return the <b>non-</b><code>NO_WAIT</code>
    * variants of a lock type.  That is, even if the actual lock type is
    * <code>LockType.EXCLUSIVE_NO_WAIT</code> or
    * <code>LockType.SHARE_NO_WAIT</code>, the simpler types of
    * <code>LockType.EXCLUSIVE</code> and <code>LockType.SHARE</code>,
    * respectively, are returned.  The lack of a lock returns
    * <code>LockType.NONE</code> rather than <code>null</code>.
    *
    * @param   ident
    *          ID which uniquely identifies the record being queried.
    *
    * @return  The lock type currently held by the current context, never
    *          <code>null</code>.
    */
   public LockType queryLock(RecordIdentifier ident)
   {
      if (lockManager == null)
      {
         return LockType.NONE;
      }
      
      return lockManager.queryLock(ident);
   }

   /**
    * Obtain the next available primary key ID on a per-table basis.
    *
    * @param   table
    *          Table for which next primary key will be returned. 
    *
    * @return  The next available ID for the given table.
    * 
    * @throws  PersistenceException
    *          if there is an error determining the next primary key ID.
    * @throws  IllegalStateException
    *          if identity manager was not initialized.
    */
   public Serializable nextPrimaryKey(String table)
   throws PersistenceException
   {
      if (identityManager != null)
      {
         return identityManager.nextPrimaryKey(table);
      }
      else
      {
         throw new IllegalStateException(
            "identity manager should be initialized");
      }
   }
   
   /**
    * Persist the specified object to the backing database session, using the
    * given ID.  It is assumed this method will only be invoked within a
    * database transaction.
    *
    * @param   dmo
    *          DMO to be persisted.
    * @param   id
    *          Primary key to associate with the persisted DMO.
    *
    * @throws  PersistenceException
    *          if a problem occurs saving the object in the current session.
    */
   public void save(Persistable dmo, Serializable id)
   throws PersistenceException
   {
      long t1 = System.currentTimeMillis();
      
      Context local = context.get();
      boolean commit = false;
      
      try
      {
         commit = beginTransaction(local);
         
         if (dmo.getId() == null)
         {
            dmo.setId((Long) id);
         }
         
         local.getSession().save(dmo);
         
         if (LOG.isDebugEnabled())
         {
            StringBuilder buf = new StringBuilder("Persisted DMO (");
            buf.append(dmo);
            buf.append(") with ID ");
            buf.append(id);
            
            LOG.debug(message(buf.toString()));
         }
      }
      catch (HibernateException exc)
      {
         commit = false;
         
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            message("Error persisting object"),
            exc);
      }
      finally
      {
         if (commit)
         {
            local.commit(true);
         }
         long t2 = System.currentTimeMillis();
         
         saveTxTime += (t2 - t1);
         ++saveTxCount;
      }
   }
   
   /**
    * Ensure the specified object is being managed by the session associated
    * with the backing database in the current context.
    *
    * @param   object
    *          DMO to be persisted.
    *
    * @throws  PersistenceException
    *          if a problem occurs attaching the object to the current
    *          session.
    */
   public void attach(Object object)
   throws PersistenceException
   {
      long t1 = System.currentTimeMillis();
      
      Context local = context.get();

      try
      {
         Session session = local.getSession();
         if (!session.contains(object))
         {
            session.lock(object, LockMode.NONE);
         }
      }
      catch (HibernateException exc)
      {
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            message("error attaching record to database session"),
            exc);
      }
      finally
      {
         long t2 = System.currentTimeMillis();
         
         attachTime += (t2 - t1);
         ++attachCount;
      }
   }
   
   /**
    * Evict the specified object from the session associated with the backing
    * database in the current context.  This is a no-op if there is no active
    * session in the context.
    *
    * @param   object
    *          Persistent entity to be evicted from the session.
    *
    * @throws  PersistenceException
    *          if a problem occurs evicting the object from the current
    *          session.
    */
   public void detach(Object object)
   throws PersistenceException
   {
      long t1 = System.currentTimeMillis();
      
      Context local = context.get();
      boolean flush = false;
      boolean commit = false;

      try
      {
         Session session = local.getSessionNoCreate();
         if (session != null && session.contains(object))
         {
            Class<?> dmoClass = object.getClass();
            dmoClass = DatabaseManager.normalizeDMOClass(dmoClass);
            flush = local.isFlushRequired(dmoClass.getName());
            
            if (flush)
            {
               commit = beginTransaction(local);
               session.flush();
            }
            
            if (LOG.isTraceEnabled())
            {
               Serializable id = ((Persistable) object).getId();
               String table = DatabaseManager.getBackingTableName(database,
                                                                  dmoClass);
               RecordIdentifier ident = new RecordIdentifier(table, id);
               LockType lt = null;
               if (table != null)
               {
                  lt = queryLock(ident);
               }
               
               StringBuilder buf = new StringBuilder();
               buf.append(flush ? "Flushing " : "NOT flushing ");
               buf.append(dmoClass.getName());
               buf.append(" before eviction");
               LOG.trace(buf.toString());
               
               buf.setLength(0);
               buf.append("Evicting ");
               buf.append(ident);
               buf.append(" (");
               buf.append(lt);
               buf.append(")");
               LOG.trace(buf.toString());
            }
            
            session.evict(object);
         }
      }
      catch (HibernateException exc)
      {
         commit = false;
         
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            message("error evicting object from database session"),
            exc);
      }
      finally
      {
         if (commit)
         {
            local.commit(true);
         }
         
         long t2 = System.currentTimeMillis();
         detachTime += (t2 - t1);
         ++detachCount;
      }
   }
   
   /**
    * Delete the persisted object from the specified database.
    * <p>
    * This action must take place within a database transaction, so this
    * method attempts to begin one and commit it, if one is not already
    * active.
    *
    * @param   object
    *          Persistent entity to be deleted.
    *
    * @throws  PersistenceException
    *          if a problem occurs deleting the object.
    */
   public void delete(Object object)
   throws PersistenceException
   {
      long t1 = System.currentTimeMillis();
      
      Context local = context.get();
      boolean commit = false;
      
      try
      {
         commit = beginTransaction(local);
         local.getSession().delete(object);
      }
      catch (HibernateException exc)
      {
         commit = false;
         
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            message("error deleting persistent object"),
            exc);
      }
      finally
      {
         if (commit)
         {
            local.commit(true);
         }
         
         long t2 = System.currentTimeMillis();
         deleteTxTime += (t2 - t1);
         ++deleteTxCount;
      }
   }
   
   /**
    * Perform a bulk delete or update of all objects which meet the specified 
    * criteria. The criteria must indicate a single DMO type only.
    * <p>
    * This action must take place within a database transaction, so this
    * method attempts to begin one and commit it, if one is not already
    * active.
    * <p>
    * This method should only be used for temp-tables (where an exclusive
    * lock is irrelevant) or in cases where the current context already holds
    * locks on all records which meet the specified criteria.
    *
    * @param   hql
    *          Query string which defines the criteria which determine which
    *          records are deleted or updated.
    * @param   values
    *          Query substitution parameters, if any.
    * @param   types
    *          Types of the query substitution parameters, if any.
    *
    * @return  The number of entities deleted or updated.
    *
    * @throws  PersistenceException
    *          if a problem occurs performing the bulk delete or update.
    */
   public int deleteOrUpdate(String hql, Object[] values, Type[] types)
   throws PersistenceException
   {
      boolean trace = LOG.isTraceEnabled();
      long start = System.currentTimeMillis();
      
      Context local = context.get();
      boolean commit = false;
      
      try
      {
         Query query = getQuery(local, hql, 0, 0, values, types, true);
         commit = beginTransaction(local);
         int count = query.executeUpdate();
         
         if (LOG.isDebugEnabled())
         {
            StringBuilder buf = new StringBuilder("[");
            buf.append(hql);
            buf.append("] count = ");
            buf.append(count);
            LOG.debug(buf.toString());
         }
         
         return count;
      }
      catch (HibernateException exc)
      {
         commit = false;
         
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         throw new PersistenceException(
            message("error performing bulk delete [" + hql + "]"),
            exc);
      }
      finally
      {
         if (commit)
         {
            local.commit(true);
         }
         
         if (trace)
         {
            long elapsed = System.currentTimeMillis() - start;
            StringBuilder buf = new StringBuilder("TIMING:DELETE [");
            buf.append(elapsed);
            buf.append("] ");
            buf.append(hql);
            
            LOG.trace(message(buf.toString()));
         }
         
         deleteUTxTime += (System.currentTimeMillis() - start);
         ++deleteUTxCount;
      }
   }
   
   /**
    * Begin a new Hibernate transaction and store the instance in the local
    * context.  A new database session is created if one does not already
    * exist in the current context.
    *
    * @return  <code>true</code> if a new transaction was begun;
    *          <code>false</code> if there was already an active transaction.
    *
    * @throws  PersistenceException
    *          if there is an error beginning the new transaction.
    */
   public boolean beginTransaction()
   throws PersistenceException
   {
      Context local = context.get();
      
      return beginTransaction(local);
   }
   
   /**
    * Commit the current Hibernate transaction and close the current session.
    * If the transaction cannot be committed, roll it back and raise an error.
    *
    * @throws  IllegalStateException
    *          if no transaction is active in the current session;  this
    *          represents a programming error.
    * @throws  PersistenceException
    *          if there is an error committing the transaction.
    */
   public void commit()
   throws PersistenceException
   {
      Context local = context.get();
      local.commit(true);
   }
      
   /**
    * Roll back the current Hibernate transaction and close the current
    * session.  If the transaction cannot be rolled back, raise an error.
    *
    * @throws  IllegalStateException
    *          if no transaction is active in the current session;  this
    *          represents a programming error.
    * @throws  PersistenceException
    *          if there is an error rolling back the transaction.
    */
   public void rollback()
   throws PersistenceException
   {
      Context local = context.get();
      local.rollback(true, false);
   }
   
   /**
    * Flush a batch of records to the database, typically while inserting
    * records in bulk.  After the flush, the session is cleared of all
    * records, to prevent a progressive memory leak.
    * <p>
    * This request bypasses the current client context's setting as to
    * whether flushing is enabled.
    * <p>
    * It is expected that this method will only be called within an active
    * transaction.  If no session is active, this method will raise an
    * exception.
    *
    * @throws  IllegalStateException
    *          if no session currently is active.
    * @throws  PersistenceException
    *          if an error occurs during session flush.
    *
    * @see     #enableFlush
    */
   public void flushBatch()
   throws PersistenceException
   {
      flush(true);
   }
   
   /**
    * Flush the current session, which synchronizes the in-memory state of
    * DMOs to the database.
    * <p>
    * This request bypasses the current client context's setting as to
    * whether flushing is enabled.
    * <p>
    * It is expected that this method will only be called within an active
    * transaction.  If no session is active, this method will raise an
    * exception.
    *
    * @throws  IllegalStateException
    *          if no session currently is active.
    * @throws  PersistenceException
    *          if an error occurs during session flush.
    *
    * @see     #enableFlush
    */
   public void flush()
   throws PersistenceException
   {
      flush(false);
   }
   
   /**
    * Indicate whether an explicit, database-level transaction currently is
    * open.
    *
    * @return  <code>true</code> if open, else <code>false</code>.
    */
   public boolean isTransactionOpen()
   {
      return context.get().isTransactionOpen(true);
   }
   
   /**
    * Format the given message text by prepending a descriptor including the
    * user's current context and the physical name of the database.
    *
    * @param   message
    *          Core message
    *
    * @return  Descriptor.
    */
   public String message(String message)
   {
      StringBuilder buf = new StringBuilder("[");
      buf.append(Utils.describeContext());
      buf.append("-->");
      buf.append(database);
      buf.append("] ");
      buf.append(message);
      
      return buf.toString();
   }
   
   /**
    * Initialize the helper objects to which this object delegates certain
    * service requests.
    */
   protected void initializeInstance()
   {
      try
      {
         lockManager = createLockManager();
         identityManager = createIdentityManager();
         
         // If using embedded H2 dialect, prepare the database.
         if (dialect instanceof P2JH2Dialect)
         {
            H2Helper.prepareDatabase(database);
         }
      }
      catch (StopConditionException sce)
      {
         throw sce;
      }
      catch (Exception exc)
      {
         throw new RuntimeException(
            "Error initializing persistence services", exc);
      }
   }
   
   /**
    * Create and initialize the lock manager to be used with this instance of
    * persistence services for the given database.  The lock manager
    * implementation to use is read from the directory.  The default if no
    * entry is found is the {@link InMemoryLockManager}.
    *
    * @return  Lock manager instance.
    *
    * @throws  IllegalAccessException
    *          if the <code>LockManager</code> concrete implementation's
    *          default constructor cannot be accessed.
    * @throws  java.lang.InstantiationException
    *          if the <code>LockManager</code> concrete implementation cannot
    *          be instantiated.
    * @throws  ClassNotFoundException
    *          if the <code>LockManager</code> concrete implementation class
    *          cannot be located.
    */
   protected LockManager createLockManager()
   throws IllegalAccessException,
          InstantiationException,
          ClassNotFoundException
   {
      if (DatabaseManager.TEMP_TABLE_DB.equals(database))
      {
         return null;
      }
      
      // Retrieve fully qualified name of LockManager implementation class
      // and instantiate it.
      String path = String.format("database/%s/p2j/lock_manager/class",
                                  database.getName());
      String className = Utils.getDirectoryNodeString(null,
                                                      path,
                                                      DEFAULT_LOCK_MANAGER,
                                                      false);
      
      LockManager mgr = (LockManager) Class.forName(className).newInstance();
      mgr.setDatabase(database);
      
      if (LOG.isInfoEnabled())
      {
         String msg = String.format("lock manager initialized [%s]",
                                    className);
         LOG.info(message(msg));
      }
      
      return mgr;
   }
   
   /**
    * Create and initialize the identity manager to be used with this
    * instance of persistence services for the given database.  The identity
    * manager implementation to use is read from the directory. The default
    * if no entry is found is {@link #DEFAULT_IDENTITY_MANAGER}.
    *
    * @return  Identity manager instance or <code>null</code> if given
    *          database is temporary.
    *
    * @throws  IllegalAccessException
    *          if the <code>IdentityManager</code> concrete implementation's
    *          default constructor cannot be accessed.
    * @throws  java.lang.InstantiationException
    *          if the <code>IdentityManager</code> concrete implementation
    *          cannot be instantiated.
    * @throws  ClassNotFoundException
    *          if the <code>IdentityManager</code> concrete implementation
    *          class cannot be located.
    */
   protected IdentityManager createIdentityManager()
   throws ClassNotFoundException,
          IllegalAccessException,
          InstantiationException
   {
      if (DatabaseManager.TEMP_TABLE_DB.equals(database))
      {
         return null;
      }

      // Retrieve fully qualified name of IdentityManager implementation
      // class and instantiate it.
      String path = String.format("database/%s/p2j/identity_manager/class",
                                  database.getName());
      String className =
         Utils.getDirectoryNodeString(null,
                                      path,
                                      DEFAULT_IDENTITY_MANAGER,
                                      false);

      IdentityPoolManager mgr =
         (IdentityPoolManager) Class.forName(className).newInstance();

      mgr.setPersistence(this);

      IdentityPool identityPool = new IdentityPool(database);
      mgr.setIdentityPool(identityPool);

      if (LOG.isInfoEnabled())
      {
         String msg = String.format("identity manager initialized [%s]",
                                    className);
         LOG.info(message(msg));
      }

      return mgr;
   }
   
   /**
    * Get a JDBC connection from the current, context local session, creating
    * the session first if necessary.  The session associated with the
    * connection is remembered so that it can be cleaned up properly when the
    * connection is later checked back in.
    * <p>
    * Note:  the caller must not call <code>close()</code> on the returned
    * connection, since the connection may be "borrowed" from the current,
    * active session;  rather, {@link #checkInConnection(Connection)} must be
    * used.
    * 
    * @return  JDBC connection.
    * 
    * @throws  PersistenceException
    *          if a session cannot be created.
    * 
    * @see     #checkInConnection(Connection)
    */
   Connection checkOutConnection()
   throws PersistenceException
   {
      Context local = context.get();
      Session session = local.getSession();
      Connection connection = session.connection();
      borrowedSessions.put(connection, session);
      
      return connection;
   }
   
   /**
    * Check in a JDBC connection previously obtained with {@link
    * #checkOutConnection()}.  This will close the connection if it is not
    * otherwise in use in the current, context local session.
    * 
    * @param   connection
    *          JDBC connection which is no longer needed by the caller.
    *          
    * @throws  PersistenceException
    *          if there is an error closing the session associated with the
    *          connection.
    */
   void checkInConnection(Connection connection)
   throws PersistenceException
   {
      Session session = borrowedSessions.get(connection);
      if (session != null)
      {
         Session localSession = context.get().getSessionNoCreate();
         if (session != localSession && session.isOpen())
         {
            try
            {
               session.close();
            }
            catch (HibernateException exc)
            {
               throw new PersistenceException(exc);
            }
         }
      }
   }
   
   /**
    * Increment the implicit transaction use counter and begin an implicit
    * transaction if none is currently active, and the use count currently
    * is at zero.
    * 
    * @return  The implicit transaction use count after incrementing it.
    * 
    * @throws  PersistenceException
    *          if there is an error beginning a new transaction.
    */
   int pushImplicitTransaction()
   throws PersistenceException
   {
      if (true) return 0;
      
      Context local = context.get();
      
      return local.pushImplicitTransaction();
   }
   
   /**
    * Decrement the implicit transaction use counter.  If it reaches zero
    * as a result, remember the current system time, so that the implicit
    * transaction commit timer can commit the current, implicit transaction
    * when necessary.
    * 
    * @return  The implicit transaction use count after decrementing it.
    */
   int popImplicitTransaction()
   {
      if (true) return 0;
      
      Context local = context.get();
      
      return local.popImplicitTransaction();
   }
   
   /**
    * Bind this instance to a backing, Hibernate session, in case it is not
    * currently backed by one.
    * <p>
    * This is the only method in this class which exposes the Hibernate
    * session to the outside world.  Because of this, this method is
    * intentionally package private.  Use with care.
    *
    * @throws  PersistenceException
    *          if an error occurs creating a session.
    * 
    * @return  Hibernate session.
    */
   Session bind()
   throws PersistenceException
   {
      Context local = context.get();
      Session session = local.getSession();
      
      return session;
   }
   
   /**
    * Evict a collection of DMOs from the current Hibernate session.
    * 
    * @param   evictees
    *          Collection of DMOs to be evicted.
    */
   void evict(Collection<Persistable> evictees)
   {
      Context local = context.get();
      Session session = local.getSessionNoCreate();
      if (session != null)
      {
         for (Persistable dmo : evictees)
         {
            session.evict(dmo);
         }
      }
   }
   
   /**
    * Flush the current session (if any), if there are any pending changes.
    * If there is no active Hibernate session, or no pending changes, no
    * action is taken.
    * 
    * @throws  PersistenceException
    *          if there is any error flushing the session.
    */
   void maybeFlush()
   throws PersistenceException
   {
      Context local = context.get();
      Session session = local.getSessionNoCreate();
      if (session != null && local.isFlushRequired(null))
      {
         flush(local, session, false);
      }
   }
   
   /**
    * Enable/disable non-forced session flushes in the current client
    * context.  Forced flushes via {@link #flush} and {@link #flushBatch} are
    * not affected by this setting.
    *
    * @param   enable
    *          <code>true</code> to enable flushes;  <code>false</code> to
    *          disable.
    */
   void enableFlush(boolean enable)
   {
      context.get().enableFlush(enable);
   }
   
   /**
    * Get the record lock context associated with the current client session.
    *
    * @return  Record lock context for the current session.
    */
   RecordLockContext getRecordLockContext()
   {
      return context.get().getRecordLockContext();
   }
   
   /**
    * Register a <code>SessionListener</code> to receive session life
    * cycle event notifications.  The listener will be deregistered
    * according to the following rules at the earlier of:
    * <ul>
    *   <li>the point at which it is collected by the garbage collector
    *       (once business logic no longer strongly references it);  or
    *   <li>the end of the scope indicated by the <code>scope</code>
    *       parameter:
    *       <ul>
    *         <li><code>ENCLOSING_EXTERNAL_SCOPE</code>
    *         <li><code>CURRENT_SCOPE</code>
    *         <li><code>NEXT_EXTERNAL_SCOPE</code>
    *       </ul>
    * </ul>
    *
    * @param   listener
    *          Session listener to be registered.
    * @param   scope
    *          <code>SessionListener.Scope</code> enum indicating the scope to
    *          which this listener should be registered.  When the scope ends,
    *          the listener is deregistered as described above.
    */
   void registerSessionListener(final SessionListener listener,
                                SessionListener.Scope scope)
   {
      Context local = context.get();
      
      local.registerSessionListener(listener, scope);
   }
   
   /**
    * Execute an HQL query and load a single record from the database or from
    * the session-level cache, in read-only mode.  The record, if present,
    * must be unique.  No locking is attempted.
    * <p>
    * This method is provided to support housekeeping queries internal to the
    * persistence framework.  It is expected that <code>hql</code> is the
    * minimal query necessary to return the primary key of the DMO which
    * matches the query criteria. The Hibernate session is then used to
    * create and return an uninitialized proxy instance of the DMO.  This
    * implementation is used to avoid the overhead of:
    * <ul>
    *   <li>retrieving all data for the matching record from the database;
    *   <li>hydrating the DMO with that data.
    * </ul>
    *
    * @param   dmoClass
    *          Expected return type;  the DMO implementation class being
    *          queried.
    * @param   hql
    *          HQL query statement passed to Hibernate.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    *
    * @return  A single DMO instance representing the query result, or
    *          <code>null</code> if no result was found.
    *
    * @throws  PersistenceException
    *          if there was an error executing the query or if the query
    *          returned multiple results.
    */
   /*
   Persistable loadReadOnly(Class<?> dmoClass,
                            String hql,
                            Object[] values,
                            Type[] types)
   throws PersistenceException
   {
      Context local = context.get();
      Persistable result = null;
      
      try
      {
         Session session = local.getSession();
         Query query = getQuery(local, hql, 2, 0, values, types);
         query.setReadOnly(true);
         if (!local.isFlushRequired(dmoClass))
         {
            query.setFlushMode(FlushMode.NEVER);
         }
         
         Serializable id = (Serializable) query.uniqueResult();
         if (id == null)
         {
            return null;
         }
         
         result = (Persistable) session.load(dmoClass, id);
      }
      catch (HibernateException exc)
      {
         DBUtils.handleInterrupt(exc);
         
         local.closeSession(true);
         
         String msg = message("error loading record");
         
         if (LOG.isWarnEnabled())
         {
            LOG.warn(msg + System.getProperty("line.separator") + hql);
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Details:", exc);
            }
         }
         
         throw new PersistenceException(msg, exc);
      }
      
      return result;
   }
   */
   
   /**
    * Process an error condition in which a unique result should have been
    * found, but multiple results occurred.
    *
    * @param   buffer
    *          Record buffer associated with the failed query.
    * @param   exc
    *          Optional root cause exception.
    *
    * @throws  PersistenceException
    *          always;  indicates the nature of the error.
    */
   void uniqueResultViolation(RecordBuffer buffer, Exception exc)
   throws PersistenceException
   {
      int errnum = 3166;
      buffer.setAmbiguous(true);
      
      String msg = "More than one " + buffer.getDMOAlias() +
                   " records found by a unique FIND";
      
      if (LOG.isWarnEnabled())
      {
         LOG.warn(message(msg));
         if (LOG.isDebugEnabled())
         {
            LOG.debug("Details:", exc);
         }
      }
      
      if (exc != null)
      {
         throw new PersistenceException(msg, errnum, exc);
      }
      else
      {
         throw new PersistenceException(msg, errnum);
      }
   }
   
   /**
    * Flush the current session, which synchronizes the in-memory state of
    * DMOs to the database.  Optionally, the session is cleared as well.
    * This request bypasses the current client context's setting as to
    * whether flushing is enabled.
    * <p>
    * It is expected that this method will only be called within an active
    * transaction.  If no session is active, this method will raise an
    * exception.
    *
    * @param   clear
    *          <code>true</code> to clear the session (used during bulk
    *          inserts or updates to prevent a progressive memory leak) after
    *          the flush.
    *
    * @throws  PersistenceException
    *          if no session currently is active or if an error occurs during
    *          session flush.
    *
    * @see     #enableFlush
    */
   private void flush(boolean clear)
   throws PersistenceException
   {
      Context local = context.get();
      Session session = local.getSessionNoCreate();
      
      // Sanity.
      if (session == null)
      {
         throw new PersistenceException("No active session to flush");
      }
      
      flush(local, session, clear);
   }
   
   /**
    * Flush the current session, which synchronizes the in-memory state of
    * DMOs to the database.  Optionally, the session is cleared as well.
    * This request bypasses the current client context's setting as to
    * whether flushing is enabled.
    * 
    * @param   local
    *          Context-local state.
    * @param   session
    *          Hibernate session
    * @param   clear
    *          <code>true</code> to clear the session (used during bulk
    *          inserts or updates to prevent a progressive memory leak) after
    *          the flush.
    *
    * @throws  PersistenceException
    *          if an error occurs during session flush.
    */
   private void flush(Context local, Session session, boolean clear)
   throws PersistenceException
   {
      boolean commit = false;
      long t1 = System.currentTimeMillis();
      
      try
      {
         commit = beginTransaction(local);
         
         session.flush();
         
         if (clear)
         {
            session.clear();
         }
      }
      catch (HibernateException exc)
      {
         commit = false;
         
         DBUtils.handleException(database, exc);
         
         local.closeSession(true);
         
         String msg = "error flushing " +
                      (clear ? "and clearing " : "") +
                      "session";
                      
         throw new PersistenceException(message(msg), exc);
      }
      finally
      {
         if (commit)
         {
            local.commit(true);
         }
         
         long t2 = System.currentTimeMillis();
         
         flushTxTime += (t2 - t1);
         ++flushTxCount;
      }
   }
   
   /**
    * Perform special handling on query substitution parameters to avoid
    * errors during query execution.  This entails:
    * <ul>
    *   <li>Trimming all trailing spaces, tabs, line feeds, and carriage
    *       returns from any parameter of type <code>character</code>.  Any
    *       parameter affected is replaced with a new <code>character</code>
    *       instance at the same index position within <code>args</code>.
    *   <li>Clipping out of bounds date values to the range supported by the
    *       backing database's implementation of a date data type.  Any
    *       parameter affected is replaced with a new <code>date</code>
    *       instance at the same index position within <code>args</code>.
    * </ul>
    * <p>
    * Note:  this method has special processing for <code>null</code> bytes
    * embedded in <code>character</code> variables.  The text of the variable
    * is truncated starting at the first embedded <code>null</code> byte.
    * This means that the resulting substitution parameters are guaranteed
    * to never include <code>null</code> bytes.
    *
    * @param   args
    *          Array of query substitution parameters or <code>null</code> if
    *          none.
    */
   private void preprocessQueryParameters(Object[] args)
   {
      if (args != null)
      {
         int len = args.length;
         for (int i = 0; i < len; i++)
         {
            args[i] = preprocessQueryParameter(args[i], (P2JDialect) dialect);
         }
      }
   }
   
   /**
    * Begin a new Hibernate transaction and store the instance in the local
    * context.  A new database session is created if one does not already
    * exist in the current context.
    *
    * @param   local
    *          Context-local state.
    *
    * @return  <code>true</code> if a new transaction was begun;
    *          <code>false</code> if there was already an active transaction.
    *
    * @throws  PersistenceException
    *          if there is an error beginning the new transaction.
    */
   private boolean beginTransaction(Context local)
   throws PersistenceException
   {      
      if (local.beginTransaction(true) == null)
      {
         return false;
      }
      
      if (LOG.isDebugEnabled())
      {
         LOG.debug(message("beginning transaction"));
      }
      
      return true;
   }
   
   /**
    * Get or create a query for the given HQL query string.  A cached query
    * is used if it exists, and if there is no limit on the number of rows to
    * return, and there is no non-default starting row set.  A new query is
    * created if no such query is cached, or if max results or starting row
    * need to be set.  The latter query is not cached, since once these values
    * are set, they cannot be reset to their defaults, but can only be set to
    * a new, specific range.
    *
    * @param   local
    *          Context-local state.
    * @param   hql
    *          HQL query string.
    * @param   maxResults
    *          The maximum number of results to be returned by the query.
    *          If this value is non-positive, no upper limit is applied.
    * @param   startOffset
    *          The 0-based offset of the first record to retrieve.  If this
    *          value is non-positive, an offset of 0 is used by default.
    * @param   values
    *          Substitution values for the query.  If none, this should be an
    *          empty array.
    * @param   types
    *          Hibernate types for the <code>values</code> array.  If none,
    *          this should be an empty array.
    * @param   noCache
    *          If <code>true</code>, do not cache the query;  otherwise, use
    *          normal rules to determine whether query should be cached or
    *          not.
    *
    * @return  Hibernate query object.
    *
    * @throws  PersistenceException
    *          if there was an error creating a Hibernate session.
    */
   private Query getQuery(Context local,
                          String hql,
                          int maxResults,
                          int startOffset,
                          Object[] values,
                          Type[] types,
                          boolean noCache)
   throws PersistenceException
   {
      Query query = local.getQuery(hql, maxResults, startOffset, !noCache);
      
      // Bind parameters if any.  It is OK to do this even with cached
      // queries, because each use is guaranteed to reset the parameters.
      if (values != null && values.length > 0)
      {
         preprocessQueryParameters(values);
         query.setParameters(values, types);
      }
      
      // Allow the second level cache to store the query's results.  This is
      // a separate issue than whether the Query object itself is cached
      // (i.e., the noCache flag).
      query.setCacheable(true);
      
      return query;
   }
   
   /**
    * A storage and work area for context-local variables used by the
    * enclosing class.  One instance of this class is created for each
    * client context.  It is used to:
    * <ul>
    *   <li>create and cache Hibernate query objects;
    *   <li>create, store and close the active Hibernate session;
    *   <li>manage the active Hibernate transaction.
    * </ul>
    * <p>
    * Special care is taken to not close a Hibernate session while other
    * entities are dependent upon it for services.
    */
   private class Context
   {
      /** Context-local record lock context */
      private final RecordLockContext lockContext =
         RecordLockContext.get(Persistence.this);

      /** Context-local ChangeBroker */
      private final ChangeBroker changeBroker = ChangeBroker.get();
      
      /** Session listener objects */
      private final WeakHashMap<SessionListener, Object> sessionListeners =
         new WeakHashMap<SessionListener, Object>();
      
      /** Soft cache of HQL strings to Hibernate queries */
      private SoftReference<Map<QueryKey, Query>> queryCache = null;
      
      /** Active Hibernate session, if any */
      private Session session = null;
      
      /** Active Hibernate transaction, if any */
      private Transaction transaction = null;
      
      /** Implicit transaction use count */
      private int implicitTxUseCount = 0;
      
      /** Is session flush enabled? */
      private boolean enableFlush = true;
      
      /** Flag to prevent nested session close attempts */
      private boolean closingSession = false;
      
      /** Optional, dialect-specific, context-local data */
      private Object dialectData =
         ((P2JDialect) dialect).createContext(Persistence.this);
      
      /** Is current transaction explicit? */
      private boolean explicitTx = false;
      
      private boolean forceCommitOnImplicitPop = false;
      
      /**
       * Get the record lock context associated with the current client
       * session.
       *
       * @return  Record lock context for the current session.
       */
      RecordLockContext getRecordLockContext()
      {
         return lockContext;
      }
      
      /**
       * Get a cached query for the given HQL query string, or create a new
       * query if no such query is cached, or if caching is not permitted by
       * the <code>cache</code> parameter.
       *
       * @param   hql
       *          HQL query string.
       * @param   maxResults
       *          The maximum number of results to be returned.
       * @param   startOffset
       *          Zero-based offset of first result to be returned (undefined
       *          results unless <code>hql</code> includes an <code>order
       *          by</code> phrase).
       * @param   cache
       *          <code>true</code> to look for a matching query in the cache,
       *          and to store this query in the cache;  <code>false</code> if
       *          query should be neither retrieved from, nor stored in the
       *          cache.
       *
       * @return  Hibernate query object.
       *
       * @throws  PersistenceException
       *          if there was an error creating a Hibernate session.
       */
      Query getQuery(String hql,
                     int maxResults,
                     int startOffset,
                     boolean cache)
      throws PersistenceException
      {
         long t1 = System.currentTimeMillis();
         Query query = null;
         QueryKey key = null;
         
         try
         {
            Map<QueryKey, Query> cacheMap = null;
            
            if (cache)
            {
               cacheMap = getQueryCacheMap();
               key = new QueryKey(hql, maxResults, startOffset);
               query = cacheMap.get(key);
            }
            
            if (query == null)
            {
               Session session = getSession();
               query = session.createQuery(hql);
               
               if (maxResults > 0)
               {
                  query.setMaxResults(maxResults);
               }
               if (startOffset > 0)
               {
                  query.setFirstResult(startOffset);
               }
               
               if (cache)
               {
                  cacheMap.put(key, query);
               }
            }
            else
            {
               // Query came from cache;  reset defaults.
               query.setFlushMode(FlushMode.AUTO);
               query.setReadOnly(false);
            }
         }
         finally
         {
            long t2 = System.currentTimeMillis();
            
            getQueryTime += (t2 - t1);
            ++getQueryCount;
         }
         
         return query;
      }
      
      /**
       * Commit the current Hibernate transaction and possibly close the
       * current session.  If the transaction cannot be committed, roll it
       * back and raise an error.
       *
       * @param   explicit
       *          <code>true</code> if the current transaction is expected to
       *          be explicit, <code>false</code> if implicit.
       *
       * @throws  IllegalStateException
       *          if no transaction is active in the current session;  this
       *          represents a programming error.
       * @throws  PersistenceException
       *          if there is an error committing the transaction.
       */
      void commit(boolean explicit)
      throws PersistenceException
      {
         if (transaction == null)
         {
            throw new IllegalStateException(
               message("no current transaction available to commit"));
         }
         long t1 = System.currentTimeMillis();
         
         try
         {
            if (explicitTx && !isFlushRequired(null))
            {
               // If there are no changes since the last flush, prevent the
               // commit from having to flush.  This setting need not be set
               // back after the commit;  the session is about to be closed
               // anyway.
               session.setFlushMode(FlushMode.NEVER);
            }
            
            try
            {
               // Increment use count so SessionListeners do not commit
               // implicit transaction.
               implicitTxUseCount++;
               
               if (implicitTxUseCount > 1)
               {
                  notifySessionEvent(
                     SessionListener.Event.TRANSACTION_COMMITTING);
               }
            }
            finally
            {
               // Decrement use count back to balance increment above.
               implicitTxUseCount--;
               
               if (DatabaseManager.TEMP_TABLE_DB.equals(Persistence.this.database))
                  ++commitTxCountT;
               else
                  ++commitTxCountP;
               transaction.commit();
            }
            
            if (LOG.isDebugEnabled())
            {
               LOG.debug(message((explicit ? "explicit" : "implicit") +
                                 " transaction committed"));
            }
            
            endTransaction(explicit, false, false);
         }
         catch (HibernateException exc)
         {
            DBUtils.handleException(database, exc);
            
            try
            {
               rollback(explicit, true);
            }
            catch (PersistenceException exc1)
            {
               // Don't need to throw an exception here;  this would hide the
               // commit exception, so just log it.
               if (LOG.isErrorEnabled())
               {
                  LOG.error(message("rollback error"), exc1);
               }
            }
            
            throw new PersistenceException(
               message("unable to commit " +
                       (explicit ? "explicit" : "implicit") +
                       " transaction"),
               exc);
         }
         finally
         {
            long t2 = System.currentTimeMillis();
            commitTxTime += (t2 -t1);
         }
      }
      
      /**
       * Roll back the current Hibernate transaction and possibly close the
       * current session.  If the transaction cannot be rolled back, raise an
       * error.
       *
       * @param   explicit
       *          <code>true</code> if the current transaction is expected to
       *          be explicit, <code>false</code> if implicit.
       * @param   error
       *          Indicates whether the rollback is due to an error.
       *
       * @throws  IllegalStateException
       *          if no transaction is active in the current session;  this
       *          represents a programming error.
       * @throws  PersistenceException
       *          if there is an error rolling back the transaction.
       */
      void rollback(boolean explicit, boolean error)
      throws PersistenceException
      {
         long t1 = System.currentTimeMillis();
         
         if (transaction == null)
         {
            throw new IllegalStateException(
               message("no current transaction available to rollback"));
         }
         
         try
         {
            maybeFlushSession();
            
            transaction.rollback();
            
            if (LOG.isDebugEnabled())
            {
               LOG.debug(message("transaction rolled back"));
            }
         }
         catch (HibernateException exc)
         {
            DBUtils.handleException(database, exc);
            
            closeSession(true);
            
            throw new PersistenceException(
               message("unable to rollback current transaction"),
               exc);
         }
         finally
         {
            endTransaction(explicit, error, true);
            
            long t2 = System.currentTimeMillis();
            rollbackTime += (t2 - t1);
            ++rollbackCount;
         }
      }
      
      /**
       * Begin a new Hibernate transaction and store the instance for future
       * use.  If an explicit transaction already exists, it is returned, and
       * a new one is not created.  If an implicit transaction already exists,
       * and a new, explicit transaction is being requested, the implicit
       * transaction is first committed, then the new, explicit transaction is
       * begun.  A new database session is created if one does not already
       * exist in the current context.
       *
       * @param   explicit
       *          <code>true</code> if an explicit transaction is being
       *          requested;  <code>false</code> if the request is for an
       *          implicit transaction.
       *
       * @return  The active transaction, whether an existing explicit
       *          transaction, or a new implicit or explicit transaction.
       *
       * @throws  PersistenceException
       *          if there is an error beginning the new transaction or
       *          committing an existing one.
       */
      Transaction beginTransaction(boolean explicit)
      throws PersistenceException
      {
         long t1 = System.currentTimeMillis();
         try
         {
            if (transaction != null)
            {
               if (explicitTx)
               {
                  // Already an explicit transaction active.
                  return null;
               }
               
               // An implicit transaction is active and must be committed before
               // we can begin a new explicit transaction (we know the new
               // transaction is explicit because we would not be inside this
               // block otherwise).
               if (transaction.isActive())
               {
                  commit(false);
               }
            }
            
            try
            {
               Session session = getSession();
               
               // Implicit transactions are read-only, explicit ones most likely
               // not.
               session.setFlushMode(explicit ? FlushMode.AUTO : FlushMode.NEVER);
               transaction = session.beginTransaction();
               
               if (DatabaseManager.TEMP_TABLE_DB.equals(Persistence.this.database))
                  ++openTxCountT;
               else
                  ++openTxCountP;
               
               return transaction;
            }
            catch (HibernateException exc)
            {
               DBUtils.handleException(database, exc);
               
               closeSession(true);
               
               // TODO:  use proper error message/number.
               throw new PersistenceException(
                  message("error beginning database transaction"),
                  exc);
            }
            finally
            {
               if (explicit && transaction != null)
               {
                  explicitTx = true;
               }
            }
         }
         finally
         {
            long t2 = System.currentTimeMillis();
            openTxTime += (t2 -t1);
         }
      }
      
      /**
       * Discard the current transaction, which was previously committed or
       * rolled back.
       * <p>
       * For an implicit transaction, the currently scheduled, implicit commit
       * task is cancelled and nulled out.  In case of an error, the current
       * session is marked broken, so it will be closed and replaced upon the
       * next attempt to use it on the user sesion thread (the current thread
       * must be the implicit commit timer if we get here for an implicit
       * transaction).
       * <p>
       * For an explicit transaction, the current session is closed and the
       * active dialect is allowed to do some post-transaction processing.
       * Also, the implicit transaction use count is reset.
       * 
       * @param   explicit
       *          Indicates whether the transaction being ended is explicit
       *          or implicit.
       * @param   error
       *          Indicates whether the transaction ended with an error.
       * @param   rollback
       *          Indicates whether the transaction is being ended after a
       *          rollback (as opposed to a commit).
       * 
       * @throws  PersistenceException
       *          if an error occurs closing the session or performing
       *          post-transaction processing.
       */
      void endTransaction(boolean explicit, boolean error, boolean rollback)
      throws PersistenceException
      {
         try
         {
            try
            {
               if (explicit)
               {
                  closeSession(error);
               }
            }
            finally
            {
               transaction = null;
            }
            
            // Give the dialect a chance to perform post-transaction
            // processing.
            P2JDialect p2jDialect = (P2JDialect) dialect;
            p2jDialect.postTransaction(Persistence.this,
                                       dialectData,
                                       rollback);
            
            // If an implicit transaction previously was interrupted to begin
            // an explicit transaction, begin a new implicit transaction now.
            if (explicitTx && implicitTxUseCount > 0)
            {
               beginTransaction(false);
            }
         }
         finally
         {
            explicitTx = false;
         }
      }
      
      private int pushPopCount = 0;
      
      /**
       * Increment the implicit transaction use counter and begin an implicit
       * transaction if none is currently active, and the use count currently
       * is at zero.
       * 
       * @return  The implicit transaction use count after incrementing it.
       * 
       * @throws  PersistenceException
       *          if there is an error beginning a new transaction.
       */
      int pushImplicitTransaction()
      throws PersistenceException
      {
         ++pushTxCount;
         
         if (implicitTxUseCount >= 0)
         {
            if (implicitTxUseCount == 0 && transaction == null)
            {
               beginTransaction(false);
            }
            
            implicitTxUseCount++;
         }
         
         return implicitTxUseCount;
      }
      
      /**
       * Decrement the implicit transaction use counter.  If it reaches zero
       * as a result, remember the current system time, so that the implicit
       * transaction commit timer can commit the current, implicit transaction
       * when necessary.
       * 
       * @return  The implicit transaction use count after decrementing it.
       */
      int popImplicitTransaction()
      {
         ++popTxCount;
         
         if (implicitTxUseCount > 0)
         {
            implicitTxUseCount--;
         }
         
         if (!explicitTx && implicitTxUseCount == 0 && transaction != null)
         {
            try
            {
               if (LOG.isDebugEnabled())
               {
                  LOG.debug("Performing implicit transaction commit");
               }
               
               commit(false);
               
               forceCommitOnImplicitPop = false;
            }
            catch (PersistenceException exc)
            {
               if (LOG.isErrorEnabled())
               {
                  LOG.error("Error commiting implicit transaction", exc);
               }
            }
         }
         
         return implicitTxUseCount;
      }
      
      /**
       * Determine whether a session flush is required for the given DMO
       * class.  A flush is required if flushing currently is enabled and the
       * {@link ChangeBroker} has tracked a change to a DMO of type
       * <code>dmoClass</code> since the last flush.
       *
       * @param   entityName
       *          A DMO class name, or <code>null</code> to check whether a
       *          flush is required for <i>any</i> DMO class.
       *
       * @return  <code>true</code> if a flush is required, else
       *          <code>false</code>.
       */
      boolean isFlushRequired(String entityName)
      {
         return (enableFlush &&
                 changeBroker.requiresFlush(Persistence.this, entityName));
      }
      
      /**
       * Enable/disable non-forced session flushes for this client context.
       * Non-forced flushes are those which Hibernate would initiate in
       * auto-flush mode.  Forced flushes are not affected by this setting.
       *
       * @param   enable
       *          <code>true</code> to enable flushes;  <code>false</code> to
       *          disable.
       */
      void enableFlush(boolean enable)
      {
         enableFlush = enable;
      }
      
      /**
       * Register a <code>SessionListener</code> to receive session life
       * cycle event notifications.  The listener will be deregistered
       * according to the following rules at the earlier of:
       * <ul>
       *   <li>the point at which it is collected by the garbage collector
       *       (once business logic no longer strongly references it);  or
       *   <li>the end of the scope indicated by the <code>scope</code>
       *       parameter:
       *       <ul>
       *         <li><code>ENCLOSING_EXTERNAL_SCOPE</code>
       *         <li><code>CURRENT_SCOPE</code>
       *         <li><code>NEXT_EXTERNAL_SCOPE</code>
       *       </ul>
       * </ul>
       *
       * @param   listener
       *          Session listener to be registered.
       * @param   scope
       *          <code>SessionListener.Scope</code> enum indicating the scope
       *          to which this listener should be registered.  When the
       *          scope ends, the listener is deregistered as described above.
       */
      void registerSessionListener(final SessionListener listener,
                                   SessionListener.Scope scope)
      {
         // Store listener as a key in the map, so it may be released when no
         // longer referenced elsewhere.
         sessionListeners.put(listener, null);
         
         if (SessionListener.Scope.NONE.equals(scope))
         {
            // Never deregister the listener by scope.
            return;
         }
         
         // Finalizable which will remove listener at a safe time.  The
         // listener may be removed earlier if not referenced in business
         // logic by the time this Finalizable executes its finished method,
         // and the GC collects it.  In the latter case, the listener will not
         // be notified, but it doesn't matter, since it is no longer
         // referenced by business logic at that point.
         final Finalizable cleaner = new Finalizable()
         {
            public void finished()
            {
               sessionListeners.remove(listener);
               listener.deregisteredSessionListener();
            }
            public void iterate() {}
            public void retry() {}
         };
         
         switch (scope)
         {
            case GLOBAL:
               TransactionManager.registerFinalizable(cleaner, true);
               break;
            case ENCLOSING_EXTERNAL:
               TransactionManager.registerTopLevelFinalizable(cleaner, true);
               break;
            case CURRENT:
               TransactionManager.registerFinalizable(cleaner, false);
               break;
            case NEXT_EXTERNAL:
               TransactionManager.registerNextExternal(cleaner);
               break;
         }
      }
      
      /**
       * Retrieve the Hibernate <code>Session</code> object associated with
       * this context, if any.
       * <p>
       * If the current session is broken, return <code>null</code>.
       *
       * @return  The session object associated with the current context, or
       *          <code>null</code> if none has been created.
       */
      Session getSessionNoCreate()
      {
         return session;
      }
      
      /**
       * Retrieve the Hibernate <code>Session</code> object associated with
       * this context, creating it if necessary.  The session can be closed
       * by {@link #closeSession}.
       *
       * @return  The session object associated with the current context.
       *
       * @throws  PersistenceException
       *          if an error occurs associating a detached record with the
       *          new session.
       * @throws  IllegalArgumentException
       *          if <code>database</code> has not been registered with the
       *          persistence service.
       */
      Session getSession()
      throws PersistenceException
      {
         if (session == null)
         {
            SessionFactory factory =
               DatabaseManager.getSessionFactory(database);
            if (factory == null)
            {
               throw new IllegalArgumentException(
                  message(
                     "database not registered with persistence service"));
            }
            
            Interceptor interceptor =
               changeBroker.getInterceptor(Persistence.this);
            session = factory.openSession(interceptor);
            
            ++sessOpenCount;

            // Associate all detached records in the current context with the
            // new session.
            BufferManager bufMgr = BufferManager.get();
            Iterator<RecordBuffer> iter =
               bufMgr.activeBuffers(Persistence.this);
            boolean debug = LOG.isDebugEnabled();
            RecordBuffer buffer = null;
            Persistable record = null;
            try
            {
               while (iter.hasNext())
               {
                  buffer = iter.next();
                  record = buffer.getCurrentRecord();
                  
                  if (debug)
                  {
                     LOG.debug(message("associating record with session:  " +
                                       record.getClass().getName() + "#" +
                                       record.getId()));
                  }
                  
                  session.lock(record, LockMode.NONE);
               }
            }
            catch (HibernateException exc)
            {
               DBUtils.handleException(database, exc);
               
               closeSession(true);
               
               String msg =
                  "Error reassociating detached DMO with new session";
               if (record != null)
               {
                  msg += (" [" +
                          record.getClass().getName() +
                          "#" +
                          record.getId() +
                          "]");
               }
               
               LOG.error(message(msg), exc);
               
               throw exc;
            }
            
            if (debug)
            {
               LOG.debug(message("opened Hibernate session"));
               /*
               if (LOG.isTraceEnabled())
               {
                  LOG.trace("Stack trace:  ", new Throwable());
               }
               */
            }
         }
         
         return session;
      }
      
      /**
       * Indicate whether a database-level transaction currently is open.
       * 
       * @param   explicit
       *          <code>true</code> to limit the check to an explicit
       *          transaction;  <code>false</code> to check for either an
       *          implicit or explicit transaction.
       *
       * @return  <code>true</code> if open, else <code>false</code>.
       */
      boolean isTransactionOpen(boolean explicit)
      {
         return ((!explicit || explicitTx) &&
                 transaction != null &&
                 transaction.isActive());
      }
      
      /**
       * Close the Hibernate <code>Session</code> object associated with this
       * context.
       * <p>
       * This implementation is tolerant of multiple closes on the same
       * session. If the session was already closed or could not be found,
       * the method simply returns.
       *
       * @param   error
       *          <code>true</code> to indicate the session is closing due to
       *          an error;  <code>false</code> to indicate the session is
       *          closing in the course of normal processing.
       * 
       * @throws  PersistenceException
       *          if an open session was found but the attempt to close it
       *          failed.
       */
      void closeSession(boolean error)
      throws PersistenceException
      {
         // Prevent nested close attempts.
         if (closingSession)
         {
            return;
         }
         
         try
         {
            closingSession = true;
            
            if (session != null &&
                session.isOpen() &&
                isTransactionOpen(false))
            {
               if (LOG.isWarnEnabled())
               {
                  LOG.warn(message("rolling back orphaned transaction"));
               }
               rollback(explicitTx, error);
            }
            
            if (!error)
            {
               // If session is not being closed, attempt a flush to ensure
               // consistency in case we just ended a transaction with a
               // rollback.
               maybeFlushSession();
               
               // Special treatment for temp table sessions.  These must
               // remain open as long as any table exists, otherwise the
               // table and its data will be lost.
               if (DatabaseManager.TEMP_TABLE_DB.equals(database) &&
                   TemporaryBuffer.getTableCount() > 0)
               {
                  return;
               }
            }
            
            closeSessionImpl(error);
         }
         finally
         {
            closingSession = false;
         }
      }
      
      /**
       * Get the context-local, dialect-specific data object, if any.
       * 
       * @return  Dialect specific data object, or <code>null</code> if no
       *          such data exists.
       */
      Object getDialectData()
      {
         return dialectData;
      }
      
      /**
       * Retrieve the map underlying the soft query cache.  If the cache has
       * been garbage collected due to low memory availability, an empty map
       * will be returned.
       * 
       * @return  The map underlying the soft query cache, which may be empty,
       *          but will not be <code>null</code>.
       */
      private Map<QueryKey, Query> getQueryCacheMap()
      {
         Map<QueryKey, Query> map = null;
         
         if (queryCache != null)
         {
            map = queryCache.get();
         }
         
         if (map == null)
         {
            map = new HashMap<QueryKey, Query>();
         }
         
         if (queryCache == null)
         {
            queryCache = new SoftReference<Map<QueryKey, Query>>(map);
         }
         
         return map;
      }
      
      /**
       * If an open session exists and the change broker indicates a flush is
       * required, force a session flush now.
       *
       * @throws  PersistenceException
       *          if an error occurs during session flush.
       */
      private void maybeFlushSession()
      throws PersistenceException
      {
         long t1 = System.currentTimeMillis();
         
         if (session == null || !session.isOpen() || !isFlushRequired(null))
         {
            return;
         }
         
         try
         {
            session.flush();
         }
         catch (HibernateException exc)
         {
            DBUtils.handleException(database, exc);
            
            closeSession(true);
            
            throw new PersistenceException(exc);
         }
         finally
         {
            long t2 = System.currentTimeMillis();
            
            maybeFlushTime += (t2 - t1);
            ++maybeFlushCount;
         }
      }
      
      /**
       * Close the Hibernate <code>Session</code> object associated with this
       * context.
       * <p>
       * This implementation is tolerant of multiple closes on the same
       * session. If the session was already closed or could not be found,
       * the method simply returns.
       * <p>
       * All registered session listeners are notified before the session is
       * closed.
       *
       * @param   error
       *          <code>true</code> to indicate the session is closing due to
       *          an error;  <code>false</code> to indicate the session is
       *          closing in the course of normal processing.
       *
       * @throws  PersistenceException
       *          if an open session was found but the attempt to close it
       *          failed, or if any error occurs during a registered session
       *          listener's session closing processing.
       */
      private void closeSessionImpl(boolean error)
      throws PersistenceException
      {
         queryCache = null;
         
         if (session != null)
         {
            try
            {
               if (session.isOpen())
               {
                  SessionListener.Event event =
                     (error
                      ? SessionListener.Event.SESSION_CLOSING_WITH_ERROR
                      : SessionListener.Event.SESSION_CLOSING_NORMALLY);
                  try
                  {
                     notifySessionEvent(event);
                  }
                  finally
                  {
                     ++sessCloseCount;
                     session.close();
                  }
                  
                  if (LOG.isDebugEnabled())
                  {
                     LOG.debug(message("closed Hibernate session"));
                     /*
                     if (LOG.isTraceEnabled())
                     {
                        LOG.trace("Stack trace:  ", new Throwable());
                     }
                     */
                 }
               }
            }
            catch (HibernateException exc)
            {
               DBUtils.handleException(database, exc);
               
               throw new PersistenceException(
                  message("error closing Hibernate session"),
                  exc);
            }
            finally
            {
               session = null;
            }
         }
      }
      
      /**
       * Notify all registered session listeners that the current Hibernate
       * session is about to close.  Listeners which were set to be
       * deregistered upon this event are deregistered after they are
       * notified.  There is no guaranteed order in which listeners are
       * notified.
       * <p>
       * All registered listeners are given a chance to process this event,
       * even if one or more raise a checked exception (there is no such
       * guarantee for unchecked exceptions).  The first exception caught
       * is wrapped in a <code>PersistenceException</code> and rethrown after
       * all listeners have been notified;  subsequent exceptions are logged
       * but otherwise ignored.
       *
       * @param   event
       *          Session event type.
       */
      private void notifySessionEvent(SessionListener.Event event)
      {
         Iterator<SessionListener> iter =
            sessionListeners.keySet().iterator();
         while (iter.hasNext())
         {
            SessionListener listener = iter.next();
            boolean deregister = false;
            
            try
            {
               // Allow listener to process session event.
               deregister = listener.sessionEvent(event);
            }
            catch (PersistenceException exc)
            {
               // Deregister listener if there was an error.
               deregister = true;
               
               LOG.error("Error processing session event", exc);
            }
            finally
            {
               // Remove listeners if there was an error, or if the listener
               // indicated it should be removed in response to the session
               // closing notification.
               if (deregister)
               {
                  iter.remove();
                  listener.deregisteredSessionListener();
               }
            }
         }
      }
      
      /**
       * Cleanup hook invoked at the end of the current client context.  We
       * close the currently open Hibernate session, if any.  This is the
       * only context local object which pins memory outside of this class
       * (within the Hibernate framework), so it requires explicit closure.
       * All other variables are garbage collected naturally when the context
       * ends.
       */
      void cleanup()
      {
         try
         {
            if (session != null &&
                session.isOpen() &&
                isTransactionOpen(false))
            {
               // Under normal circumstances, there should not be any tx open
               // at the context's end of life, but just in case...
               if (LOG.isWarnEnabled())
               {
                  LOG.warn(
                     message(
                        "rolling back open transaction during context cleanup"));
               }
               rollback(explicitTx, false);
            }
            else
            {
               // We may very well have an open session at the context's end
               // of life, since these are only closed upon tx commits or
               // rollbacks, or in response to a database error.
               if (LOG.isDebugEnabled() &&
                   session != null &&
                   session.isOpen())
               {
                  LOG.debug(
                     message(
                        "closing Hibernate session during context cleanup"));
               }
               closeSession(false);
            }
         }
         catch (PersistenceException exc)
         {
            LOG.error(
               "Error closing Hibernate session during context cleanup",
               exc);
         }
      }
   }
   
   /**
    * A hash key for a unique combination of table and column name.
    */
   private static class ColumnKey
   {
      /** Table name */
      private final String table;
      
      /** Column name */
      private final String column;
      
      /**
       * Constructor.
       * 
       * @param   table
       *          Table name.
       * @param   column
       *          Column name.
       */
      ColumnKey(String table, String column)
      {
         this.table = table;
         this.column = column;
      }
      
      /**
       * Test for equivalence with another object.
       * 
       * @param   o
       *          Another <code>ColumnKey</code> instance.
       * 
       * @return  <code>true</code> if <code>o</code> has same table and
       *          column name as this object;  else <code>false</code>.
       */
      public boolean equals(Object o)
      {
         ColumnKey that = (ColumnKey) o;
         
         return (this.table.equals(that.table) &&
                 this.column.equals(that.column));
      }
      
      /**
       * A hash code implementation consisent with our overridden {@link
       * #equals(Object)} method.
       * 
       * @return  Hash code.
       */
      public int hashCode()
      {
         int result = 17;
         
         result = 37 * result + table.hashCode();
         result = 37 * result + column.hashCode();
         
         return result;
      }
      
      /**
       * Get string representation of this object.
       * 
       * @return  String representation.
       */
      public String toString()
      {
         StringBuilder buf = new StringBuilder();
         buf.append(table);
         buf.append(":");
         buf.append(column);
         
         return buf.toString();
      }
   }
   
   /**
    * A hash key for a unique combination of HQL and maximum results and
    * starting row offset.
    */
   private static class QueryKey
   {
      /** HQL statement */
      private final String hql;
      
      /** Maximum results */
      private final int maxResults;
      
      /** Starting row offset */
      private final int startOffset;
      
      /**
       * Constructor.
       * 
       * @param   hql
       *          HQL statement
       * @param   maxResults
       *          Maximum results.
       * @param   startOffset
       *          Starting row offset.
       */
      QueryKey(String hql, int maxResults, int startOffset)
      {
         this.hql = hql;
         this.maxResults = maxResults;
         this.startOffset = startOffset;
      }
      
      /**
       * Test for equivalence with another object.
       * 
       * @param   o
       *          Another <code>ColumnKey</code> instance.
       * 
       * @return  <code>true</code> if <code>o</code> has same internal
       *          state as this object;  else <code>false</code>.
       */
      public boolean equals(Object o)
      {
         QueryKey that = (QueryKey) o;
         
         return (this.maxResults == that.maxResults   &&
                 this.startOffset == that.startOffset &&
                 this.hql.equals(that.hql));
      }
      
      /**
       * A hash code implementation consisent with our overridden {@link
       * #equals(Object)} method.
       * 
       * @return  Hash code.
       */
      public int hashCode()
      {
         int result = 17;
         
         result = 37 * result + hql.hashCode();
         result = 37 * result + maxResults;
         result = 37 * result + startOffset;
         
         return result;
      }
      
      /**
       * Get string representation of this object.
       * 
       * @return  String representation.
       */
      public String toString()
      {
         StringBuilder buf = new StringBuilder();
         buf.append(hql);
         buf.append("[");
         buf.append(startOffset);
         buf.append(":");
         buf.append(maxResults);
         buf.append("]");
         
         return buf.toString();
      }
   }

   public void commitOnImplicitTx()
   {
      context.get().forceCommitOnImplicitPop = true;
   }
}
