/*
** Module   : BlockManager.java
** Abstract : common "scaffolding" for executing Progress-compatible code
**            blocks
**
** Copyright (c) 2008-2009, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20080509   @38517 Created initial version. Implements the core
**                           "scaffolding" for executing 4GL-compatible
**                           code blocks. The API is designed to handle
**                           all possible block constructs and for the
**                           converted code to delegate the execution of
**                           those code blocks to this class. This greatly
**                           reduces the amount of code in the converted
**                           business logic.
** 002 GES 20080619   @38884 Now provides more detailed block types to the
**                           transaction manager. ON ENDKEY UNDO, RETRY is
**                           now signalled to the TM as well.
** 003 GES 20080625   @38958 Shifted the TM to the BlockType enum and
**                           moved the enum to be a separate class.
** 004 GES 20080701   @39058 FOR blocks/loops must honor user-defined
**                           ENDKEY behavior in preference to the default
**                           behavior for a query "off end".
** 005 GES 20080818   @39433 Added some UNDO stmt signatures. Fix to add
**                           the error prop to DO blocks/loops that are
**                           full transactions (it is implicit).
** 006 GES 20080822   @39519 Added missing function for GO-ON editing
**                           blocks, implicit ERROR property support for
**                           DO blocks/loops that are marked as FULL
**                           transaction types AND ILP support to convert
**                           NEXT to LEAVE.
** 007 GES 20080829   @39694 Added honorStop() support.
** 008 GES 20080924   @39934 Fix for QueryOffEndException processing.
**                           It must always cause a FOR loop or block
**                           to UNDO, LEAVE.
** 009 SIY 20081003   @40016 Added notification of listeners about 
**                           conditions triggered in blocks to which they
**                           are bound.
** 010 GES 20081106   @40361 Editing blocks must handle an error generated
**                           by the END-ERROR key by re-throwing it (just
**                           as is done in the UPDATE stmt). This needs
**                           to be done for all forms of editing block,
**                           not just one with the UPDATE stmt.
** 011 SIY 20081111   @40412 Added UNKNOWN Condition type, changed 
**                           visibility of exceptionToCondition, other 
**                           minor cleanups.
** 012 GES 20081126   @40740 Fixes for interactions counter and END-ERROR
**                           processing in EDITING and DO blocks/loops.
** 013 GES 20090115   @41132 Fix for loophole in condition processing.
**                           If processCondition() is about to re-throw
**                           an exception, it must trigger abnormalEnd()
**                           processing first.
** 014 SVL 20090415   @41799 Fixed error handing for editing blocks.
** 015 GES 20090512   @42148 Added execution statistics support.
** 016 CA  20090727   @43396 When a query associated or nested with a FOR 
**                           or FOR EACH block goes off end, its 
**                           associated buffers must not process the 
**                           undoables (the buffer must not reference any 
**                           record after the UNDO, LEAVE).
** 017 GES 20090729   @43443 Added validation processing hooks at the end of
**                           every block or loop, just before normal exit or
**                           iteration.
** 018 CA  20090731   @43471 Fixed query off-end processing. When the query 
**                           for a FOR or FOR EACH block finds no more records,
**                           the loop must be allowed to finish, by performing
**                           the implicit END-KEY action (UNDO, LEAVE), even
**                           if the user has defined an explicit END-KEY 
**                           action. For END-KEY conditions generated by a FIND
**                           statement, the block must execute the default
**                           END-KEY action if no explicit one is defined. 
**                           Else, it must execute the user-defined action.
** 019 CA  20090804   @43505 Fixed undo, leave when used inside a FOR block. 
**                           Cleanup TM related static imports. Fixed PRESELECT
**                           query in DO/REPEAT blocks - the record buffers 
**                           associated with the query must not reference any 
**                           record until a FIND statement is executed on the 
**                           buffer.
** 020 ECF 20090816   @43656 Fixed output parameter support for database
**                           fields. Modified topLevelBlock() and
**                           functionBlock() to call out to the current
**                           block's OutputParameterAssigner, if any, to copy
**                           output parameters' values back to the database
**                           fields with which they are associated.
** 021 GES 20090831   @43801 Don't raise ERROR in caller if the called code is
**                           a function.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.io.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.security.ContextLocal;
import com.goldencode.p2j.persist.Persistence;
import com.goldencode.p2j.persist.QueryOffEndException;
import com.goldencode.p2j.persist.QueryOffEndListener;

// the following are dependencies on the UI package which we would prefer to
// eliminate if the services can be cleanly abstracted without making the
// calling code worse from a readability perspective
import com.goldencode.p2j.ui.LogicalTerminal;
import com.goldencode.p2j.ui.CommonFrame;
import com.goldencode.p2j.ui.FrameElement;
import com.goldencode.p2j.ui.EventList;

import static com.goldencode.p2j.util.TransactionManager.*;

/**
 * Provides an external API that allows the Progress-compatible code block
 * semantics to be provided with a minimum of code in the caller. This class
 * encodes the "scaffolding" necessary to properly process code blocks with
 * the right behavior.  The external interface is designed to be statically
 * imported to make access simple, short and easy to read. The API has been
 * designed to minimize name conflicts.  The scaffolding code uses the
 * {@link TransactionManager} extensively to implement the proper behavior.
 */
public class BlockManager
{
   /** Stores context-local state variables. */
   private static final ContextContainer work = new ContextContainer();
   
   /** Encodes the possible levels of transaction support for each block. */
   public static enum TransactionType { NONE, SUB, FULL };
   
   /** Encodes the possible condition types. */
   public static enum Condition { ERROR, ENDKEY, STOP, QUIT, UNKNOWN };
   
   /** Encodes the possible actions in response to given conditions. */
   public static enum Action
   {
      LEAVE,
      NEXT, 
      RETRY, 
      RETURN_NORMAL,
      RETURN_CONSUME,
      RETURN_ERROR
   };
   
   /** Encodes the possible types of return statement. */
   private static enum ReturnType { NORMAL, CONSUME, ERROR };
   
   /** Encodes the possible responses to a flow control event. */
   private static enum Disposition { BREAK, CONTINUE, THROW, NONE };
   
   /** Directory entry to define an execution statistics output file. */
   private static final String STATS_FILE = "block-execution-statistics";
   
   /** Output filename in which to write execution statistics. */
   private static String statsFile = null;
   
   /** Stores execution statistics. */
   private static Map<String, Integer> stats = null;
   
   /** Stores execution statistics. */
   private static Map<String, Deque<Long>> times = null;

   /** Stores execution statistics. */
   private static Map<String, Long> time = null;
   
   static
   {
      Directory dir = DirectoryManager.getInstance();
      statsFile = dir.getString(Directory.ID_RELATIVE, STATS_FILE, null);
      
      if (statsFile != null)
      {
         stats = new HashMap<String, Integer>();
         StandardServer.register(new StatsHelper());
      }
   }
   
   /**
    * Execute the given external procedure block with a default transaction
    * level of <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    */
   public static void externalProcedure(Block block)
   {
      externalProcedure(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given external procedure block with the given transaction
    * level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    */
   public static void externalProcedure(TransactionType lvl, Block block)
   {
      topLevelBlock(lvl, BlockType.EXTERNAL_PROC, block);
   }
   
   /**
    * Execute the given internal procedure block with a default transaction
    * level of <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    */
   public static void internalProcedure(Block block)
   {
      internalProcedure(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given internal procedure block with the given transaction
    * level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    */
   public static void internalProcedure(TransactionType lvl, Block block)
   {
      topLevelBlock(lvl, BlockType.INTERNAL_PROC, block);
   }
   
   /**
    * Execute the given function block with a default transaction level of
    * <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static character characterFunction(Block block)
   {
      return characterFunction(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static character characterFunction(TransactionType lvl, Block block)
   {
      return (character) functionBlock(lvl, block, character.class);
   }
   
   /**
    * Execute the given function block with a default transaction level of
    * <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static logical logicalFunction(Block block)
   {
      return logicalFunction(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static logical logicalFunction(TransactionType lvl, Block block)
   {
      return (logical) functionBlock(lvl, block, logical.class);
   }
   
   /**
    * Execute the given function block with a default transaction level of
    * <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static integer integerFunction(Block block)
   {
      return integerFunction(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static integer integerFunction(TransactionType lvl, Block block)
   {
      return (integer) functionBlock(lvl, block, integer.class);
   }
   
   /**
    * Execute the given function block with a default transaction level of
    * <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static decimal decimalFunction(Block block)
   {
      return decimalFunction(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static decimal decimalFunction(TransactionType lvl, Block block)
   {
      return (decimal) functionBlock(lvl, block, decimal.class);
   }
   
   /**
    * Execute the given function block with a default transaction level of
    * <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static date dateFunction(Block block)
   {
      return dateFunction(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static date dateFunction(TransactionType lvl, Block block)
   {
      return (date) functionBlock(lvl, block, date.class);
   }
   
   /**
    * Execute the given function block with a default transaction level of
    * <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static handle handleFunction(Block block)
   {
      return handleFunction(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static handle handleFunction(TransactionType lvl, Block block)
   {
      return (handle) functionBlock(lvl, block, handle.class);
   }
   
   /**
    * Execute the given function block with a default transaction level of
    * <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static raw rawFunction(Block block)
   {
      return rawFunction(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static raw rawFunction(TransactionType lvl, Block block)
   {
      return (raw) functionBlock(lvl, block, raw.class);
   }
   
   /**
    * Execute the given function block with a default transaction level of
    * <code>TransactionType.SUB</code>.
    *
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static memptr memptrFunction(Block block)
   {
      return memptrFunction(TransactionType.SUB, block);
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    lvl
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   public static memptr memptrFunction(TransactionType lvl, Block block)
   {
      return (memptr) functionBlock(lvl, block, memptr.class);
   }
   
   /**
    * Execute a simple DO block (not a loop) with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    *
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void doBlock(String label, Block block)
   {
      doBlockWorker(TransactionType.NONE, null, label, null, block);
   }
   
   /**
    * Execute a simple DO block (not a loop) with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no user-defined properties.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void doBlock(String[] enclosing, String label, Block block)
   {
      doBlockWorker(TransactionType.NONE, enclosing, label, null, block);
   }
   
   /**
    * Execute a simple DO block (not a loop) with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void doBlock(TransactionType lvl, String label, Block block)
   {
      doBlockWorker(lvl, null, label, null, block);
   }
   
   /**
    * Execute a simple DO block (not a loop) with the given configuration
    * and no user-defined or default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void doBlock(TransactionType lvl,
                              String[]        enclosing,
                              String          label,
                              Block           block)
   {
      doBlockWorker(lvl, enclosing, label, null, block);
   }

   /**
    * Execute a simple DO block (not a loop) with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doBlock(String label, OnPhrase[] on, Block block)
   {
      doBlockWorker(TransactionType.NONE, null, label, on, block);
   }
   
   /**
    * Execute a simple DO block (not a loop) with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doBlock(String[]   enclosing,
                              String     label,
                              OnPhrase[] on,
                              Block      block)
   {
      doBlockWorker(TransactionType.NONE, enclosing, label, on, block);
   }
   
   /**
    * Execute a simple DO block (not a loop) with the given configuration
    * and no default properties. In this case, there are no enclosing
    * non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doBlock(TransactionType lvl,
                              String          label,
                              OnPhrase[]      on,
                              Block           block)
   {
      doBlockWorker(lvl, null, label, on, block);
   }
   
   /**                                              
    * Execute a simple DO block (not a loop) with the given configuration
    * and no default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doBlock(TransactionType lvl,
                              String[]        enclosing,
                              String          label,
                              OnPhrase[]      on,
                              Block           block)
   {
      doBlockWorker(lvl, enclosing, label, on, block);
   }
   
   /**
    * Execute a DO WHILE expression loop with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    *
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void doWhile(String            label,
                              LogicalExpression expr,
                              Block             block)
   {
      doLoopWorker(TransactionType.NONE,
                   null,
                   label,
                   null,
                   expr,
                   null,
                   block);
   }
   
   /**
    * Execute a DO WHILE expression loop with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no user-defined properties.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void doWhile(String[]          enclosing,
                              String            label,
                              LogicalExpression expr,
                              Block             block)
   {
      doLoopWorker(TransactionType.NONE,
                   enclosing,
                   label,
                   null,
                   expr,
                   null,
                   block);
   }
   
   /**
    * Execute a DO WHILE expression loop with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void doWhile(TransactionType   lvl,
                              String            label,
                              LogicalExpression expr,
                              Block             block)
   {
      doLoopWorker(lvl, null, label, null, expr, null, block);
   }
   
   /**
    * Execute a DO WHILE expression loop with the given configuration
    * and no user-defined or default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void doWhile(TransactionType   lvl,
                              String[]          enclosing,
                              String            label,
                              LogicalExpression expr,
                              Block             block)
   {
      doLoopWorker(lvl, enclosing, label, null, expr, null, block);
   }

   /**
    * Execute a DO WHILE expression loop with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doWhile(String            label,
                              LogicalExpression expr,
                              OnPhrase[]        on,
                              Block             block)
   {
      doLoopWorker(TransactionType.NONE, null, label, null, expr, on, block);
   }
   
   /**
    * Execute a DO WHILE expression loop with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doWhile(String[]          enclosing,
                              String            label,
                              LogicalExpression expr,
                              OnPhrase[]        on,
                              Block             block)
   {
      doLoopWorker(TransactionType.NONE,
                   enclosing,
                   label,
                   null,
                   expr,
                   on,
                   block);
   }
   
   /**
    * Execute a DO WHILE expression loop with the given configuration
    * and no default properties. In this case, there are no enclosing
    * non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doWhile(TransactionType   lvl,
                              String            label,
                              LogicalExpression expr,
                              OnPhrase[]        on,
                              Block             block)
   {
      doLoopWorker(lvl, null, label, null, expr, on, block);
   }
   
   /**
    * Execute a DO WHILE expression loop with the given configuration
    * and no default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr      
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doWhile(TransactionType   lvl,
                              String[]          enclosing,
                              String            label,
                              LogicalExpression expr,
                              OnPhrase[]        on,
                              Block             block)
   {
      doLoopWorker(lvl, enclosing, label, null, expr, on, block);
   }
             
   /**
    * Execute a DO TO expression loop with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void doTo(String label, ToClause to, Block block)
   {
      doLoopWorker(TransactionType.NONE, null, label, to, null, null, block);
   }
   
   /**
    * Execute a DO TO expression loop with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no user-defined properties.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void doTo(String[] enclosing,
                           String   label,
                           ToClause to,
                           Block    block)
   {
      doLoopWorker(TransactionType.NONE,
                   enclosing,
                   label,
                   to,
                   null,
                   null,
                   block);
   }
   
   /**
    * Execute a DO TO expression loop with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void doTo(TransactionType lvl,
                           String          label,
                           ToClause        to,
                           Block           block)
   {
      doLoopWorker(lvl, null, label, to, null, null, block);
   }
   
   /**
    * Execute a DO TO expression loop with the given configuration
    * and no user-defined or default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void doTo(TransactionType lvl,
                           String[]        enclosing,
                           String          label,
                           ToClause        to,
                           Block           block)
   {
      doLoopWorker(lvl, enclosing, label, to, null, null, block);
   }

   /**
    * Execute a DO TO expression loop with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used and there are no enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doTo(String     label,
                           ToClause   to,
                           OnPhrase[] on,
                           Block      block)
   {
      doLoopWorker(TransactionType.NONE, null, label, to, null, on, block);
   }
   
   /**
    * Execute a DO TO expression loop with the given configuration
    * and no default properties. <code>TransactionType.NONE</code> will be
    * used.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doTo(String[]   enclosing,
                           String     label,
                           ToClause   to,
                           OnPhrase[] on,
                           Block      block)
   {
      doLoopWorker(TransactionType.NONE,
                   enclosing,
                   label,
                   to,
                   null,
                   on,
                   block);
   }
   
   /**
    * Execute a DO TO expression loop with the given configuration
    * and no default properties. In this case, there are no enclosing
    * non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doTo(TransactionType lvl,
                           String          label,
                           ToClause        to,
                           OnPhrase[]      on,
                           Block           block)
   {
      doLoopWorker(lvl, null, label, to, null, on, block);
   }
   
   /**
    * Execute a DO TO expression loop with the given configuration
    * and no default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doTo(TransactionType lvl,
                           String[]        enclosing,
                           String          label,
                           ToClause        to,
                           OnPhrase[]      on,
                           Block           block)
   {
      doLoopWorker(lvl, enclosing, label, to, null, on, block);
   }
   
   /**
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.NONE</code> will be used and there are no
    * enclosing non-managed blocks. In addition, there are no user-defined
    * properties.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    block
    *           The code to execute.
    */
   public static void doToWhile(String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                Block             block)
   {
      doLoopWorker(TransactionType.NONE,
                   null,
                   label,
                   to,
                   expr,
                   null,
                   block);
   }
   
   /**
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.NONE</code> will be used and there are no
    * user-defined properties.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    block
    *           The code to execute.
    */
   public static void doToWhile(String[]          enclosing,
                                String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                Block             block)
   {
      doLoopWorker(TransactionType.NONE,
                   enclosing,
                   label,
                   to,
                   expr,
                   null,
                   block);
   }
   
   /**
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    * <p>
    * There are no user-defined properties and no enclosing non-managed
    * blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    block
    *           The code to execute.
    */
   public static void doToWhile(TransactionType   lvl,
                                String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                Block             block)
   {
      doLoopWorker(lvl, null, label, to, expr, null, block);
   }
   
   /**
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    * <p>
    * There are no user-defined properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    block
    *           The code to execute.
    */
   public static void doToWhile(TransactionType   lvl,
                                String[]          enclosing,
                                String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                Block             block)
   {
      doLoopWorker(lvl, enclosing, label, to, expr, null, block);
   }

   /**
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.NONE</code> will be used and there are no
    * enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doToWhile(String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                OnPhrase[]        on,
                                Block             block)
   {
      doLoopWorker(TransactionType.NONE, null, label, to, expr, on, block);
   }
   
   /**
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.NONE</code> will be used.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doToWhile(String[]          enclosing,
                                String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                OnPhrase[]        on,
                                Block             block)
   {
      doLoopWorker(TransactionType.NONE,
                   enclosing,
                   label,
                   to,
                   expr,
                   on,
                   block);
   }
   
   /**
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    * <p>
    * There are no enclosing non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doToWhile(TransactionType   lvl,
                                String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                OnPhrase[]        on,
                                Block             block)
   {
      doLoopWorker(lvl, null, label, to, expr, on, block);
   }
   
   /**
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void doToWhile(TransactionType   lvl,
                                String[]          enclosing,
                                String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                OnPhrase[]        on,
                                Block             block)
   {
      doLoopWorker(lvl, enclosing, label, to, expr, on, block);
   }
             
   /**
    * Execute a simple REPEAT (infinite loop) with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    *
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void repeat(String label, Block block)
   {
      repeatWorker(TransactionType.SUB, null, label, null, null, null, block);
   }
   
   /**
    * Execute a simple REPEAT (infinite loop) with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void repeat(String[] enclosing, String label, Block block)
   {
      repeatWorker(TransactionType.SUB,
                   enclosing,
                   label,
                   null,
                   null,
                   null,
                   block);
   }
   
   /**
    * Execute a simple REPEAT (infinite loop) with the given configuration
    * and default properties.  There are no user-defined properties and
    * there are no enclosing non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void repeat(TransactionType lvl, String label, Block block)
   {
      repeatWorker(lvl, null, label, null, null, null, block);
   }
   
   /**
    * Execute a simple REPEAT (infinite loop) with the given configuration
    * and default properties.  There are no user-defined properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void repeat(TransactionType lvl,
                             String[]        enclosing,
                             String          label,
                             Block           block)
   {
      repeatWorker(lvl, enclosing, label, null, null, null, block);
   }

   /**
    * Execute a simple REPEAT (infinite loop) with the given configuration
    * and default properties.  <code>TransactionType.SUB</code> will be used
    * and there are no enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeat(String label, OnPhrase[] on, Block block)
   {
      repeatWorker(TransactionType.SUB, null, label, null, null, on, block);
   }
   
   /**
    * Execute a simple REPEAT (infinite loop) with the given configuration
    * and default properties.  <code>TransactionType.SUB</code> will be used.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeat(String[]   enclosing,
                             String     label,
                             OnPhrase[] on,
                             Block      block)
   {
      repeatWorker(TransactionType.SUB,
                   enclosing,
                   label,
                   null,
                   null,
                   on,
                   block);
   }
   
   /**
    * Execute a simple REPEAT (infinite loop) with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeat(TransactionType lvl,
                             String          label,
                             OnPhrase[]      on,
                             Block           block)
   {
      repeatWorker(lvl, null, label, null, null, on, block);
   }
   
   /**                                              
    * Execute a simple REPEAT (infinite loop) with the given configuration
    * and default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeat(TransactionType lvl,
                             String[]        enclosing,
                             String          label,
                             OnPhrase[]      on,
                             Block           block)
   {
      repeatWorker(lvl, enclosing, label, null, null, on, block);
   }
   
   /**
    * Execute a REPEAT WHILE expression loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    *
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void repeatWhile(String            label,
                                  LogicalExpression expr,
                                  Block             block)
   {
      repeatWorker(TransactionType.SUB,
                   null,
                   label,
                   null,
                   expr,
                   null,
                   block);
   }
   
   /**
    * Execute a REPEAT WHILE expression loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void repeatWhile(String[]          enclosing,
                                  String            label,
                                  LogicalExpression expr,
                                  Block             block)
   {
      repeatWorker(TransactionType.SUB,
                   enclosing,
                   label,
                   null,
                   expr,
                   null,
                   block);
   }
   
   /**
    * Execute a REPEAT WHILE expression loop with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void repeatWhile(TransactionType   lvl,
                                  String            label,
                                  LogicalExpression expr,
                                  Block             block)
   {
      repeatWorker(lvl, null, label, null, expr, null, block);
   }
   
   /**
    * Execute a REPEAT WHILE expression loop with the given configuration
    * and no user-defined or default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void repeatWhile(TransactionType   lvl,
                                  String[]          enclosing,
                                  String            label,
                                  LogicalExpression expr,
                                  Block             block)
   {
      repeatWorker(lvl, enclosing, label, null, expr, null, block);
   }

   /**
    * Execute a REPEAT WHILE expression loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatWhile(String            label,
                                  LogicalExpression expr,
                                  OnPhrase[]        on,
                                  Block             block)
   {
      repeatWorker(TransactionType.SUB, null, label, null, expr, on, block);
   }
   
   /**
    * Execute a REPEAT WHILE expression loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatWhile(String[]          enclosing,
                                  String            label,
                                  LogicalExpression expr,
                                  OnPhrase[]        on,
                                  Block             block)
   {
      repeatWorker(TransactionType.SUB,
                   enclosing,
                   label,
                   null,
                   expr,
                   on,
                   block);
   }
   
   /**
    * Execute a REPEAT WHILE expression loop with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatWhile(TransactionType   lvl,
                                  String            label,
                                  LogicalExpression expr,
                                  OnPhrase[]        on,
                                  Block             block)
   {
      repeatWorker(lvl, null, label, null, expr, on, block);
   }
   
   /**
    * Execute a REPEAT WHILE expression loop with the given configuration
    * and default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr      
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatWhile(TransactionType   lvl,
                                  String[]          enclosing,
                                  String            label,
                                  LogicalExpression expr,
                                  OnPhrase[]        on,
                                  Block             block)
   {
      repeatWorker(lvl, enclosing, label, null, expr, on, block);
   }
             
   /**
    * Execute a REPEAT TO expression loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void repeatTo(String label, ToClause to, Block block)
   {
      repeatWorker(TransactionType.SUB, null, label, to, null, null, block);
   }
   
   /**
    * Execute a REPEAT TO expression loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void repeatTo(String[] enclosing,
                               String   label,
                               ToClause to,
                               Block    block)
   {
      repeatWorker(TransactionType.SUB,
                   enclosing,
                   label,
                   to,
                   null,
                   null,
                   block);
   }
   
   /**
    * Execute a REPEAT TO expression loop with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void repeatTo(TransactionType lvl,
                               String          label,
                               ToClause        to,
                               Block           block)
   {
      repeatWorker(lvl, null, label, to, null, null, block);
   }
   
   /**
    * Execute a REPEAT TO expression loop with the given configuration
    * and no user-defined or default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void repeatTo(TransactionType lvl,
                               String[]        enclosing,
                               String          label,
                               ToClause        to,
                               Block           block)
   {
      repeatWorker(lvl, enclosing, label, to, null, null, block);
   }

   /**
    * Execute a REPEAT TO expression loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatTo(String     label,
                               ToClause   to,
                               OnPhrase[] on,
                               Block      block)
   {
      repeatWorker(TransactionType.SUB, null, label, to, null, on, block);
   }
   
   /**
    * Execute a REPEAT TO expression loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatTo(String[]   enclosing,
                               String     label,
                               ToClause   to,
                               OnPhrase[] on,
                               Block      block)
   {
      repeatWorker(TransactionType.SUB,
                   enclosing,
                   label,
                   to,
                   null,
                   on,
                   block);
   }
   
   /**
    * Execute a REPEAT TO expression loop with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatTo(TransactionType lvl,
                               String          label,
                               ToClause        to,
                               OnPhrase[]      on,
                               Block           block)
   {
      repeatWorker(lvl, null, label, to, null, on, block);
   }
   
   /**
    * Execute a REPEAT TO expression loop with the given configuration
    * and default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatTo(TransactionType lvl,
                               String[]        enclosing,
                               String          label,
                               ToClause        to,
                               OnPhrase[]      on,
                               Block           block)
   {
      repeatWorker(lvl, enclosing, label, to, null, on, block);
   }
   
   /**
    * Execute a REPEAT WHILE expression, REPEAT TO var = val1 to val2 by val3
    * or REPEAT TO WHILE expression loop with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * enclosing non-managed blocks. In addition, there are no user-defined
    * properties.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    block
    *           The code to execute.
    */
   public static void repeatToWhile(String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    Block             block)
   {
      repeatWorker(TransactionType.SUB,
                   null,
                   label,
                   to,
                   expr,
                   null,
                   block);
   }
   
   /**
    * Execute a REPEAT WHILE expression, REPEAT TO var = val1 to val2 by val3
    * or REPEAT TO WHILE expression loop with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * user-defined properties.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    block
    *           The code to execute.
    */
   public static void repeatToWhile(String[]          enclosing,
                                    String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    Block             block)
   {
      repeatWorker(TransactionType.SUB,
                   enclosing,
                   label,
                   to,
                   expr,
                   null,
                   block);
   }
   
   /**
    * Execute a REPEAT WHILE expression, REPEAT TO var = val1 to val2 by val3
    * or REPEAT TO WHILE expression loop with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no user-defined properties and no enclosing non-managed
    * blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    block
    *           The code to execute.
    */
   public static void repeatToWhile(TransactionType   lvl,
                                    String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    Block             block)
   {
      repeatWorker(lvl, null, label, to, expr, null, block);
   }
   
   /**
    * Execute a REPEAT WHILE expression, REPEAT TO var = val1 to val2 by val3
    * or REPEAT TO WHILE expression loop with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no user-defined properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    block
    *           The code to execute.
    */
   public static void repeatToWhile(TransactionType   lvl,
                                    String[]          enclosing,
                                    String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    Block             block)
   {
      repeatWorker(lvl, enclosing, label, to, expr, null, block);
   }

   /**
    * Execute a REPEAT WHILE expression, REPEAT TO var = val1 to val2 by val3
    * or REPEAT TO WHILE expression loop with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatToWhile(String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      repeatWorker(TransactionType.SUB, null, label, to, expr, on, block);
   }
   
   /**
    * Execute a REPEAT WHILE expression, REPEAT TO var = val1 to val2 by val3
    * or REPEAT TO WHILE expression loop with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatToWhile(String[]          enclosing,
                                    String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      repeatWorker(TransactionType.SUB,
                   enclosing,
                   label,
                   to,
                   expr,
                   on,
                   block);
   }
   
   /**
    * Execute a REPEAT WHILE expression, REPEAT TO var = val1 to val2 by val3
    * or REPEAT TO WHILE expression loop with the given configuration and
    * default properties.  At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no enclosing non-managed blocks.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatToWhile(TransactionType   lvl,
                                    String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      repeatWorker(lvl, null, label, to, expr, on, block);
   }
   
   /**
    * Execute a REPEAT WHILE expression, REPEAT TO var = val1 to val2 by val3
    * or REPEAT TO WHILE expression loop with the given configuration and
    * default properties.  At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void repeatToWhile(TransactionType   lvl,
                                    String[]          enclosing,
                                    String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      repeatWorker(lvl, enclosing, label, to, expr, on, block);
   }
             
   /**
    * Execute a FOR FIRST or FOR LAST block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void forBlock(String label, Block block)
   {
      forBlockWorker(TransactionType.SUB,
                     null,
                     label,
                     null,
                     null,
                     null,
                     block);
   }
   
   /**
    * Execute a FOR FIRST or FOR LAST block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void forBlock(String[] enclosing, String label, Block block)
   {
      forBlockWorker(TransactionType.SUB,
                     enclosing,
                     label,
                     null,
                     null,
                     null,
                     block);
   }
   
   /**
    * Execute a FOR FIRST or FOR LAST block with the given configuration
    * and default properties.  There are no user-defined properties and
    * there are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void forBlock(TransactionType lvl, String label, Block block)
   {
      forBlockWorker(lvl, null, label, null, null, null, block);
   }
   
   /**
    * Execute a FOR FIRST or FOR LAST block with the given configuration
    * and default properties.  There are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void forBlock(TransactionType lvl,
                               String[]        enclosing,
                               String          label,
                               Block           block)
   {
      forBlockWorker(lvl, enclosing, label, null, null, null, block);
   }

   /**
    * Execute a FOR FIRST or FOR LAST block with the given configuration
    * and default properties.  <code>TransactionType.SUB</code> will be used
    * and there are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlock(String label, OnPhrase[] on, Block block)
   {
      forBlockWorker(TransactionType.SUB,
                     null,
                     label,
                     null,
                     null,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST or FOR LAST block with the given configuration
    * and default properties.  <code>TransactionType.SUB</code> will be used.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlock(String[]   enclosing,
                               String     label,
                               OnPhrase[] on,
                               Block      block)
   {
      forBlockWorker(TransactionType.SUB,
                     enclosing,
                     label,
                     null,
                     null,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST or FOR LAST block with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlock(TransactionType lvl,
                               String          label,
                               OnPhrase[]      on,
                               Block           block)
   {
      forBlockWorker(lvl, null, label, null, null, on, block);
   }
   
   /**                                              
    * Execute a FOR FIRST or FOR LAST block with the given configuration
    * and default properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlock(TransactionType lvl,
                               String[]        enclosing,
                               String          label,
                               OnPhrase[]      on,
                               Block           block)
   {
      forBlockWorker(lvl,
                     enclosing,
                     label,
                     null,
                     null,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockWhile(String            label,
                                    LogicalExpression expr,
                                    Block             block)
   {
      forBlockWorker(TransactionType.SUB,
                     null,
                     label,
                     null,
                     expr,
                     null,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockWhile(String[]          enclosing,
                                    String            label,
                                    LogicalExpression expr,
                                    Block             block)
   {
      forBlockWorker(TransactionType.SUB,
                     enclosing,
                     label,
                     null,
                     expr,
                     null,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockWhile(TransactionType   lvl,
                                    String            label,
                                    LogicalExpression expr,
                                    Block             block)
   {
      forBlockWorker(lvl, null, label, null, expr, null, block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block with the given configuration
    * and no user-defined or default properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockWhile(TransactionType   lvl,
                                    String[]          enclosing,
                                    String            label,
                                    LogicalExpression expr,
                                    Block             block)
   {
      forBlockWorker(lvl,
                     enclosing,
                     label,
                     null,
                     expr,
                     null,
                     block);
   }

   /**
    * Execute a FOR FIRST/LAST WHILE block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks.
    *
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockWhile(String            label,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      forBlockWorker(TransactionType.SUB,
                     null,
                     label,
                     null,
                     expr,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockWhile(String[]          enclosing,
                                    String            label,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      forBlockWorker(TransactionType.SUB,
                     enclosing,
                     label,
                     null,
                     expr,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockWhile(TransactionType   lvl,
                                    String            label,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      forBlockWorker(lvl, null, label, null, expr, on, block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block with the given configuration
    * and default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockWhile(TransactionType   lvl,
                                    String[]          enclosing,
                                    String            label,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      forBlockWorker(lvl, enclosing, label, null, expr, on, block);
   }
             
   /**
    * Execute a FOR FIRST/LAST TO block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockTo(String label, ToClause to, Block block)
   {
      forBlockWorker(TransactionType.SUB,
                     null,
                     label,
                     to,
                     null,
                     null,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST TO block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockTo(String[] enclosing,
                                 String   label,
                                 ToClause to,
                                 Block    block)
   {
      forBlockWorker(TransactionType.SUB,
                     enclosing,
                     label,
                     to,
                     null,
                     null,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST TO block with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockTo(TransactionType lvl,
                                 String          label,
                                 ToClause        to,
                                 Block           block)
   {
      forBlockWorker(lvl, null, label, to, null, null, block);
   }
   
   /**
    * Execute a FOR FIRST/LAST TO block with the given configuration
    * and no user-defined or default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockTo(TransactionType lvl,
                                 String[]        enclosing,
                                 String          label,
                                 ToClause        to,
                                 Block           block)
   {
      forBlockWorker(lvl, enclosing, label, to, null, null, block);
   }

   /**
    * Execute a FOR FIRST/LAST TO block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockTo(String     label,
                                 ToClause   to,
                                 OnPhrase[] on,
                                 Block      block)
   {
      forBlockWorker(TransactionType.SUB,
                     (String[]) null,
                     label,
                     to,
                     null,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST TO block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockTo(String[]   enclosing,
                                 String     label,
                                 ToClause   to,
                                 OnPhrase[] on,
                                 Block      block)
   {
      forBlockWorker(TransactionType.SUB,
                     enclosing,
                     label,
                     to,
                     null,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST TO block with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockTo(TransactionType lvl,
                                 String          label,
                                 ToClause        to,
                                 OnPhrase[]      on,
                                 Block           block)
   {
      forBlockWorker(lvl, null, label, to, null, on, block);
   }
   
   /**
    * Execute a FOR FIRST/LAST TO block with the given configuration
    * and default properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockTo(TransactionType lvl,
                                 String[]        enclosing,
                                 String          label,
                                 ToClause        to,
                                 OnPhrase[]      on,
                                 Block           block)
   {
      forBlockWorker(lvl, enclosing, label, to, null, on, block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * enclosing non-managed blocks. In addition, there are no user-defined
    * properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockToWhile(String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      Block             block)
   {
      forBlockWorker(TransactionType.SUB,
                     null,
                     label,
                     to,
                     expr,
                     null,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockToWhile(String[]          enclosing,
                                      String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      Block             block)
   {
      forBlockWorker(TransactionType.SUB,
                     enclosing,
                     label,
                     to,
                     expr,
                     null,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no user-defined properties and no enclosing non-managed
    * blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockToWhile(TransactionType   lvl,
                                      String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      Block             block)
   {
      forBlockWorker(lvl, null, label, to, expr, null, block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockToWhile(TransactionType   lvl,
                                      String[]          enclosing,
                                      String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      Block             block)
   {
      forBlockWorker(lvl, enclosing, label, to, expr, null, block);
   }

   /**
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockToWhile(String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      OnPhrase[]        on,
                                      Block             block)
   {
      forBlockWorker(TransactionType.SUB,
                     (String[]) null,
                     label,
                     to,
                     expr,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockToWhile(String[]          enclosing,
                                      String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      OnPhrase[]        on,
                                      Block             block)
   {
      forBlockWorker(TransactionType.SUB,
                     enclosing,
                     label,
                     to,
                     expr,
                     on,
                     block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * default properties.  At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockToWhile(TransactionType   lvl,
                                      String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      OnPhrase[]        on,
                                      Block             block)
   {
      forBlockWorker(lvl, null, label, to, expr, on, block);
   }
   
   /**
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * default properties.  At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forBlockToWhile(TransactionType   lvl,
                                      String[]          enclosing,
                                      String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      OnPhrase[]        on,
                                      Block             block)
   {
      forBlockWorker(lvl, enclosing, label, to, expr, on, block);
   }
             
   /**
    * Execute a FOR EACH loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void forEach(String label, Block block)
   {
      forEachWorker(TransactionType.SUB,
                    null,
                    label,
                    null,
                    null,
                    null,
                    block);
   }
   
   /**
    * Execute a FOR EACH loop with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void forEach(String[] enclosing, String label, Block block)
   {
      forEachWorker(TransactionType.SUB,
                    enclosing,
                    label,
                    null,
                    null,
                    null,
                    block);
   }
   
   /**
    * Execute a FOR EACH loop with the given configuration
    * and default properties.  There are no user-defined properties and
    * there are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void forEach(TransactionType lvl, String label, Block block)
   {
      forEachWorker(lvl, null, label, null, null, null, block);
   }
   
   /**
    * Execute a FOR EACH loop with the given configuration
    * and default properties.  There are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    block
    *           The code to execute.
    */
   public static void forEach(TransactionType lvl,
                              String[]        enclosing,
                              String          label,
                              Block           block)
   {
      forEachWorker(lvl, enclosing, label, null, null, null, block);
   }

   /**
    * Execute a FOR EACH loop with the given configuration
    * and default properties.  <code>TransactionType.SUB</code> will be used
    * and there are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEach(String label, OnPhrase[] on, Block block)
   {
      forEachWorker(TransactionType.SUB,
                    null,
                    label,
                    null,
                    null,
                    on,
                    block);
   }
   
   /**
    * Execute a FOR EACH loop with the given configuration
    * and default properties.  <code>TransactionType.SUB</code> will be used.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEach(String[]   enclosing,
                              String     label,
                              OnPhrase[] on,
                              Block      block)
   {
      forEachWorker(TransactionType.SUB,
                    enclosing,
                    label,
                    null,
                    null,
                    on,
                    block);
   }
   
   /**
    * Execute a FOR EACH loop with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEach(TransactionType lvl,
                              String          label,
                              OnPhrase[]      on,
                              Block           block)
   {
      forEachWorker(lvl, null, label, null, null, on, block);
   }
   
   /**                                              
    * Execute a FOR EACH loop with the given configuration
    * and default properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEach(TransactionType lvl,
                              String[]        enclosing,
                              String          label,
                              OnPhrase[]      on,
                              Block           block)
   {
      forEachWorker(lvl, enclosing, label, null, null, on, block);
   }
   
   /**
    * Execute a FOR EACH WHILE block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forEachWhile(String            label,
                                   LogicalExpression expr,
                                   Block             block)
   {
      forEachWorker(TransactionType.SUB,
                    null,
                    label,
                    null,
                    expr,
                    null,
                    block);
   }
   
   /**
    * Execute a FOR EACH WHILE block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forEachWhile(String[]          enclosing,
                                   String            label,
                                   LogicalExpression expr,
                                   Block             block)
   {
      forEachWorker(TransactionType.SUB,
                    enclosing,
                    label,
                    null,
                    expr,
                    null,
                    block);
   }
   
   /**
    * Execute a FOR EACH WHILE block with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forEachWhile(TransactionType   lvl,
                                   String            label,
                                   LogicalExpression expr,
                                   Block             block)
   {
      forEachWorker(lvl, null, label, null, expr, null, block);
   }
   
   /**
    * Execute a FOR EACH WHILE block with the given configuration
    * and no user-defined or default properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forEachWhile(TransactionType   lvl,
                                   String[]          enclosing,
                                   String            label,
                                   LogicalExpression expr,
                                   Block             block)
   {
      forEachWorker(lvl, enclosing, label, null, expr, null, block);
   }

   /**
    * Execute a FOR EACH WHILE block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachWhile(String            label,
                                   LogicalExpression expr,
                                   OnPhrase[]        on,
                                   Block             block)
   {
      forEachWorker(TransactionType.SUB,
                    null,
                    label,
                    null,
                    expr,
                    on,
                    block);
   }
   
   /**
    * Execute a FOR EACH WHILE block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachWhile(String[]          enclosing,
                                   String            label,
                                   LogicalExpression expr,
                                   OnPhrase[]        on,
                                   Block             block)
   {
      forEachWorker(TransactionType.SUB,
                    enclosing,
                    label,
                    null,
                    expr,
                    on,
                    block);
   }
   
   /**
    * Execute a FOR EACH WHILE block with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    expr
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachWhile(TransactionType   lvl,
                                   String            label,
                                   LogicalExpression expr,
                                   OnPhrase[]        on,
                                   Block             block)
   {
      forEachWorker(lvl, null, label, null, expr, on, block);
   }
   
   /**
    * Execute a FOR EACH WHILE block with the given configuration
    * and default properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    expr      
    *           The WHILE clause that defines the loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachWhile(TransactionType   lvl,
                                   String[]          enclosing,
                                   String            label,
                                   LogicalExpression expr,
                                   OnPhrase[]        on,
                                   Block             block)
   {
      forEachWorker(lvl, enclosing, label, null, expr, on, block);
   }
             
   /**
    * Execute a FOR EACH TO block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks. In addition, there
    * are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void forEachTo(String   label,
                                ToClause to,
                                Block    block)
   {
      forEachWorker(TransactionType.SUB,
                    null,
                    label,
                    to,
                    null,
                    null,
                    block);
   }
   
   /**
    * Execute a FOR EACH TO block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void forEachTo(String[] enclosing,
                                String   label,
                                ToClause to,
                                Block    block)
   {
      forEachWorker(TransactionType.SUB,
                    enclosing,
                    label,
                    to,
                    null,
                    null,
                    block);
   }
   
   /**
    * Execute a FOR EACH TO block with the given configuration
    * and no user-defined or default properties. In addition, there are no
    * enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void forEachTo(TransactionType lvl,
                                String          label,
                                ToClause        to,
                                Block           block)
   {
      forEachWorker(lvl, null, label, to, null, null, block);
   }
   
   /**
    * Execute a FOR EACH TO block with the given configuration
    * and no user-defined or default properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    block
    *           The code to execute.
    */
   public static void forEachTo(TransactionType lvl,
                                String[]        enclosing,
                                String          label,
                                ToClause        to,
                                Block           block)
   {
      forEachWorker(lvl, enclosing, label, to, null, null, block);
   }

   /**
    * Execute a FOR EACH TO block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used and there are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachTo(String     label,
                                ToClause   to,
                                OnPhrase[] on,
                                Block      block)
   {
      forEachWorker(TransactionType.SUB,
                    (String[]) null,
                    label,
                    to,
                    null,
                    on,
                    block);
   }
   
   /**
    * Execute a FOR EACH TO block with the given configuration
    * and default properties. <code>TransactionType.SUB</code> will be
    * used.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachTo(String[]   enclosing,
                                String     label,
                                ToClause   to,
                                OnPhrase[] on,
                                Block      block)
   {
      forEachWorker(TransactionType.SUB,
                    enclosing,
                    label,
                    to,
                    null,
                    on,
                    block);
   }
   
   /**
    * Execute a FOR EACH TO block with the given configuration
    * and default properties. In this case, there are no enclosing
    * non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachTo(TransactionType lvl,
                                String          label,
                                ToClause        to,
                                OnPhrase[]      on,
                                Block           block)
   {
      forEachWorker(lvl, null, label, to, null, on, block);
   }
   
   /**
    * Execute a FOR EACH TO block with the given configuration
    * and default properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachTo(TransactionType lvl,
                                String[]        enclosing,
                                String          label,
                                ToClause        to,
                                OnPhrase[]      on,
                                Block           block)
   {
      forEachWorker(lvl, enclosing, label, to, null, on, block);
   }
   
   /**
    * Execute a FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * enclosing non-managed blocks. In addition, there are no user-defined
    * properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forEachToWhile(String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     Block             block)
   {
      forEachWorker(TransactionType.SUB,
                    null,
                    label,
                    to,
                    expr,
                    null,
                    block);
   }
   
   /**
    * Execute a FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forEachToWhile(String[]          enclosing,
                                     String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     Block             block)
   {
      forEachWorker(TransactionType.SUB,
                    enclosing,
                    label,
                    to,
                    expr,
                    null,
                    block);
   }
   
   /**
    * Execute a FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no user-defined properties and no enclosing non-managed
    * blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forEachToWhile(TransactionType   lvl,
                                     String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     Block             block)
   {
      forEachWorker(lvl, null, label, to, expr, null, block);
   }
   
   /**
    * Execute a FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no user-defined properties.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    block
    *           The code to execute.
    */
   public static void forEachToWhile(TransactionType   lvl,
                                     String[]          enclosing,
                                     String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     Block             block)
   {
      forEachWorker(lvl, enclosing, label, to, expr, null, block);
   }

   /**
    * Execute a FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used and there are no
    * enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachToWhile(String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     OnPhrase[]        on,
                                     Block             block)
   {
      forEachWorker(TransactionType.SUB,
                    (String[]) null,
                    label,
                    to,
                    expr,
                    on,
                    block);
   }
   
   /**
    * Execute a FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * default properties. At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * <code>TransactionType.SUB</code> will be used.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachToWhile(String[]          enclosing,
                                     String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     OnPhrase[]        on,
                                     Block             block)
   {
      forEachWorker(TransactionType.SUB,
                    enclosing,
                    label,
                    to,
                    expr,
                    on,
                    block);
   }
   
   /**
    * Execute a FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * default properties.  At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * There are no enclosing non-managed blocks.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachToWhile(TransactionType   lvl,
                                     String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     OnPhrase[]        on,
                                     Block             block)
   {
      forEachWorker(lvl, null, label, to, expr, on, block);
   }
   
   /**
    * Execute a FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * default properties.  At least one of the to clause or while expression
    * must be passed and it is valid to have both passed in which case both
    * of the conditions are tested with a logical AND.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself.  This
    * means that the caller is responsible for all query processing. The
    * record retrieval must be done on the first line of code in the block
    * to duplicate the Progress behavior. Note that this is required to
    * handle all error cases and 4GL conditions properly.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   public static void forEachToWhile(TransactionType   lvl,
                                     String[]          enclosing,
                                     String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     OnPhrase[]        on,
                                     Block             block)
   {
      forEachWorker(lvl, enclosing, label, to, expr, on, block);
   }
             
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>PROMPT-FOR</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void promptForEditing(CommonFrame    frame,
                                       FrameElement[] widgets,
                                       String         label,
                                       Block          block)
   {
      editingBlock(frame,
                   widgets,
                   null, 
                   null,
                   label,
                   block,
                   CommonFrame.PROMPT_FOR);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>PROMPT-FOR</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    enclosing
    *           The list of enclosing blocks of which this class is unaware.
    *           Any LEAVE or NEXT that references one of those enclosing
    *           blocks must not unwind the stack further using an exception,
    *           but rather it must return and let the caller check to see if
    *           a deferred LEAVE or NEXT must be honored.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void promptForEditing(CommonFrame    frame,
                                       FrameElement[] widgets,
                                       String[]       enclosing,
                                       String         label,
                                       Block          block)
   {
      editingBlock(frame,
                   widgets,
                   null,
                   enclosing,
                   label,
                   block,
                   CommonFrame.PROMPT_FOR);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>PROMPT-FOR</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.  The exit from the
    * editing block will be based on the given <code>GO-ON</code> event list.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    goOn
    *           The exit conditions for this loop.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void promptForEditing(CommonFrame    frame,
                                       FrameElement[] widgets,
                                       EventList      goOn,
                                       String         label,
                                       Block          block)
   {
      editingBlock(frame,
                   widgets,
                   goOn,
                   null,
                   label,
                   block,
                   CommonFrame.PROMPT_FOR);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>PROMPT-FOR</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.  The exit from the
    * editing block will be based on the given <code>GO-ON</code> event list.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    goOn
    *           The exit conditions for this loop.
    * @param    enclosing
    *           The list of enclosing blocks of which this class is unaware.
    *           Any LEAVE or NEXT that references one of those enclosing
    *           blocks must not unwind the stack further using an exception,
    *           but rather it must return and let the caller check to see if
    *           a deferred LEAVE or NEXT must be honored.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void promptForEditing(CommonFrame    frame,
                                       FrameElement[] widgets,
                                       EventList      goOn,
                                       String[]       enclosing,
                                       String         label,
                                       Block          block)
   {
      editingBlock(frame,
                   widgets,
                   goOn,
                   enclosing,
                   label,
                   block,
                   CommonFrame.PROMPT_FOR);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>SET</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void setEditing(CommonFrame    frame,
                                 FrameElement[] widgets,
                                 String         label,
                                 Block          block)
   {
      editingBlock(frame,
                   widgets,
                   null,
                   null,
                   label,
                   block,
                   CommonFrame.SET);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>SET</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    enclosing
    *           The list of enclosing blocks of which this class is unaware.
    *           Any LEAVE or NEXT that references one of those enclosing
    *           blocks must not unwind the stack further using an exception,
    *           but rather it must return and let the caller check to see if
    *           a deferred LEAVE or NEXT must be honored.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void setEditing(CommonFrame    frame,
                                 FrameElement[] widgets,
                                 String[]       enclosing,
                                 String         label,
                                 Block          block)
   {
      editingBlock(frame,
                   widgets,
                   null,
                   enclosing,
                   label,
                   block,
                   CommonFrame.SET);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>SET</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.  The exit from the
    * editing block will be based on the given <code>GO-ON</code> event list.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    goOn
    *           The exit conditions for this loop.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void setEditing(CommonFrame    frame,
                                 FrameElement[] widgets,
                                 EventList      goOn,
                                 String         label,
                                 Block          block)
   {
      editingBlock(frame,
                   widgets,
                   goOn,
                   null,
                   label,
                   block,
                   CommonFrame.SET);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>SET</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.  The exit from the
    * editing block will be based on the given <code>GO-ON</code> event list.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    goOn
    *           The exit conditions for this loop.
    * @param    enclosing
    *           The list of enclosing blocks of which this class is unaware.
    *           Any LEAVE or NEXT that references one of those enclosing
    *           blocks must not unwind the stack further using an exception,
    *           but rather it must return and let the caller check to see if
    *           a deferred LEAVE or NEXT must be honored.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void setEditing(CommonFrame    frame,
                                 FrameElement[] widgets,
                                 EventList      goOn,
                                 String[]       enclosing,
                                 String         label,
                                 Block          block)
   {
      editingBlock(frame,
                   widgets,
                   goOn,
                   enclosing,
                   label,
                   block,
                   CommonFrame.SET);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>UPDATE</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void updateEditing(CommonFrame    frame,
                                    FrameElement[] widgets,
                                    String         label,
                                    Block          block)
   {
      editingBlock(frame,
                   widgets,
                   null,
                   null,
                   label,
                   block,
                   CommonFrame.UPDATE);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>UPDATE</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    enclosing
    *           The list of enclosing blocks of which this class is unaware.
    *           Any LEAVE or NEXT that references one of those enclosing
    *           blocks must not unwind the stack further using an exception,
    *           but rather it must return and let the caller check to see if
    *           a deferred LEAVE or NEXT must be honored.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void updateEditing(CommonFrame    frame,
                                    FrameElement[] widgets,
                                    String[]       enclosing,
                                    String         label,
                                    Block          block)
   {
      editingBlock(frame,
                   widgets,
                   null,
                   enclosing,
                   label,
                   block,
                   CommonFrame.UPDATE);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>UPDATE</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.  The exit from the
    * editing block will be based on the given <code>GO-ON</code> event list.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    goOn
    *           The exit conditions for this loop.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void updateEditing(CommonFrame    frame,
                                    FrameElement[] widgets,
                                    EventList      goOn,
                                    String         label,
                                    Block          block)
   {
      editingBlock(frame,
                   widgets,
                   goOn,
                   null,
                   label,
                   block,
                   CommonFrame.UPDATE);
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>UPDATE</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.  The exit from the
    * editing block will be based on the given <code>GO-ON</code> event list.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.
    * @param    goOn
    *           The exit conditions for this loop.
    * @param    enclosing
    *           The list of enclosing blocks of which this class is unaware.
    *           Any LEAVE or NEXT that references one of those enclosing
    *           blocks must not unwind the stack further using an exception,
    *           but rather it must return and let the caller check to see if
    *           a deferred LEAVE or NEXT must be honored.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    */
   public static void updateEditing(CommonFrame    frame,
                                    FrameElement[] widgets,
                                    EventList      goOn,
                                    String[]       enclosing,
                                    String         label,
                                    Block          block)
   {
      editingBlock(frame,
                   widgets,
                   goOn,
                   enclosing,
                   label,
                   block,
                   CommonFrame.UPDATE);
   }
   
   /**
    * Break out of the nearest enclosing block.
    * 
    * @throws   LeaveUnwindException 
    */
   public static void leave()
   throws LeaveUnwindException
   {
      leave("");
   }
   
   /**
    * Break out of the block associated with the given label.  This will
    * unwind the stack until the proper block has been reached.  At that
    * point the block will be left using the Java "break" statement.
    *
    * @param    label
    *           The name of the block to be left.
    *           
    * @throws   LeaveUnwindException 
    */
   public static void leave(String label)
   throws LeaveUnwindException
   {
      WorkArea wa = work.obtain();
      
      if (label != null)
      {
         wa.deferredLeave = label;
      }
      
      throw wa.leaveExc;
   }
   
   /**
    * Detect if there is a pending leave for the block named by the given
    * label. If so, this deferred leave flag will be cleared and 
    * <code>true</code> will be returned.  The caller is responsible for
    * ensuring that the leave request is fully granted as no further state
    * will be available for subsequent processing.
    *
    * @param    label
    *           The name of the block to be checked against the deferred
    *           leave block name.
    *
    * @return   <code>true</code> if there is a pending deferred leave and
    *           the name of the block to be left is the same as the given
    *           label. If the pending deferred block name is the "" (empty
    *           string), that is the equivalent of matching "any block" and
    *           so the first call to this method after that is set will cause
    *           a <code>true</code> to be returned.
    */
   public static boolean deferredLeave(String label)
   {
      return deferredLeaveWorker(null, label);
   }
   
   /**
    * Continue (iterate) the nearest enclosing block.
    * 
    * @throws   NextUnwindException 
    */
   public static void next()
   throws NextUnwindException
   {
      next("");
   }
   
   /**
    * Continue (iterate) the block associated with the given label.  This will
    * unwind the stack until the proper block has been reached.  At that
    * point the block will be iterated using the Java "continue" statement.
    *
    * @param    label
    *           The name of the block to be iterated.
    * @throws   NextUnwindException 
    */
   public static void next(String label)
   throws NextUnwindException
   {
      WorkArea wa = work.obtain();
      
      if (label != null)
      {
         wa.deferredNext = label;
      }
      
      throw wa.nextExc;
   }
   
   /**
    * Detect if there is a pending next for the block named by the given
    * label. If so, this deferred next flag will be cleared and 
    * <code>true</code> will be returned.  The caller is responsible for
    * ensuring that the next request is fully granted as no further state
    * will be available for subsequent processing.
    *
    * @param    label
    *           The name of the block to be checked against the deferred
    *           next block name.
    *
    * @return   <code>true</code> if there is a pending deferred next and
    *           the name of the block to be iterated is the same as the given
    *           label. If the pending deferred block name is the "" (empty
    *           string), that is the equivalent of matching "any block" and
    *           so the first call to this method after that is set will cause
    *           a <code>true</code> to be returned.
    */
   public static boolean deferredNext(String label)
   {
      return deferredNextWorker(null, label);
   }
   
   /**
    * Raise the STOP condition and cause control to be transferred to the
    * nearest enclosing block that handles the associated condition.
    * 
    * @throws   StopConditionException 
    */
   public static void stop()
   throws StopConditionException
   {
      throw new StopConditionException("Explicit STOP.");
   }
   
   /**
    * Raise the QUIT condition and cause control to be transferred to the
    * nearest enclosing block that handles the associated condition.
    * 
    * @throws  QuitConditionException 
    */
   public static void quit()
   throws QuitConditionException
   {
      throw new QuitConditionException("Explicit QUIT.");
   }
   
   /**
    * Return from the nearest enclosing top-level block. Sets the default
    * <code>RETURN-VALUE</code> to "" (empty string) for internal/external
    * procedures and triggers. Leaves <code>unknown value</code> as the
    * default return from a user-defined function.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void returnNormal()
   throws ReturnUnwindException
   {
      returnWorker(null, ReturnType.NORMAL);
   }
   
   /**
    * Return from the nearest enclosing top-level block. Sets the
    * <code>RETURN-VALUE</code> to the given <code>character</code> value for
    * internal/external procedures and triggers. Saves the given value as the
    * return value for a user-defined function.  If <code>val</code> is
    * <code>null</code>, then <code>RETURN-VALUE</code> will be set to the
    * "" (empty string) for internal/external procedures and triggers. If
    * <code>val</code> is <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    val
    *           The value to return or <code>null</code>.  For procedures
    *           or triggers, this must be a <code>character</code> value.
    *           
    * @throws   ReturnUnwindException 
    */
   public static void returnNormal(BaseDataType val)
   throws ReturnUnwindException
   {
      returnWorker(val, ReturnType.NORMAL);
   }
   
   /**
    * Return from the nearest enclosing top-level block. Sets the
    * <code>RETURN-VALUE</code> to the given <code>character</code> value for
    * internal/external procedures and triggers. Saves the given value as the
    * return value for a user-defined function.  If <code>val</code> is
    * <code>null</code>, then <code>RETURN-VALUE</code> will be set to the
    * "" (empty string) for internal/external procedures and triggers. If
    * <code>val</code> is <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    val
    *           The value to return or <code>null</code>.  For procedures
    *           or triggers, this must be a <code>character</code> value.
    *           
    * @throws   ReturnUnwindException 
    */
   public static void returnNormal(String val)
   throws ReturnUnwindException
   {
      returnWorker(new character(val), ReturnType.NORMAL);
   }
   
   /**
    * Return from the nearest enclosing top-level block. This should not be
    * used for internal/external procedures and triggers, since the
    * <code>RETURN-VALUE</code> can only be set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    val
    *           The value to return.
    *           
    * @throws   ReturnUnwindException 
    */
   public static void returnNormal(int val)
   throws ReturnUnwindException
   {
      returnWorker(new integer(val), ReturnType.NORMAL);
   }
   
   /**
    * Return from the nearest enclosing top-level block. This should not be
    * used for internal/external procedures and triggers, since the
    * <code>RETURN-VALUE</code> can only be set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    val
    *           The value to return.
    *           
    * @throws   ReturnUnwindException 
    */
   public static void returnNormal(double val)
   throws ReturnUnwindException
   {
      returnWorker(new decimal(val), ReturnType.NORMAL);
   }
   
   /**
    * Return from the nearest enclosing top-level block. This should not be
    * used for internal/external procedures and triggers, since the
    * <code>RETURN-VALUE</code> can only be set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    val
    *           The value to return.
    *           
    * @throws   ReturnUnwindException 
    */
   public static void returnNormal(boolean val)
   throws ReturnUnwindException
   {
      returnWorker(new logical(val), ReturnType.NORMAL);
   }
   
   /**
    * Return from the nearest enclosing top-level block. Sets the default
    * <code>RETURN-VALUE</code> to "" (empty string) for internal/external
    * procedures and triggers. Leaves <code>unknown value</code> as the
    * default return from a user-defined function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *           
    * @throws   ReturnUnwindException 
    */
   public static void returnConsume()
   throws ReturnUnwindException
   {
      returnWorker(null, ReturnType.CONSUME);
   }
   
   /**
    * Return from the nearest enclosing top-level block, consuming any pending
    * event processing. Sets the <code>RETURN-VALUE</code> to the given value
    * for internal/external procedures and triggers.  Should not be called
    * for a user-defined function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *
    * @param    val
    *           The value to return or <code>null</code>.
    *           
    * @throws   ReturnUnwindException 
    */
   public static void returnConsume(character val)
   throws ReturnUnwindException
   {
      returnWorker(val, ReturnType.CONSUME);
   }
   
   /**
    * Return from the nearest enclosing top-level block, consuming any pending
    * event processing. Sets the <code>RETURN-VALUE</code> to the given value
    * for internal/external procedures and triggers. Sets the given value as
    * the return value for a user-defined function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *
    * @param    val
    *           The value to return or <code>null</code>.
    *           
    * @throws   ReturnUnwindException 
    */
   public static void returnConsume(String val)
   throws ReturnUnwindException
   {
      returnWorker(new character(val), ReturnType.CONSUME);
   }
   
   /**
    * Return from the nearest enclosing top-level block. Sets the default
    * <code>RETURN-VALUE</code> to "" (empty string) for internal/external
    * procedures and triggers. Leaves <code>unknown value</code> as the
    * default return from a user-defined function.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *           
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void returnError()
   throws ReturnUnwindException,
          ErrorConditionException
   {
      returnWorker(null, ReturnType.ERROR);
   }
   
   /**
    * Return from the nearest enclosing top-level block. If the given value
    * is <code>null</code>, this sets the default <code>RETURN-VALUE</code>
    * to "" (empty string) for internal/external procedures and triggers.
    * It leaves <code>unknown value</code> as the default return from a
    * user-defined function.  If the value is not <code>null</code>, then
    * that value (if it is a <code>character</code> type) will be set as the
    * <code>RETURN-VALUE</code> for internal/external procedures and triggers.
    * If not <code>null</code>, the value will be returned as the function
    * return value for user-defined functions.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *
    * @param    val
    *           The value to return.
    *           
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void returnError(character val)
   throws ReturnUnwindException,
          ErrorConditionException
   {
      returnWorker(val, ReturnType.ERROR);
   }
   
   /**
    * Return from the nearest enclosing top-level block. If the given value
    * is <code>null</code>, this sets the default <code>RETURN-VALUE</code>
    * to "" (empty string) for internal/external procedures and triggers.
    * It leaves <code>unknown value</code> as the default return from a
    * user-defined function.  If the value is not <code>null</code>, then
    * that value will be set as the <code>RETURN-VALUE</code> for internal/
    * external procedures and triggers. If not <code>null</code>, the value
    * will be returned as the function return value for user-defined
    * functions.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *
    * @param    val
    *           The value to return.
    *           
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void returnError(String val)
   throws ReturnUnwindException,
          ErrorConditionException
   {
      returnWorker(new character(val), ReturnType.ERROR);
   }
   
   /**
    * Trigger a rollback of the current block and then leave that block.
    * 
    * @throws   LeaveUnwindException 
    */
   public static void undoLeave()
   throws LeaveUnwindException
   {
      undoLeave(null, "");
   }
   
   /**
    * Trigger a rollback of the given block and then leave that block.
    *
    * @param    label
    *           The name of the block that is the target of the undo and
    *           leave operations.
    * 
    * @throws   LeaveUnwindException 
    */
   public static void undoLeave(String label)
   throws LeaveUnwindException
   {
      undoLeave(label, label);
   }
   
   /**
    * Trigger a rollback of the given block and then leave that block or a
    * more enclosing block.
    * <p>
    * If the current block is a variant of a FOR block, then the buffers
    * associated with the block's query must not change the referenced 
    * record.
    *
    * @param    undo
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    leave
    *           The name of the block that is the target of the leave
    *           operation.
    * 
    * @throws   LeaveUnwindException 
    */
   public static void undoLeave(String undo, String leave)
   throws LeaveUnwindException
   {
      BlockType blk = getBlockType();
      
      boolean forBlock = blk == BlockType.FOR_BLOCK          ||
                         blk == BlockType.FOR_BLOCK_WHILE    ||
                         blk == BlockType.FOR_BLOCK_TO       ||
                         blk == BlockType.FOR_BLOCK_TO_WHILE ||
                         blk == BlockType.FOR_LOOP           ||
                         blk == BlockType.FOR_LOOP_WHILE     ||
                         blk == BlockType.FOR_LOOP_TO        ||
                         blk == BlockType.FOR_LOOP_TO_WHILE;
      
      try
      {
         if (forBlock)
         {
            processOffEndListeners(true);
         }

         rollback(undo);
         leave(leave);
      }
      finally
      {
         if (forBlock)
         {
            processOffEndListeners(false);
         }
      }
   }
   
   /**
    * Trigger a rollback of the current block and then iterate that block.
    * <p>
    * Infinite loop protection may convert the NEXT action into a LEAVE
    * depending on runtime state.
    * 
    * @throws   NextUnwindException 
    */
   public static void undoNext()
   throws NextUnwindException
   {
      undoNext(null, "");
   }
   
   /**
    * Trigger a rollback of the given block and then iterate that block.
    * <p>
    * Infinite loop protection may convert the NEXT action into a LEAVE
    * depending on runtime state.
    *
    * @param    label
    *           The name of the block that is the target of the undo and
    *           next operations.
    * 
    * @throws   NextUnwindException 
    */
   public static void undoNext(String label)
   throws NextUnwindException
   {
      undoNext(label, label);
   }
   
   /**
    * Trigger a rollback of the given block and then iterate that block or
    * a more enclosing block.
    * <p>
    * Infinite loop protection may convert the NEXT action into a LEAVE
    * depending on runtime state.
    *
    * @param    undo
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    next
    *           The name of the block that is the target of the next
    *           operation.
    * 
    * @throws   NextUnwindException 
    */
   public static void undoNext(String undo, String next)
   throws NextUnwindException
   {
      rollback(undo);
      
      if (honorNext(next))
      {
         next(next);
      }
      else
      {
         leave(next);
      }
   }
   
   /**
    * Trigger a rollback of the current block and then retry that block.
    * 
    * @throws   RetryUnwindException 
    */
   public static void undoRetry()
   throws RetryUnwindException
   {
      rollback();
      forceRetry(null);
   }
   
   /**
    * Trigger a rollback of the given block and then retry that block.
    *
    * @param    label
    *           The name of the block that is the target of the undo and
    *           retry operations.
    * 
    * @throws   RetryUnwindException 
    */
   public static void undoRetry(String label)
   throws RetryUnwindException
   {
      rollback(label);
      forceRetry(label);
   }
   
   /**
    * Trigger a rollback of the nearest enclosing top-level block and then
    * retry that block.
    * 
    * @throws   RetryUnwindException 
    */
   public static void undoRetryTopLevel()
   throws RetryUnwindException
   {
      rollbackTopLevel();
      forceRetry(null);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block. Sets the default <code>RETURN-VALUE</code> to
    * "" (empty string) for internal/external procedures and triggers. Leaves
    * <code>unknown value</code> as the default return from a user-defined
    * function.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal()
   throws ReturnUnwindException
   {
      rollback();
      returnWorker(null, ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback and then return from the nearest enclosing top-level
    * block. Sets the default <code>RETURN-VALUE</code> to "" (empty string)
    * for internal/external procedures and triggers. Leaves
    * <code>unknown value</code> as the default return from a user-defined
    * function.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormalTopLevel()
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(null, ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. Sets the default <code>RETURN-VALUE</code> to
    * "" (empty string) for internal/external procedures and triggers. Leaves
    * <code>unknown value</code> as the default return from a user-defined
    * function.
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(String label)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(null, ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block. Sets the <code>RETURN-VALUE</code> to the
    * given <code>character</code> value for internal/external procedures and
    * triggers. Saves the given value as the return value for a user-defined
    * function.  If <code>val</code> is <code>null</code>, then
    * <code>RETURN-VALUE</code> will be set to the "" (empty string) for
    * internal/external procedures and triggers. If <code>val</code> is
    * <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    val
    *           The value to return or <code>null</code>.  For procedures
    *           or triggers, this must be a <code>character</code> value.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(BaseDataType val)
   throws ReturnUnwindException
   {
      rollback();
      returnWorker(val, ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the nearest enclosing top-level block and then
    * return from that block. Sets the <code>RETURN-VALUE</code> to the
    * given <code>character</code> value for internal/external procedures and
    * triggers. Saves the given value as the return value for a user-defined
    * function.  If <code>val</code> is <code>null</code>, then
    * <code>RETURN-VALUE</code> will be set to the "" (empty string) for
    * internal/external procedures and triggers. If <code>val</code> is
    * <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    val
    *           The value to return or <code>null</code>.  For procedures
    *           or triggers, this must be a <code>character</code> value.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormalTopLevel(BaseDataType val)
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(val, ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. Sets the <code>RETURN-VALUE</code> to the
    * given <code>character</code> value for internal/external procedures and
    * triggers. Saves the given value as the return value for a user-defined
    * function.  If <code>val</code> is <code>null</code>, then
    * <code>RETURN-VALUE</code> will be set to the "" (empty string) for
    * internal/external procedures and triggers. If <code>val</code> is
    * <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return or <code>null</code>.  For procedures
    *           or triggers, this must be a <code>character</code> value.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(String label, BaseDataType val)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(val, ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. Sets the <code>RETURN-VALUE</code> to the
    * given <code>character</code> value for internal/external procedures and
    * triggers. Saves the given value as the return value for a user-defined
    * function.  If <code>val</code> is <code>null</code>, then
    * <code>RETURN-VALUE</code> will be set to the "" (empty string) for
    * internal/external procedures and triggers. If <code>val</code> is
    * <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return or <code>null</code>.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(String label, String val)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(new character(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the nearest enclosing top-level block and then
    * return from that block. Sets the <code>RETURN-VALUE</code> to the
    * given <code>character</code> value for internal/external procedures and
    * triggers. Saves the given value as the return value for a user-defined
    * function.  If <code>val</code> is <code>null</code>, then
    * <code>RETURN-VALUE</code> will be set to the "" (empty string) for
    * internal/external procedures and triggers. If <code>val</code> is
    * <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    val
    *           The value to return or <code>null</code>.  For procedures
    *           or triggers, this must be a <code>character</code> value.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormalTopLevel(String val)
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(new character(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block. This should not be used for internal/external
    * procedures and triggers, since the <code>RETURN-VALUE</code> can only be
    * set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(int val)
   throws ReturnUnwindException
   {
      rollback();
      returnWorker(new integer(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the nearest enclosing top-level block and then
    * return from that block. Sets the <code>RETURN-VALUE</code> to the
    * given <code>character</code> value for internal/external procedures and
    * triggers. Saves the given value as the return value for a user-defined
    * function.  If <code>val</code> is <code>null</code>, then
    * <code>RETURN-VALUE</code> will be set to the "" (empty string) for
    * internal/external procedures and triggers. If <code>val</code> is
    * <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormalTopLevel(int val)
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(new integer(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. This should not be used for internal/external
    * procedures and triggers, since the <code>RETURN-VALUE</code> can only be
    * set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(String label, int val)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(new integer(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block. This should not be used for internal/external
    * procedures and triggers, since the <code>RETURN-VALUE</code> can only be
    * set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(double val)
   throws ReturnUnwindException
   {
      rollback();
      returnWorker(new decimal(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the nearest enclosing top-level block and then
    * return from that block. Sets the <code>RETURN-VALUE</code> to the
    * given <code>character</code> value for internal/external procedures and
    * triggers. Saves the given value as the return value for a user-defined
    * function.  If <code>val</code> is <code>null</code>, then
    * <code>RETURN-VALUE</code> will be set to the "" (empty string) for
    * internal/external procedures and triggers. If <code>val</code> is
    * <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormalTopLevel(double val)
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(new decimal(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. This should not be used for internal/external
    * procedures and triggers, since the <code>RETURN-VALUE</code> can only be
    * set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(String label, double val)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(new decimal(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block. This should not be used for internal/external
    * procedures and triggers, since the <code>RETURN-VALUE</code> can only be
    * set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(boolean val)
   throws ReturnUnwindException
   {
      rollback();
      returnWorker(new logical(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the nearest enclosing top-level block and then
    * return from that block. Sets the <code>RETURN-VALUE</code> to the
    * given <code>character</code> value for internal/external procedures and
    * triggers. Saves the given value as the return value for a user-defined
    * function.  If <code>val</code> is <code>null</code>, then
    * <code>RETURN-VALUE</code> will be set to the "" (empty string) for
    * internal/external procedures and triggers. If <code>val</code> is
    * <code>null</code> for a user-defined function,
    * <code>unknown value</code> will be the default return.
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormalTopLevel(boolean val)
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(new logical(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. This should not be used for internal/external
    * procedures and triggers, since the <code>RETURN-VALUE</code> can only be
    * set to a character expression.
    * <p>
    * For a user-defined function, this saves the given value as the function
    * return value.
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnNormal(String label, boolean val)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(new logical(val), ReturnType.NORMAL);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block. Sets the default <code>RETURN-VALUE</code>
    * to "" (empty string) for internal/external procedures and triggers.
    * Leaves <code>unknown value</code> as the default return from a
    * user-defined function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnConsume()
   throws ReturnUnwindException
   {
      rollback();
      returnWorker(null, ReturnType.CONSUME);
   }
   
   /**
    * Trigger a rollback of and then return from the nearest
    * enclosing top-level block. Sets the default <code>RETURN-VALUE</code>
    * to "" (empty string) for internal/external procedures and triggers.
    * Leaves <code>unknown value</code> as the default return from a
    * user-defined function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnConsumeTopLevel()
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(null, ReturnType.CONSUME);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block, consuming any pending event processing. Sets
    * the <code>RETURN-VALUE</code> to the given value for internal/external
    * procedures and triggers. Should not be called for a user-defined
    * function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *
    * @param    val
    *           The value to return or <code>null</code>.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnConsume(character val)
   throws ReturnUnwindException
   {
      rollback();
      returnWorker(val, ReturnType.CONSUME);
   }
   
   /**
    * Trigger a rollback of and then return from the nearest
    * enclosing top-level block, consuming any pending event processing. Sets
    * the <code>RETURN-VALUE</code> to the given value for internal/external
    * procedures and triggers. Should not be called for a user-defined
    * function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *
    * @param    val
    *           The value to return or <code>null</code>.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnConsumeTopLevel(character val)
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(val, ReturnType.CONSUME);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. Sets the default <code>RETURN-VALUE</code>
    * to "" (empty string) for internal/external procedures and triggers.
    * Leaves <code>unknown value</code> as the default return from a
    * user-defined function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnConsume(String label)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(null, ReturnType.CONSUME);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block, consuming any pending event processing. Sets
    * the <code>RETURN-VALUE</code> to the given value for internal/external
    * procedures and triggers. Should not be called for a user-defined
    * function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return or <code>null</code>.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnConsume(String label, character val)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(val, ReturnType.CONSUME);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block, consuming any pending event processing. Sets
    * the <code>RETURN-VALUE</code> to the given value for internal/external
    * procedures and triggers. Should not be called for a user-defined
    * function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return or <code>null</code>.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnConsume(String label, String val)
   throws ReturnUnwindException
   {
      rollback(label);
      returnWorker(new character(val), ReturnType.CONSUME);
   }
   
   /**
    * Trigger a rollback of and then return from the nearest
    * enclosing top-level block, consuming any pending event processing. Sets
    * the <code>RETURN-VALUE</code> to the given value for internal/external
    * procedures and triggers. Should not be called for a user-defined
    * function.
    * <p>
    * If executed for a trigger, force the event which caused the trigger
    * execution to be consumed (no default processing of that event should
    * occur).
    *
    * @param    val
    *           The value to return or <code>null</code>.
    * 
    * @throws   ReturnUnwindException 
    */
   public static void undoReturnConsumeTopLevel(String val)
   throws ReturnUnwindException
   {
      rollbackTopLevel();
      returnWorker(new character(val), ReturnType.CONSUME);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block. Sets the default <code>RETURN-VALUE</code>
    * to "" (empty string) for internal/external procedures and triggers.
    * Leaves <code>unknown value</code> as the default return from a
    * user-defined function.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    * 
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void undoReturnError()
   throws ReturnUnwindException,
          ErrorConditionException
   {
      rollback();
      returnWorker(null, ReturnType.ERROR);
   }
   
   /**
    * Trigger a rollback of and then return from the nearest
    * enclosing top-level block. Sets the default <code>RETURN-VALUE</code>
    * to "" (empty string) for internal/external procedures and triggers.
    * Leaves <code>unknown value</code> as the default return from a
    * user-defined function.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    * 
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void undoReturnErrorTopLevel()
   throws ReturnUnwindException,
          ErrorConditionException
   {
      rollbackTopLevel();
      returnWorker(null, ReturnType.ERROR);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. Sets the default <code>RETURN-VALUE</code>
    * to "" (empty string) for internal/external procedures and triggers.
    * Leaves <code>unknown value</code> as the default return from a
    * user-defined function.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * 
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void undoReturnError(String label)
   throws ReturnUnwindException,
          ErrorConditionException
   {
      rollback(label);
      returnWorker(null, ReturnType.ERROR);
   }
   
   /**
    * Trigger a rollback of the current block and then return from the nearest
    * enclosing top-level block. If the given value is <code>null</code>,
    * this sets the default <code>RETURN-VALUE</code> to "" (empty string)
    * for internal/external procedures and triggers. It leaves
    * <code>unknown value</code> as the default return from a user-defined
    * function.  If the value is not <code>null</code>, then that value (if
    * it is a <code>character</code> type) will be set as the
    * <code>RETURN-VALUE</code> for internal/external procedures and triggers.
    * If not <code>null</code>, the value will be returned as the function
    * return value for user-defined functions.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void undoReturnError(character val)
   throws ReturnUnwindException,
          ErrorConditionException
   {
      rollback();
      returnWorker(val, ReturnType.ERROR);
   }
   
   /**
    * Trigger a rollback of and then return from the nearest
    * enclosing top-level block. If the given value is <code>null</code>,
    * this sets the default <code>RETURN-VALUE</code> to "" (empty string)
    * for internal/external procedures and triggers. It leaves
    * <code>unknown value</code> as the default return from a user-defined
    * function.  If the value is not <code>null</code>, then that value (if
    * it is a <code>character</code> type) will be set as the
    * <code>RETURN-VALUE</code> for internal/external procedures and triggers.
    * If not <code>null</code>, the value will be returned as the function
    * return value for user-defined functions.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void undoReturnErrorTopLevel(character val)
   throws ReturnUnwindException,
          ErrorConditionException
   {
      rollbackTopLevel();
      returnWorker(val, ReturnType.ERROR);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. If the given value is <code>null</code>,
    * this sets the default <code>RETURN-VALUE</code> to "" (empty string)
    * for internal/external procedures and triggers. It leaves
    * <code>unknown value</code> as the default return from a user-defined
    * function.  If the value is not <code>null</code>, then that value (if
    * it is a <code>character</code> type) will be set as the
    * <code>RETURN-VALUE</code> for internal/external procedures and triggers.
    * If not <code>null</code>, the value will be returned as the function
    * return value for user-defined functions.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void undoReturnError(String label, character val)
   throws ReturnUnwindException,
          ErrorConditionException
   {
      rollback(label);
      returnWorker(val, ReturnType.ERROR);
   }
   
   /**
    * Trigger a rollback of the given block and then return from the nearest
    * enclosing top-level block. If the given value is <code>null</code>,
    * this sets the default <code>RETURN-VALUE</code> to "" (empty string)
    * for internal/external procedures and triggers. It leaves
    * <code>unknown value</code> as the default return from a user-defined
    * function.  If the value is not <code>null</code>, then that value (if
    * it is a <code>character</code> type) will be set as the
    * <code>RETURN-VALUE</code> for internal/external procedures and triggers.
    * If not <code>null</code>, the value will be returned as the function
    * return value for user-defined functions.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *
    * @param    label
    *           The name of the block that is the target of the undo
    *           operation.
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void undoReturnError(String label, String val)
   throws ReturnUnwindException,
          ErrorConditionException
   {
      rollback(label);
      returnWorker(new character(val), ReturnType.ERROR);
   }
   
   /**
    * Trigger a rollback of and then return from the nearest
    * enclosing top-level block. If the given value is <code>null</code>,
    * this sets the default <code>RETURN-VALUE</code> to "" (empty string)
    * for internal/external procedures and triggers. It leaves
    * <code>unknown value</code> as the default return from a user-defined
    * function.  If the value is not <code>null</code>, then that value (if
    * it is a <code>character</code> type) will be set as the
    * <code>RETURN-VALUE</code> for internal/external procedures and triggers.
    * If not <code>null</code>, the value will be returned as the function
    * return value for user-defined functions.
    * <p>
    * If silent error mode is not enabled in the caller, an <code>ERROR</code>
    * condition will be raised here and will unwind the stack (it will not
    * be processed, but will instead it will be rethrown) until it reaches
    * the caller's context at which time the error will be honored. If silent
    * error mode is enabled, the error data will be stored in the 
    * <code>ERROR-STATUS</code> facility and a return will be executed to
    * return processing to the caller without raising the error condition
    * (without throwing an exception).
    *
    * @param    val
    *           The value to return.
    * 
    * @throws   ReturnUnwindException 
    * @throws   ErrorConditionException 
    */
   public static void undoReturnErrorTopLevel(String val)
   throws ReturnUnwindException,
          ErrorConditionException
   {
      rollbackTopLevel();
      returnWorker(new character(val), ReturnType.ERROR);
   }
   
   /**
    * Execute the given top-level block with the given transaction level.
    *
    * @param    level
    *           The transaction level for the given block.
    * @param    btype
    *           The type of block represented by the given code.
    * @param    block
    *           The code to execute.
    */
   static void topLevelBlock(TransactionType level,
                             BlockType       btype,
                             Block           block)
   {
      if (statsFile != null)
      {
         recordExecutionStats(btype, block);
      }
      
      try
      {
         OutputParameterAssigner opa =
            TransactionManager.deregisterOutputParameterAssigner();
         
         int     ttype = convertTransType(level, btype);
         boolean etype = (btype == BlockType.EXTERNAL_PROC);
         
         pushScope("methodScope",
                   ttype,
                   (PROP_ERROR | PROP_ENDKEY),
                   etype,
                   true,
                   false,
                   false,
                   btype);
         
         // add a scope to our list of function return values and block types
         WorkArea wa = work.obtain();
         wa.topType.push(btype);
         
         try
         {
            methodScope:
            {
               // give the block a chance to initialize its state and don't
               // repeat that during retry
               block.init();
               
               blockSetup();
               
               do
               {
                  try
                  {
                     processBody(block, null);
                  }
                  
                  catch (ErrorConditionException err)
                  {
                     notifyCondition(err);
                     honorError(err);
                     rollback("methodScope");
                     triggerRetry("methodScope");
                  }
                  
                  catch (EndConditionException end)
                  {
                     notifyCondition(end);
                     rollback("methodScope");
                     break methodScope;
                  }
                  
                  catch (RetryUnwindException rue)
                  {
                     honorRetry(rue);
                  }
                  
                  catch (StackUnwindException sue)
                  {
                     // we have unwound to the exit point, break out (it doesn't
                     // matter whether we need to exit because of a LEAVE, NEXT
                     // or a RETURN, they all have the same result on a top-level
                     // block
                     break methodScope;
                  }
                  
                  catch (Throwable thr)
                  {
                     abnormalEnd(thr);
                  }
               }
               while (needsRetry());
            }
         }
         
         catch (RuntimeException exc)
         {
            // cleaning up the OPA inside this catch only works because "normal"
            // control flow exceptions are caught and handled before they would
            // cross the top-level block boundary
            if (opa != null)
            {
               opa.abort();
            }
            
            throw exc;
         }
         
         finally
         {
            popScope();
            
            // cleanup our top level block state
            exitTopLevel(wa);
         }
         
         // process assign-backs to database fields from output parameters, if
         // any
         if (opa != null &&
             (btype == BlockType.EXTERNAL_PROC ||
              btype == BlockType.INTERNAL_PROC))
         {
            opa.processAssignments();
         }
      }
      finally
      {
         recordExecutionTime(btype, block);
      }
   }
   
   /**
    * Implements the body of all blocks or loops that are not related to the
    * <code>FOR</code> language statement. The main feature is to ensure that
    * any normal exit from the body will notify any <code>Commitable</code>
    * instances for validation purposes. A normal exit is a non-exception
    * return from {@link Block#body}, or certain exception cases related to
    * <code>LEAVE</code>, <code>NEXT</code>, <code>RETURN</code> or
    * <code>QUIT</code> when they are not associated with an <code>UNDO</code>.
    *
    * @param    block
    *           The block to execute.
    * @param    on                  
    *           List of user-defined on phrases.
    */
   private static void processBody(Block block, OnPhrase[] on)
   throws Throwable
   {
      Throwable reason = null;
      
      try
      {
         block.body();
      }
      
      catch (Throwable thr)
      {
         reason = thr;
      }
      
      finally
      {
         if (isTransaction() && (reason == null || isNormal(reason, on)))
         {
            processValidate();
         }
         
         if (reason != null)
         {
            throw reason;
         }
      }
   }
   
   /**
    * Implements the body of all <code>FOR</code> language statement blocks or
    * loops. The main feature is to ensure that any normal exit from the body
    * will notify any <code>Commitable</code> instances for validation
    * purposes. A normal exit is a non-exception return from {@link Block#body},
    * or certain exception cases related to <code>LEAVE</code>,
    * <code>NEXT</code>, <code>RETURN</code> or <code>QUIT</code> when they are
    * not associated with an <code>UNDO</code>.
    *
    * @param    block
    *           The block to execute.
    * @param    to                        
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on                  
    *           List of user-defined on phrases.
    *
    * @return   <code>true</code> if the code should break out of the block
    *           on return (because the loop/block control expressions or
    *           incrementer forced an exit).
    */
   private static boolean processForBody(Block             block,
                                         ToClause          to,
                                         LogicalExpression expr,
                                         OnPhrase[]        on)
   throws Throwable
   {
      Throwable reason     = null;
      boolean   needsBreak = false;   
      
      try
      {
         // retrieve the record (only used in cases where there is
         // a TO and/or WHILE)
         if (to != null || expr != null)
            block.enter();
         
         // evaluate any TO or WHILE condition and execute the block
         // body if both are null or when all non-null expressions
         // evaluate true; otherwise never execute the block body
         if (needsIterate(to, expr))
         {
            block.body();
         }
         else
         {
            needsBreak = true;
         }
      }
      
      catch (Throwable thr)
      {
         reason = thr;
      }
      
      finally
      {
         if (isTransaction() && (reason == null || isNormal(reason, on)))
         {
            processValidate();
         }
         
         if (reason != null)
         {
            throw reason;
         }
      }
      
      return needsBreak;
   }
   
   /**
    * Detect if an exception represents a normal or abnormal condition.
    *
    * @param    thr
    *           The exception to analyze.
    * @param    on                  
    *           List of user-defined on phrases.
    *
    * @return   <code>true</code> if the exception is a normal control-flow
    *           condition.
    */
   private static boolean isNormal(Throwable thr, OnPhrase[] on)
   {
      // no exception means normal
      if (thr == null)
      {
         return true;
      }
      
      // LEAVE, NEXT and RETURN
      if (thr instanceof StackUnwindException)
      {
         // check if an UNDO is NOT in effect
         if (!hadRollback())
         {
            return true;
         }
      }
      
      // QUIT
      if (thr instanceof QuitConditionException)
      {
         // if there are no user-defined ON phrases, then a QUIT is always
         // a commit instead of a rollback
         if (on == null)
         {
            return true;
         }
         
         // look for a user-defined ON QUIT phrase
         OnPhrase imple = matchCondition(on, null, Condition.QUIT);
         
         // if none, then QUIT will be a commit not a rollback
         if (imple == null)
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Detect if there is a pending leave for the block named by the given
    * label. If so, this deferred leave flag will be cleared and 
    * <code>true</code> will be returned.  The caller is responsible for
    * ensuring that the leave request is fully granted as no further state
    * will be available for subsequent processing.
    *
    * @param    wa
    *           Local context.
    * @param    label
    *           The name of the block to be checked against the deferred
    *           leave block name.
    *
    * @return   <code>true</code> if there is a pending deferred leave and
    *           the name of the block to be left is the same as the given
    *           label. If the pending deferred block name is the "" (empty
    *           string), that is the equivalent of matching "any block" and
    *           so the first call to this method after that is set will cause
    *           a <code>true</code> to be returned.
    */
   private static boolean deferredLeaveWorker(WorkArea wa, String label)
   {
      if (wa == null)
         wa = work.obtain();
      
      if (wa.deferredLeave != null && label != null)
      {
         if (label.equals(wa.deferredLeave) || wa.deferredLeave.length() == 0)
         {
            // clear state
            wa.deferredLeave = null;
            
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Detect if there is a pending next for the block named by the given
    * label. If so, this deferred next flag will be cleared and 
    * <code>true</code> will be returned.  The caller is responsible for
    * ensuring that the next request is fully granted as no further state
    * will be available for subsequent processing.
    *
    * @param    wa
    *           Local context.
    * @param    label
    *           The name of the block to be checked against the deferred
    *           next block name.
    *
    * @return   <code>true</code> if there is a pending deferred next and
    *           the name of the block to be iterated is the same as the given
    *           label. If the pending deferred block name is the "" (empty
    *           string), that is the equivalent of matching "any block" and
    *           so the first call to this method after that is set will cause
    *           a <code>true</code> to be returned.
    */
   private static boolean deferredNextWorker(WorkArea wa, String label)
   {
      if (wa == null)
         wa = work.obtain();
      
      if (wa.deferredNext != null && label != null)
      {
         if (label.equals(wa.deferredNext) || wa.deferredNext.length() == 0)
         {
            // clear state
            wa.deferredNext = null;
            
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Return from the nearest enclosing top-level block. This will process
    * differently based on the type of the top-level block (internal/external
    * procedures, triggers or user-defined functions), the type of return
    * processing that is requested (see {@link ReturnType}) and the presence
    * or absence of a value to return.
    * <p>
    * The following behavior is provided:
    * <p>
    * <pre>
    * Codes   Meaning                
    * ------  -----------------------------------------------------
    * C       Invoke LogicalTerminal.consume()
    * E       Invoke TransactionManager.triggerErrorInCaller()
    * V       Invoke ControlFlowOps.setReturnValue(val)
    * N       Invoke ControlFlowOps.setReturnValue("")
    * F       Invoke setFuncReturn(val)
    * U       Invoke setFuncReturn(null) (set return value as unknown)
    * I       Invalid (compile error in Progress)
    *
    *  Return    Return    Internal  External  Trigger  Function
    *  Value      Type       Proc      Proc
    * --------  ---------  --------  --------  -------  --------
    * null      NORMAL     N         N         N        U
    * non-null  NORMAL     V         V         V        F
    * null      CONSUME    N         N         NC       U
    * non-null  CONSUME    V         V         VC       F
    * null      ERROR      NE        NE        I        U
    * non-null  ERROR      VE        VE        I        U
    * </pre>
    * <p>
    * In all cases, a {@link ReturnUnwindException} will be thrown and this
    * will cause the stack to unwind to the nearest enclosing top-level
    * block. That block will then handle the return in a block specific
    * manner.  For example, in a user-defined function, the given value or 
    * by default the <code>unknown value</code> will be returned to the
    * caller. Procedures and triggers have void return values.
    *
    * @param    val
    *           The value to return.
    * @param    rtype
    *           The type of return processing.
    */
   private static void returnWorker(BaseDataType val, ReturnType rtype)
   throws ReturnUnwindException
   {
      WorkArea  wa      = work.obtain();
      BlockType nearest = topLevelType(wa);
      boolean   inFunc  = false;
         
      if (nearest != null)
      {
         switch (nearest)
         {
            case TRIGGER:
               if (rtype == ReturnType.CONSUME)
               {
                  LogicalTerminal.consumeEvent();
               }
               else if (rtype == ReturnType.ERROR)
               {
                  String err = "RETURN ERROR is not allowed in a TRIGGER";
                  throw new IllegalArgumentException(err);
               }
               // fall through so that return-value is also set 
            case INTERNAL_PROC:
            case EXTERNAL_PROC:
               if (val == null)
               {
                  ControlFlowOps.setReturnValue("");
               }
               else
               {
                  // the return value should always be of type character
                  // but this code just makes it safer
                  if (val instanceof character)
                  {
                     character retVal = (character) val;
                     ControlFlowOps.setReturnValue(retVal);
                  }
               }
               break;
            case FUNCTION:
               // in the error case, even if a value is passed, it will be
               // forced to null; in the non-error cases if the value is null
               // then the result will be setting the return to unknown value
               // and if the value is not null, then that value will be set as
               // the return value
               setFuncReturn((rtype == ReturnType.ERROR) ? null : val);
               inFunc = true;
               break;
         }
      }
      
      // functions don't raise error in the calling scope 
      if (rtype == ReturnType.ERROR && !inFunc)
      {
         triggerErrorInCaller();
      }
      
      throw wa.retExc;
   }
   
   /**
    * Execute the given function block with the given transaction level.
    *
    * @param    level
    *           The transaction level for the given block.
    * @param    block
    *           The code to execute.
    * @param    retType
    *           The type of the return value which is used in the case the
    *           <code>unknown value</code> must be returned.
    *
    * @return   The return value from the block (or its nested blocks) which
    *           is the <code>unknown value</code> by default or on error.
    */
   private static BaseDataType functionBlock(TransactionType level,
                                             Block           block,
                                             Class<?>        retType)
   {
      if (statsFile != null)
      {
         recordExecutionStats(BlockType.FUNCTION, block);
      }
      
      try
      {
         OutputParameterAssigner opa =
            TransactionManager.deregisterOutputParameterAssigner();
   
         BaseDataType retVal = null;
         
         int ttype = convertTransType(level, BlockType.FUNCTION);
         pushScope("methodScope",
                   ttype,
                   (PROP_ERROR | PROP_ENDKEY),
                   false,
                   true,
                   false,
                   false,
                   BlockType.FUNCTION);
         
         // add a scope to our list of function return values and block types
         WorkArea wa = work.obtain();
         wa.funcRet.push(null);
         wa.topType.push(BlockType.FUNCTION);
         
         try
         {
            methodScope:
            {
               // give the block a chance to initialize its state and don't
               // repeat that during retry
               block.init();
               
               blockSetup();
               
               do
               {
                  try
                  {
                     processBody(block, null);
                  }
                  
                  catch (ErrorConditionException err)
                  {
                     // we deliberately do not test honorError() here since
                     // functions in Progress do not implement the "trigger
                     // error in caller" behavior on a return error
                     notifyCondition(err);
                     rollback("methodScope");
                     triggerRetry("methodScope");
                  }
                  
                  catch (EndConditionException end)
                  {
                     notifyCondition(end);
                     rollback("methodScope");
                     break methodScope;
                  }
                  
                  catch (RetryUnwindException rue)
                  {
                     honorRetry(rue);
                  }
                  
                  catch (StackUnwindException sue)
                  {
                     // we have unwound to the exit point, break out (it doesn't
                     // matter whether we need to exit because of a LEAVE, NEXT or
                     // a RETURN, they all have the same result on a function block
                     break methodScope;
                  }
                  
                  catch (Throwable thr)
                  {
                     abnormalEnd(thr);
                  }
               }
               while (needsRetry());
            }
         }
         
         catch (RuntimeException exc)
         {
            // cleaning up the OPA inside this catch only works because "normal"
            // control flow exceptions are caught and handled before they would
            // cross the top-level block boundary
            if (opa != null)
            {
               opa.abort();
            }
            
            throw exc;
         }
         
         finally
         {
            popScope();
            
            // cleanup return value stack and obtain the current value
            retVal = wa.funcRet.pop();
            
            // cleanup our top level block state
            exitTopLevel(wa);
         }
         
         // process assign-backs to database fields from output parameters, if
         // any
         if (opa != null)
         {
            opa.processAssignments();
         }
         
         return (retVal == null) ? BaseDataType.generateUnknown(retType)
                                 : retVal;
      }
      finally
      {
         recordExecutionTime(BlockType.FUNCTION, block);
      }
   }
   
   /**
    * Execute the given editing block for the given frame and widget list
    * with <code>UPDATE</code> semanics. All looping, sub-transaction
    * properties and block properties will be honored.  The exit from the
    * editing block will be based on the given <code>GO-ON</code> event list.
    *
    * @param    frame
    *           The frame for which the editing block is associated.
    * @param    widgets
    *           The widgets in the given frame that are to be enabled for
    *           editing.   
    * @param    goOn       
    *           The exit conditions for this loop or <code>null</code> if
    *           there are no explicit exit conditions.
    * @param    enclosing
    *           The list of enclosing blocks of which this class is unaware.
    *           Any LEAVE or NEXT that references one of those enclosing
    *           blocks must not unwind the stack further using an exception,
    *           but rather it must return and let the caller check to see if
    *           a deferred LEAVE or NEXT must be honored.
    * @param    label
    *           The label of the block for use in control flow statements
    *           such as <code>LEAVE</code> or <code>NEXT</code>.
    * @param    block
    *           The code for the editing block body.
    * @param    etype
    *           Editing block semantics (<code>CommonFrame.PROMPT-FOR</code>,
    *           <code>CommonFrame.SET</code> or
    *           <code>CommonFrame.UPDATE</code>).
    */
   private static void editingBlock(CommonFrame    frame,
                                    FrameElement[] widgets,
                                    EventList      goOn,
                                    String[]       enclosing,
                                    String         label,
                                    Block          block,
                                    int            etype)
   {
      pushScope(label,
                convertTransType(TransactionType.SUB, BlockType.EDITING),
                PROP_ERROR,
                false,
                false,
                true,
                true,
                BlockType.EDITING);
      
      try
      {
         // give the block a chance to initialize its state and don't
         // repeat that during retry
         block.init();
         
         // prepare the frame and the client for the editing block, this
         // must only occur once (outside of the loop body and not being
         // re-executed on retry)
         if (goOn == null)
         {
            frame.startEditingMode(widgets, etype);
         }
         else
         {
            frame.startEditingMode(widgets, etype, goOn);
         }
         
         editing:
         do
         {
            blockSetup();
            
            do
            {
               try
               {
                  processBody(block, null);
               }
               
               catch (ErrorConditionException err)
               {
                  notifyCondition(err);
                  
                  // EDITING blocks propagate ERROR conditions to the
                  // containing block, in 2 cases:
                  // 1. PROMPT-FOR or SET semantics
                  // 2. No matter the type of editing block (even if it is
                  //    an UPDATE), if the condition was raised as a result
                  //    of the END-ERROR key AND the editing block's "local"
                  //    interaction number < 2
                  if (etype != CommonFrame.UPDATE ||
                      (err.isEndError() && (getInteractions() < 2)))
                  {
                     throw err;
                  }

                  // If the EDITING semantic is for an UPDATE type block AND
                  // the ERROR condition was NOT raised as a result of an
                  // END-ERROR key press if "local" interaction number < 2,
                  // then AND ONLY THEN we will act as if there was an UNDO,
                  // RETRY of the EDITING block.
                  honorError(err);
                  rollback(label);
                  triggerRetry(null);
               }
               
               catch (RetryUnwindException rue)
               {
                  honorRetry(rue);
               }
               
               catch (LeaveUnwindException lue)
               {
                  if (testLeave(enclosing, label) == Disposition.BREAK)
                  {
                     // the LEAVE targetted this block or an enclosing
                     // non-managed block
                     break editing;
                  }
                  
                  // the LEAVE must be referencing an enclosing block which IS
                  // known to the block manager OR to one of the other
                  // enclosing managed blocks, so keep unwinding the stack
                  throw lue;
               }
               
               catch (NextUnwindException nue)
               {
                  switch (testNext(enclosing, label))
                  {
                     case BREAK:
                        // a NEXT targetted at a directly enclosing block that
                        // is not managed
                        break editing;
                     case CONTINUE:
                        // a NEXT targetted at this block
                        continue editing;
                     default:
                        // the NEXT must be referencing an enclosing block
                        // which IS known to the block manager OR to one of
                        // the other enclosing managed blocks, so keep
                        // unwinding the stack
                        throw nue;
                  }
               }
               
               catch (ReturnUnwindException ret)
               {
                  // this isn't a top level block, keep unwinding without
                  // treating this as an abend
                  throw ret;
               }
               
               catch (Throwable thr)
               {
                  abnormalEnd(thr);
               }
            }
            while (needsRetry());
            
            if (isBreakPending())
            {
               break editing;
            }
         }
         while (frame.continueEditing(widgets));
      }
      
      finally
      {
         popScope();
      }
   }
   
   /**                   
    * Execute a simple DO block (not a loop) with the given configuration
    * and no default properties.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    on                  
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   private static void doBlockWorker(TransactionType lvl,
                                     String[]        enclosing,
                                     String          label,
                                     OnPhrase[]      on,
                                     Block           block)
   {
      int ttype = convertTransType(lvl, BlockType.DO);
      
      OnPhrase[] defOn = null;
      
      if (lvl == TransactionType.FULL)
      {
         defOn = new OnPhrase[1];
         defOn[0] = new OnPhrase(Condition.ERROR, Action.RETRY, label);
      }
      
      int props = decodeConditions(defOn, on);
      
      pushScope(label,
                ttype,
                props,
                false,
                false,
                false,
                false,
                BlockType.DO);
      
      try
      {
         try
         {
            // each constructed query during block initialization will be
            // collected
            startOffEndRegistration();
            
            // give the block a chance to initialize its state and don't
            // repeat that during retry
            block.init();
         }
         finally
         {
            stopOffEndRegistration();

            // after all queries were created and registered, they must be
            // reset so the buffers reference no record.
            Set<QueryOffEndListener> listeners = getOffEndListeners(true);
            
            for (QueryOffEndListener qoel : listeners)
            {
               qoel.reset();
            }
         }
         
         blockSetup();
         
         blockLabel:
         do
         {
            try
            {
               processBody(block, on);
            }
            
            catch (ConditionException ce)
            {
               notifyCondition(ce);
               
               switch (processCondition(on, defOn, ce, enclosing, label))
               {
                  // note that this is not an iterating block so a leave or
                  // next targeted at this block must always act the same
                  case BREAK:
                  case CONTINUE:
                     break blockLabel;
               }
            }
            
            catch (RetryUnwindException rue)
            {
               honorRetry(rue);
            }
            
            catch (LeaveUnwindException lue)
            {
               if (testLeave(enclosing, label) == Disposition.BREAK)
               {
                  // the LEAVE targeted this block or an enclosing
                  // non-managed block
                  break blockLabel;
               }
               
               // the LEAVE must be referencing an enclosing block which IS
               // known to the block manager OR to one of the other enclosing
               // managed blocks, so keep unwinding the stack
               throw lue;
            }
            
            catch (NextUnwindException nue)
            {
               if (testNext(enclosing, label) == Disposition.BREAK)
               {
                  // the NEXT can't be targeted at this block BUT it can be
                  // targeted at an enclosing non-managed block
                  break blockLabel;
               }
               
               // the NEXT must be referencing an enclosing block which IS
               // known to the block manager OR to one of the other enclosing
               // managed blocks, so keep unwinding the stack
               throw nue;
            }
            
            catch (ReturnUnwindException rue)
            {
               // RETURN can't be targeted at a DO block so keep unwinding
               throw rue;
            }
            
            catch (Throwable thr)
            {
               abnormalEnd(thr);
            }
         }
         while (needsRetry());
      }
      
      finally
      {
         popScope();
      }
   }
   
   /**                   
    * Execute a DO WHILE expression, DO TO var = val1 to val2 by val3 or DO
    * TO WHILE expression loop with the given configuration and no default
    * properties.  At least one of the to clause or while expression must be
    * passed and it is valid to have both passed in which case both of the
    * conditions are tested with a logical AND.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   private static void doLoopWorker(TransactionType   lvl,
                                    String[]          enclosing,
                                    String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      BlockType btype = null;
      boolean   next  = false;
      
      if (to == null && expr != null)
      {
         btype = BlockType.DO_WHILE;
      }
      else if (to != null && expr == null)
      {
         btype = BlockType.DO_TO;
         next  = true;
      }
      else if (to != null && expr != null)
      {
         btype = BlockType.DO_TO_WHILE;
         next  = true;
      }
      else
      {
         String err = "DO loop MUST have a TO or WHILE clause.";
         throw new IllegalStateException(err);
      }
      
      OnPhrase[] defOn = null;
      
      if (lvl == TransactionType.FULL)
      {
         defOn = new OnPhrase[1];
         defOn[0] = new OnPhrase(Condition.ERROR, Action.RETRY, label);
      }
      
      // process the loop
      coreLoop(btype,
               next,
               lvl,
               enclosing,
               label,
               to,
               expr,
               on,
               defOn,
               block);
   }
   
   /**                   
    * Execute a REPEAT, REPEAT WHILE expression, REPEAT TO var = val1 to val2
    * by val3 or REPEAT TO WHILE expression loop with the given configuration
    * and with default properties.  If neither the to clause or while
    * expression is specified, this is an infinite loop. If both are specified
    * then both of the conditions are tested with a logical AND. By default,
    * the ON ENDKEY UNDO, LEAVE and ON ERROR UNDO, RETRY properties are
    * supported.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   private static void repeatWorker(TransactionType   lvl,
                                    String[]          enclosing,
                                    String            label,
                                    ToClause          to,
                                    LogicalExpression expr,
                                    OnPhrase[]        on,
                                    Block             block)
   {
      BlockType btype = null;
      boolean   next  = false;
      
      if (to == null && expr == null)
      {
         btype = BlockType.REPEAT;
      }
      else if (to == null && expr != null)
      {
         btype = BlockType.REPEAT_WHILE;
      }
      else if (to != null && expr == null)
      {
         btype = BlockType.REPEAT_TO;
         next  = true;
      }
      else if (to != null && expr != null)
      {
         btype = BlockType.REPEAT_TO_WHILE;
         next  = true;
      }
      
      OnPhrase[] defOn = new OnPhrase[2];
      defOn[0] = new OnPhrase(Condition.ENDKEY, Action.LEAVE, label);
      defOn[1] = new OnPhrase(Condition.ERROR, Action.RETRY, label);
      
      // process the loop
      coreLoop(btype,
               next,
               lvl,
               enclosing,
               label,
               to,
               expr,
               on,
               defOn,
               block);
   }
      
   /**                   
    * Execute a DO or REPEAT loop with all possible loop constructs, the
    * given configuration and with default properties.
    *
    * @param    btype
    *           Current block type.
    * @param    next
    *           <code>true</code> if on infinite loop protection a retry
    *           should be converted to a NEXT, <code>false</code> to convert
    *           to a LEAVE.
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    * @param    on
    *           List of user-defined on phrases.
    * @param    defOn
    *           List of default properties for the block. Use
    *           <code>null</code> to indicate no default properties for the
    *           block.
    * @param    block
    *           The code to execute.
    */
   private static void coreLoop(BlockType         btype,
                                boolean           next,
                                TransactionType   lvl,
                                String[]          enclosing,
                                String            label,
                                ToClause          to,
                                LogicalExpression expr,
                                OnPhrase[]        on,
                                OnPhrase[]        defOn,
                                Block             block)
   {
      int ttype = convertTransType(lvl, btype);
      int props = decodeConditions(defOn, on);
      
      pushScope(label, ttype, props, false, false, true, next, btype);
      
      try
      {
         try
         {
            // each constructed query during block initialization will be
            // collected
            startOffEndRegistration();
            
            // give the block a chance to initialize its state and don't
            // repeat that during iteration or retry
            block.init();
         }
         finally
         {
            stopOffEndRegistration();

            // after all queries were created and registered, they must be
            // reset so the buffers reference no record.
            Set<QueryOffEndListener> listeners = getOffEndListeners(true);
            
            for (QueryOffEndListener qoel : listeners)
            {
               qoel.reset();
            }
         }
         
         loop:     
         while (needsIterate(to, expr))
         {
            blockSetup();
         
            do
            {
               try
               {
                  processBody(block, on);
               }
               
               catch (ConditionException ce)
               {
                  notifyCondition(ce);
                  
                  switch (processCondition(on, defOn, ce, enclosing, label))
                  {
                     case BREAK:
                        break loop;
                     case CONTINUE:
                        continue loop;
                  }
               }
               
               catch (RetryUnwindException rue)
               {
                  honorRetry(rue);
               }
               
               catch (LeaveUnwindException lue)
               {
                  if (testLeave(enclosing, label) == Disposition.BREAK)
                  {
                     // the LEAVE targeted this block or an enclosing
                     // non-managed block
                     break loop;
                  }
                  
                  // the LEAVE must be referencing an enclosing block which IS
                  // known to the block manager OR to one of the other
                  // enclosing managed blocks, so keep unwinding the stack
                  throw lue;
               }
               
               catch (NextUnwindException nue)
               {
                  switch (testNext(enclosing, label))
                  {
                     case BREAK:
                        // the NEXT is targetted at an enclosing non-managed
                        // block
                        break loop;
                        
                     case CONTINUE:
                        // the NEXT is targetted at the current loop
                        continue loop;
                  }
                  
                  // the NEXT must be referencing an enclosing block which IS
                  // known to the block manager OR to one of the other
                  // enclosing managed blocks, so keep unwinding the stack
                  throw nue;
               }
               
               catch (ReturnUnwindException rue)
               {
                  // RETURN can't be targetted at a DO loop so keep
                  // unwinding
                  throw rue;
               }
               
               catch (Throwable thr)
               {
                  abnormalEnd(thr);
               }
            }
            while (needsRetry());
            
            if (isBreakPending())
            {
               break loop;
            }
         }
      }
      
      finally
      {
         popScope();
      }
   }
   
   /**                   
    * Execute a FOR FIRST/LAST WHILE block, FOR FIRST/LAST TO block
    * or FOR FIRST/LAST TO WHILE block with the given configuration and
    * with default properties.  Any TO or WHILE clause is treated as an
    * additional condition that must be met before the block body is executed.
    * Note that the presence of a TO or WHILE does not make this a loop. If
    * neither the TO clause or WHILE expression is specified, this is a
    * simple block. If both are specified then both of the conditions are
    * tested with a logical AND. By default, the ON ENDKEY UNDO, LEAVE and ON
    * ERROR UNDO, RETRY properties are supported.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself or if there
    * is a TO or WHILE clause then the query processing should be located in
    * the {@link Block#enter} method.  This means that the caller is
    * responsible for all query processing. The record retrieval must be done
    * on the first line of code in the block or in the <code>enter</code>
    * method to duplicate the Progress behavior. Note that this is required
    * to handle all error cases and 4GL conditions properly.
    * <p>
    * The evaluation of the TO or WHILE block termination conditions is done
    * AFTER the record retrieval.  This causes any found record to remain
    * available if the block is never entered because of the TO or WHILE
    * conditions.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to                        
    *           The TO clause that defines the block condition. The block
    *           control variable is initialized before entering the block. 
    *           Before the block body is entered, the control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at that moment). if the termination condition is
    *           met, the block is never entered.
    * @param    expr      
    *           The WHILE clause that defines a block condition. This
    *           expression is evaluated before the block body is about to
    *           be entered.  The block body will be entered unless the
    *           expression evaluates to <code>false</code>.
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   private static void forBlockWorker(TransactionType   lvl,
                                      String[]          enclosing,
                                      String            label,
                                      ToClause          to,
                                      LogicalExpression expr,
                                      OnPhrase[]        on,
                                      Block             block)
   {
      BlockType btype = null;
      
      if (to == null && expr == null)
      {
         btype = BlockType.FOR_BLOCK;
      }
      else if (to == null && expr != null)
      {
         btype = BlockType.FOR_BLOCK_WHILE;
      }
      else if (to != null && expr == null)
      {
         btype = BlockType.FOR_BLOCK_TO;
      }
      else if (to != null && expr != null)
      {
         btype = BlockType.FOR_BLOCK_TO_WHILE;
      }
      
      OnPhrase[] defOn = new OnPhrase[2];
      defOn[0] = new OnPhrase(Condition.ENDKEY, Action.LEAVE, label);
      defOn[1] = new OnPhrase(Condition.ERROR, Action.RETRY, label);
      
      int ttype = convertTransType(lvl, btype);
      int props = decodeConditions(defOn, on);

      pushScope(label, ttype, props, false, false, false, false, btype);
      
      try
      {
         try
         {
            // each constructed query during block initialization will be
            // collected
            startOffEndRegistration();
            
            // give the block a chance to initialize its state and don't
            // repeat that during retry
            block.init();
         }
         finally
         {
            // after block initialization is finished, the off end listeners
            // for all the buffers in the collected queries are saved
            stopOffEndRegistration();
         }

         blockSetup();

         blockLabel:
         do
         {
            try
            {
               if (processForBody(block, to, expr, on))
               {
                  break blockLabel;
               }
            }
            
            catch (ConditionException ce)
            {
               notifyCondition(ce);
               
               // when there are no more records in the query, we ALWAYS
               // UNDO, LEAVE (which is the same as the default ENDKEY
               // processing) the current block even if the user has defined
               // an explicit behavior for ENDKEY
               if (ce instanceof QueryOffEndException)
               {
                  QueryOffEndException qoee = (QueryOffEndException) ce;
                  QueryOffEndListener[] listeners = qoee.getListeners();

                  // flag indicating if the off-end listeners are the one 
                  // registered during block initialization
                  boolean allOffEndListeners = true;
                  for (int i = 0; i < listeners.length; i++)
                  {
                     allOffEndListeners = allOffEndListeners && 
                                          isOffEndListener(listeners[i]);
                  }

                  if (allOffEndListeners)
                  {
                     try
                     {
                        // if the buffers are the ones used by the FOR block,
                        // process them and trigger an UNDO, LEAVE
                        processOffEndListeners(true);

                        rollback(label);
                        break blockLabel;
                     }
                     finally
                     {
                        processOffEndListeners(false);
                     }
                  }
               }

               try
               {
                  // if the exception was triggered by something else, then
                  // the off-end interested buffers must remain on the 
                  // currently referenced record
                  processOffEndListeners(true);

                  // normal condition processing
                  switch (processCondition(on, defOn, ce, enclosing, label))
                  {
                     // note that this is not an iterating block so a leave or
                     // next targeted at this block must always act the same
                     case BREAK:
                     case CONTINUE:
                        break blockLabel;
                  }
               }
               finally
               {
                  processOffEndListeners(false);
               }
            }
            
            catch (RetryUnwindException rue)
            {
               honorRetry(rue);
            }
            
            catch (LeaveUnwindException lue)
            {
               if (testLeave(enclosing, label) == Disposition.BREAK)
               {
                  // the LEAVE targeted this block or an enclosing
                  // non-managed block
                  break blockLabel;
               }
               
               // the LEAVE must be referencing an enclosing block which IS
               // known to the block manager OR to one of the other enclosing
               // managed blocks, so keep unwinding the stack
               throw lue;
            }
            
            catch (NextUnwindException nue)
            {
               if (testNext(enclosing, label) == Disposition.BREAK)
               {
                  // the NEXT can't be targeted at this block BUT it can be
                  // targeted at an enclosing non-managed block
                  break blockLabel;
               }
               
               // the NEXT must be referencing an enclosing block which IS
               // known to the block manager OR to one of the other enclosing
               // managed blocks, so keep unwinding the stack
               throw nue;
            }
            
            catch (ReturnUnwindException rue)
            {
               // RETURN can't be targeted at a DO block so keep unwinding
               throw rue;
            }
            
            catch (Throwable thr)
            {
               abnormalEnd(thr);
            }
         }
         while (needsRetry());
      }
      
      finally
      {
         popScope();
      }
   }
   
   /**                   
    * Execute a FOR EACH, FOR EACH WHILE block, FOR EACH TO block
    * or FOR EACH TO WHILE block with the given configuration and
    * with default properties.  Any TO or WHILE clause is treated as an
    * additional condition that must be met before the loop body is executed.
    * If both are specified then both of the conditions are tested with a 
    * logical AND. By default, the ON ENDKEY UNDO, LEAVE and ON ERROR UNDO,
    * RETRY properties are supported.
    * <p>
    * The record retrieval that is associated with FOR blocks or loops must
    * be implemented directly inside the code of the block itself or if there
    * is a TO or WHILE clause then the query processing should be located in
    * the {@link Block#enter} method.  This means that the caller is
    * responsible for all query processing. The record retrieval must be done
    * on the first line of code in the block or in the <code>enter</code>
    * method to duplicate the Progress behavior. Note that this is required
    * to handle all error cases and 4GL conditions properly.
    * <p>
    * The evaluation of the TO or WHILE block termination conditions is done
    * AFTER the record retrieval.  This causes any found record to remain
    * available if the loop is never entered (or re-entered) because of the
    * TO or WHILE conditions.
    *
    * @param    lvl
    *           The transaction level to be honored. 
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. 
    * @param    on
    *           List of user-defined on phrases.
    * @param    block
    *           The code to execute.
    */
   private static void forEachWorker(TransactionType   lvl,
                                     String[]          enclosing,
                                     String            label,
                                     ToClause          to,
                                     LogicalExpression expr,
                                     OnPhrase[]        on,
                                     Block             block)
   {
      BlockType btype = null;
      
      if (to == null && expr == null)
      {
         btype = BlockType.FOR_LOOP;
      }
      else if (to == null && expr != null)
      {
         btype = BlockType.FOR_LOOP_WHILE;
      }
      else if (to != null && expr == null)
      {
         btype = BlockType.FOR_LOOP_TO;
      }
      else if (to != null && expr != null)
      {
         btype = BlockType.FOR_LOOP_TO_WHILE;
      }
      
      OnPhrase[] defOn = new OnPhrase[2];
      defOn[0] = new OnPhrase(Condition.ENDKEY, Action.LEAVE, label);
      defOn[1] = new OnPhrase(Condition.ERROR, Action.RETRY, label);
      
      int ttype = convertTransType(lvl, btype);
      int props = decodeConditions(defOn, on);

      pushScope(label, ttype, props, false, false, true, true, btype);
      
      try
      {
         try
         {
            // each constructed query during block initialization will be
            // collected
            startOffEndRegistration();
            
            // give the block a chance to initialize its state and don't
            // repeat that during retry
            block.init();
         }
         finally
         {
            // after block initialization is finished, the off end listeners
            // for all the buffers in the collected queries are saved
            stopOffEndRegistration();
         }
         
         loop:     
         while (true)
         {
            blockSetup();
            
            do
            {
               try
               {
                  if (processForBody(block, to, expr, on))
                  {
                     break loop;
                  }
               }
               
               catch (ConditionException ce)
               {
                  notifyCondition(ce);
                  
                  // when there are no more records in the query, we ALWAYS
                  // UNDO, LEAVE (which is the same as the default ENDKEY
                  // processing) the current block even if the user has
                  // defined an explicit behavior for ENDKEY
                  if (ce instanceof QueryOffEndException)
                  {
                     QueryOffEndException qoee = (QueryOffEndException) ce;
                     QueryOffEndListener[] listeners = qoee.getListeners();
   
                     // flag indicating if the off-end listeners are the one 
                     // registered during block initialization
                     boolean allOffEndListeners = true;
                     for (int i = 0; i < listeners.length; i++)
                     {
                        allOffEndListeners = allOffEndListeners && 
                                             isOffEndListener(listeners[i]);
                     }

                     if (allOffEndListeners)
                     {
                        try
                        {
                           // if the buffers are the ones used by the FOR 
                           // loop, process them and trigger an UNDO, LEAVE
                           processOffEndListeners(true);

                           rollback(label);
                           break loop;
                        }
                        finally
                        {
                           processOffEndListeners(false);
                        }
                     }
                  }
                  
                  try
                  {
                     // if the exception was triggered by something else, then
                     // the off-end interested buffers must remain on the 
                     // currently referenced record
                     processOffEndListeners(true);

                     // normal condition processing
                     switch (processCondition(on, defOn, ce, enclosing, label))
                     {
                        case BREAK:
                           break loop;
                        case CONTINUE:
                           continue loop;
                     }
                  }
                  finally
                  {
                     processOffEndListeners(false);
                  }
               }
               
               catch (RetryUnwindException rue)
               {
                  honorRetry(rue);
               }
               
               catch (LeaveUnwindException lue)
               {
                  if (testLeave(enclosing, label) == Disposition.BREAK)
                  {
                     // the LEAVE targetted this block or an enclosing
                     // non-managed block
                     break loop;
                  }
                  
                  // the LEAVE must be referencing an enclosing block which IS
                  // known to the block manager OR to one of the other
                  // enclosing managed blocks, so keep unwinding the stack
                  throw lue;
               }
               
               catch (NextUnwindException nue)
               {
                  switch (testNext(enclosing, label))
                  {
                     case BREAK:
                        // the NEXT is targetted at an enclosing non-managed
                        // block
                        break loop;
                        
                     case CONTINUE:
                        // the NEXT is targetted at the current loop
                        continue loop;
                  }
                  
                  // the NEXT must be referencing an enclosing block which IS
                  // known to the block manager OR to one of the other
                  // enclosing managed blocks, so keep unwinding the stack
                  throw nue;
               }
               
               catch (ReturnUnwindException rue)
               {
                  // RETURN can't be targetted at a DO block so keep unwinding
                  throw rue;
               }
               
               catch (Throwable thr)
               {
                  abnormalEnd(thr);
               }
            }
            while (needsRetry());
            
            if (isBreakPending())
            {
               break loop;
            }
         }
      }
      
      finally
      {
         popScope();
      }
   }
   
   /**
    * Test the given expressions to determine if the calling loop should
    * iterate or if the loop should exit.  Either or both parameters
    * may be <code>null</code> individually or simultaneously.
    *
    * @param    to
    *           The TO clause that defines a loop condition. The loop control
    *           variable is initialized on the first pass through and on
    *           subsequent passes it is incremented by a given factor. Before
    *           the loop body is entered, the loop control variable is
    *           compared to the termination value (which is dynamically
    *           resolved at each loop iteration). When the termination
    *           condition is met, the loop ends. If this is <code>null</code>,
    *           then a WHILE clause MUST be present to provide the condition
    *           for maintaining the loop.
    * @param    expr      
    *           The WHILE clause that defines a loop condition. This
    *           expression is evaluated every time the loop body is about to
    *           be entered.  The loop body will be entered unless the
    *           expression evaluates to <code>false</code>. If this is
    *           <code>null</code>, then a TO clause MUST be present to
    *           provide the condition for maintaining the loop.
    *
    * @return   <code>true</code> if all non-null parameters evaluate to
    *           <code>true</code> or if both parameters are <code>null</code>.
    */
   private static boolean needsIterate(ToClause to, LogicalExpression expr)
   {
      boolean toResult    = (to == null || to.incrementAndTest());
      boolean whileResult = (expr == null || expr.execute().booleanValue());
      
      return toResult && whileResult;
   }
   
   /**
    * Core worker to handle all the possible behavior in response to a
    * condition being raised in an inner block (DO, REPEAT or FOR). The
    * given condition exception will first be matched with the user-defined
    * on phases and then if no match is found, with the default on phrases
    * for the calling block. If nothing is found, then the condition
    * exception will be re-thrown. If a match is found then condition-specific
    * processing will be done to UNDO and then either unwind the stack using
    * exceptions or a disposition will be returned so that the caller can
    * properly implement a LEAVE or NEXT that is targetted at the current
    * block or at an enclosing non-managed block.
    * <p>
    * This honors infinite loop protection for converting NEXT to LEAVE
    * based on runtime state.
    *
    * @param    user
    *           List of user-defined on phrases.
    * @param    def
    *           List of default on phrases for the current block type.
    * @param    ce
    *           The condition being raised.
    * @param    enclosing
    *           The list of non-managed block names that enclose the current
    *           block.
    * @param    label
    *           The block name of the current block.
    *
    * @return   If an exception is not thrown (or re-thrown), then the caller
    *           will get a disposition code that should be honored to
    *           properly implement a LEAVE or NEXT action.
    */
   private static Disposition processCondition(OnPhrase[]         user,
                                               OnPhrase[]         def,
                                               ConditionException ce,
                                               String[]           enclosing,
                                               String             label)
   throws ConditionException,
          StackUnwindException,
          RetryUnwindException
   {
      Condition current = exceptionToCondition(ce);
      OnPhrase imple    = matchCondition(user, def, current);
      
      if (imple != null)
      {
         // ERROR processing has a special mode (RETURN ERROR) and STOP
         // processing mode which must unwind the stack to a specific
         // point, so we must check if we are in either of those modes now
         // (before the UNDO or secondary action for the block can be
         // executed)
         if (current == Condition.ERROR)
         {
            ErrorConditionException err = (ErrorConditionException) ce;
            
            // END-ERROR is propagated to the parent block in blocks that
            // have the ERROR property but do NOT have the ENDKEY property
            if (err.isEndError() && !hasEndkeyProp())
            {
               // detect the case in which a END-ERROR would have been
               // converted to ENDKEY, in such a case we must rethrow
               if (getInteractions() < 2)
               {
                  // let the parent see this (this will re-throw the
                  // exception after handling rollback state/processing)
                  abnormalEnd(err, false);
               }
            }
            
            // if we must unwind, this call will re-throw the exception,
            // otherwise it will return and we can continue with "normal"
            // exception processing
            honorError((ErrorConditionException) ce);
         }
         else if (current == Condition.STOP)
         {
            // if we must unwind, this call will re-throw the exception,
            // otherwise it will return and we can continue with "normal"
            // exception processing
            honorStop((StopConditionException) ce);
         }
         
         // first we UNDO at the specified label (which may be null meaning
         // we UNDO the current block)
         rollback(imple.undoLabel);
         
         Action target = imple.action;
         
         // honor ILP for NEXT
         if (target == Action.NEXT)
         {
            if (!honorNext(imple.actionLabel))
            {
               // convert to a LEAVE
               target = Action.LEAVE;
            }
         }
         
         WorkArea wa = work.obtain();
         
         // second we implement the specified action
         switch (target)
         {
            case LEAVE:
               if (imple.actionLabel == null        || 
                   imple.actionLabel.length() == 0  ||
                   imple.actionLabel.equals(label))
               {
                  return Disposition.BREAK;
               }
               else
               {
                  // at this point we must have a non-null and non-empty
                  // label that is the leave target
                  wa.deferredLeave = imple.actionLabel;
                  
                  if (matches(enclosing, imple.actionLabel))
                  {
                     // the target is a non-managed block that directly
                     // encloses the current block, break out and let the
                     // calling code check to see that this needs to be
                     // honored
                     return Disposition.BREAK;
                  }
                  else
                  {
                     // unwind the stack
                     throw wa.leaveExc;
                  }
               }
            case NEXT:
               if (imple.actionLabel == null        || 
                   imple.actionLabel.length() == 0  ||
                   imple.actionLabel.equals(label))
               {
                  return Disposition.CONTINUE;
               }
               else
               {
                  // at this point we must have a non-null and non-empty
                  // label that is the next target
                  wa.deferredNext = imple.actionLabel;
                  
                  if (matches(enclosing, imple.actionLabel))
                  {
                     // the target is a non-managed block that directly
                     // encloses the current block, break out and let the
                     // calling code check to see that this needs to be
                     // honored
                     return Disposition.BREAK;
                  }
                  else
                  {
                     // unwind the stack
                     throw wa.nextExc;
                  }
               }
            case RETRY:
               // if this retry is being triggered by an ENDKEY, remember
               // that fact so that the retry processing can be customized
               // (this enables the "stuttering" quirk and some custom
               // behavior with PAUSE
               if (current == Condition.ENDKEY)
               {
                  markEndkeyRetry();
               }
               
               // RETRY is always targetted at the same block as UNDO
               triggerRetry(imple.undoLabel);
               
               return Disposition.NONE;
            case RETURN_NORMAL:
               // this will throw an exception, it is OK because this method
               // will always be called in the context of an inner block and
               // WILL NEVER be called directly from a top-level block
               returnWorker(imple.value, ReturnType.NORMAL);
               break;
            case RETURN_CONSUME:
               // this will throw an exception, it is OK because this method
               // will always be called in the context of an inner block and
               // WILL NEVER be called directly from a top-level block
               returnWorker(imple.value, ReturnType.CONSUME);
               break;
            case RETURN_ERROR:
               // this will throw an exception, it is OK because this method
               // will always be called in the context of an inner block and
               // WILL NEVER be called directly from a top-level block
               returnWorker(imple.value, ReturnType.ERROR);
               break;
         }
      }
      
      // no match so we just re-throw the exception and continue to unwind
      // the stack; but we must process it as an unexpected exception which
      // will maintain rollback state properly
      abnormalEnd(ce, false);
      
      // never reached because the code above will re-throw the exception
      return null;
   }

   /**
    * Translate exception into {@link Condition} code.
    * 
    * @param    ce
    *           Source exception.
    *          
    * @return   one of the {@link Condition} codes or <code>null</code> if no
    *           match is found.
    */
   static Condition exceptionToCondition(ConditionException ce)
   {
      Condition current = null;
      
      if (ce instanceof ErrorConditionException)
      {
         current = Condition.ERROR;
      }
      else if (ce instanceof EndConditionException)
      {
         current = Condition.ENDKEY;
      }
      else if (ce instanceof StopConditionException)
      {
         current = Condition.STOP;
      }
      else if (ce instanceof QuitConditionException)
      {
         current = Condition.QUIT;
      }
      return current;
   }
   
   /**
    * Find a matching condition in the user-defined on phrases or the default
    * on phrases for the block and return the on phrase that is matched. The
    * user-defined on phrases are searched first and only if there is no
    * match will the default on phrases be searched.
    *
    * @param    user
    *           The set of user-defined on phrases for the current block.
    * @param    def
    *           The set of default on phrases for the current block.
    * @param    cond
    *           The condition being matched.
    *
    * @return   The on phrase that defines the behavior to be implemented or
    *           <code>null</code> if there is no match in either list.
    */
   private static OnPhrase matchCondition(OnPhrase[] user,
                                          OnPhrase[] def,
                                          Condition  cond)
   {
      if (cond == null)
         return null;
      
      if (user != null)
      {
         for (OnPhrase next : user)
         {
            if (next.condition == cond)
            {
               return next;
            }
         }
      }
      
      if (def != null)
      {
         for (OnPhrase next : def)
         {
            if (next.condition == cond)
            {
               return next;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Encode the conditions supported by this block as a bitfield.
    * <p>
    * The inputs are analyzed to match <code>Condition</code> values with
    * their corresponding bits in the <code>TransactionManager</code>
    * properties bitfield.
    * <p>
    * The bits are defined by {@link TransactionManager#PROP_NONE},
    * {@link TransactionManager#PROP_ERROR},
    * {@link TransactionManager#PROP_ENDKEY},
    * {@link TransactionManager#PROP_STOP} or
    * {@link TransactionManager#PROP_QUIT}.
    *
    * @param    user
    *           The set of user-defined on phrases for the current block.
    * @param    def
    *           The set of default on phrases for the current block.
    *
    * @return   The calculated bitmask with each bit set that matches one of
    *           the ON phrase conditions. <code>PROP_NONE</code> will be
    *           returned if both parameters are <code>null</code> and/or
    *           empty.
    */
   private static int decodeConditions(OnPhrase[] user, OnPhrase[] def)
   {
      int bits = PROP_NONE;
      
      if (user != null)
      {
         for (OnPhrase next : user)
         {
            bits |= conditionMask(next.condition);
         }
      }
      
      if (def != null)
      {
         for (OnPhrase next : def)
         {
            bits |= conditionMask(next.condition);
         }
      }
      
      return bits;
   }
   
   /**
    * Matches a <code>Condition</code> with its corresponding
    * <code>TransactionManager</code> constant.
    *
    * @param    cond
    *           The condition to match.
    *
    * @return   The integer constant that represents the given condition.
    */
   private static int conditionMask(Condition cond)
   {
      int rc = PROP_NONE;
      
      switch (cond)
      {
         case ERROR:
            rc = PROP_ERROR;
            break;
         case ENDKEY:
            rc = PROP_ENDKEY;
            break;
         case STOP:
            rc = PROP_STOP;
            break;
         case QUIT:
            rc = PROP_QUIT;
            break;
      }
      
      return rc;
   }
   
   /**
    * Detect if the leave event is referencing this block OR if it is
    * referencing a directly enclosing block which is not managed by this
    * class.  In either case, the result will be to break out of the current
    * block.  In all other cases, the result will be to re-throw the exception
    * and cause the stack to unwind to the nearest enclosing managed block.
    * <p>
    * The target of the leave operation must have been stored in context
    * local data when the leave exception was thrown (which caused the
    * stack to unwind).
    * <p>
    * This method is called from within a catch block to determine how to
    * respond to the deferred leave operation for the current block.
    *
    * @param    enclosing
    *           The list of non-managed blocks that directly enclose the
    *           current block.
    * @param    label
    *           The name of the current block.  If this matches the target of
    *           the deferred leave operation, then the current block must
    *           be left.
    *
    * @return   Either <code>BREAK</code> or <code>THROW</code>.
    */
   private static Disposition testLeave(String[] enclosing, String label)
   {
      // test if the LEAVE is destined for this block
      if (deferredLeave(label))
      {
         // it is, so honor the LEAVE
         return Disposition.BREAK;
      }
      else
      {
         WorkArea wa = work.obtain();
         
         // handle of LEAVE on an enclosing block that is not one
         // of which the block manager is aware
         if (matches(enclosing, wa.deferredLeave))
         {
            return Disposition.BREAK;
         }
      }
      
      // the LEAVE must be referencing an enclosing block which IS known
      // to the block manager OR to one of the other enclosing managed
      // blocks, so keep unwinding the stack
      return Disposition.THROW;
   }
   
   /**
    * Detect if the next event is referencing this block OR if it is
    * referencing a directly enclosing block which is not managed by this
    * class.  In the first case, the result will be to continue the current
    * block.  In the second case, the caller must break out of the current
    * block.  In all other cases, the result will be to re-throw the exception
    * and cause the stack to unwind to the nearest enclosing managed block.
    * <p>
    * The target of the next operation must have been stored in context
    * local data when the next exception was thrown (which caused the
    * stack to unwind).
    * <p>
    * This method is called from within a catch block to determine how to
    * respond to the deferred next operation for the current block.
    *
    * @param    enclosing
    *           The list of non-managed blocks that directly enclose the
    *           current block.
    * @param    label
    *           The name of the current block.  If this matches the target of
    *           the deferred leave operation, then the current block must
    *           be left.
    *
    * @return   Either <code>BREAK</code>, <code>CONTINUE</code> or
    *           <code>THROW</code>.
    */
   private static Disposition testNext(String[] enclosing, String label)
   {
      // test if the NEXT is destined for this block
      if (deferredNext(label))
      {
         // it is, so honor the NEXT
         return Disposition.CONTINUE;
      }
      else
      {
         WorkArea wa = work.obtain();
         
         // handle of NEXT on an enclosing block that is not one
         // of which the block manager is aware
         if (matches(enclosing, wa.deferredNext))
         {
            return Disposition.BREAK;
         }
      }
      
      // the NEXT must be referencing an enclosing block which IS known
      // to the block manager OR to one of the other enclosing managed
      // blocks, so keep unwinding the stack
      return Disposition.THROW;
   }
   
   /**
    * Common worker to cleanup top-level block state just before exiting.
    *
    * @param    wa
    *           The local context.
    */
   private static void exitTopLevel(WorkArea wa)
   {
      wa.topType.pop();
      wa.deferredLeave = null;
      wa.deferredNext  = null;
   }
   
   /**
    * Set the return value for the nearest enclosing function. This should
    * only be called if the nearest enclosing top-level block is a
    * <code>FUNCTION</code>.  Multiple calls to this before returning from
    * the function block will cause the value passed on the last call to be
    * honored.
    *
    * @param    bdt
    *           The return value to set. Passing <code>null</code> will
    *           reset the return value to <code>unknown</code>.
    */
   private static void setFuncReturn(BaseDataType bdt)
   {
      if (bdt != null)
      {
         // replace the current latest value with the given instance
         WorkArea wa = work.obtain();
         
         BlockType nearest = topLevelType(wa);
         
         if (nearest != null && nearest == BlockType.FUNCTION)
         {
            if (wa.funcRet.size() > 0)
            {
               wa.funcRet.pop();
               wa.funcRet.push(bdt);
            }
         }
      }
   }
   
   /**
    * Given a list of strings, detect if the given entry matches (case
    * sensitively) any of the elements in the list.
    *
    * @param    list
    *           The list of strings to check against.
    * @param    entry
    *           The element for which to search.
    *
    * @return   <code>true</code> if the given entry is found in the given
    *           list.
    */
   private static boolean matches(String[] list, String entry)
   {
      if (list == null || entry == null)
         return false;
      
      for (String element : list)
      {
         if (entry.equalsIgnoreCase(element))
         {
            return true;
         }
      }
         
      return false;
   }
   
   /**
    * Helper to convert between this class' enum and the corresponding
    * {@link TransactionManager} constants.
    *
    * @param    ttype
    *           The enum to convert.
    * @param    btype
    *           The block type.
    *
    * @return   The corresponding {@link TransactionManager} constant.
    */
   private static int convertTransType(TransactionType ttype, BlockType btype)
   {
      int newType = -1;
      
      switch (ttype)
      {
         case NONE:
            newType = NO_TRANSACTION;
            break;
         case SUB:
            newType = SUB_TRANSACTION;
            break;
         case FULL:
            newType = TRANSACTION;
            break;
      }
      
      return newType;
   }
   
   /**
    * Accesses the type of the nearest enclosing top level block.
    *
    * @param    wa
    *           The current context's data or <code>null</code> to force a
    *           lookup of the context data.
    *
    * @return   The type of the enclosing top level block or <code>null</code>
    *           if there is no top-level block in the call stack.
    */
   private static BlockType topLevelType(WorkArea wa)
   {
      if (wa == null)
      {
         wa = work.obtain();
      }
      
      return wa.topType.size() > 0 ? wa.topType.peek() : null;
   }
   
   /**
    * Update block execution statistics for the given block instance. Depending
    * of the block type, either the class name of the block will be inspected
    * or the call stack will be inspected to determine the block name and then
    * the usage statistics for that block will be incremented.
    *
    * @param    type
    *           The type of the block.
    * @param    block
    *           The block that is about to be executed.
    */
   public static void recordExecutionStats(BlockType type, Block block)
   {
      String descr = getDescription(type, block);
      
      recordStats(descr);
   }
   
   public static String getDescription(BlockType type, Block block)
   {
      // the first application entry in the stack is the one that describes
      // what we are executing
      character sname = EnvironmentOps.getSourceName(1);
      
      String descr = null;
      
      // this shouldn't be null, but we will be safe anyway
      if (sname == null)
      {
         descr = "Unknown (call stack analysis failed)";
      }
      else
      {
         descr = sname.toStringMessage();
      }
      
      // triggers will appear as the same as an external procedure
      if (type == BlockType.TRIGGER)
      {
         // calculate the specific trigger that was executed
         String className = block.getClass().getName();
         
         int idx = className.lastIndexOf('$');
         
         if (idx >= 0)
         {
            className = className.substring(idx + 1);
         }
         
         // append the trigger portion of the class name to the description
         StringBuilder sb = new StringBuilder(descr);
         sb.append(' ').append(className);
         
         descr = sb.toString();
      }
      
      return descr;
   }
   
   public static void recordStats(String descr)
   {
      if (stats == null)
         return;

      // update the statistics
      Integer count = stats.get(descr);
      
      if (count == null)
      {
         count = 1;
      }
      else
      {
         count++;
      }
      
      stats.put(descr, count);
      
      
      if (times == null)
      {
         times = new HashMap<String, Deque<Long>>();
      }
      
      Deque<Long> starts = times.get(descr);
      if (starts == null)
      {
         times.put(descr, starts = new ArrayDeque<Long>());
      }
      
      starts.push(System.currentTimeMillis());
   }
   
   
   public static void recordExecutionTime(BlockType type, Block block)
   {
      String descr = getDescription(type, block);
      recordTime(descr);
   }
   
   public static void recordTime(String descr)
   {
      if (stats == null)
         return;
      
      if (times == null)
      {
         times = new HashMap<String, Deque<Long>>();
      }
      
      Deque<Long> starts = times.get(descr);
      
      if (starts != null && starts.isEmpty() == false)
      {
         if (time == null)
            time = new HashMap<String, Long>();
         
         long t2 = System.currentTimeMillis();
         long t1 = starts.pop();
         Long total = time.get(descr);
         time.put(descr, (total == null ? 0 : total) + (t2 - t1));
      }
   }
   
   
   
   /**
    * Generates a CSV file with the recorded execution statistics.
    */
   private static void dumpExecutionStats()
   {
      if (statsFile != null)
      {
         PrintWriter out = null;
         
         try
         {
            out = new PrintWriter(new File(statsFile));
         }
         
         catch (FileNotFoundException fnfe)
         {
            // TODO: log this?
         }
         
         if (out != null)
         {
            out.println("\"Entry Point\",\"Execution Counter\",\"Execution Time\"");
            
            Set<String>      keyset = stats.keySet();
            Iterator<String> keys   = keyset.iterator();
            
            while (keys.hasNext())
            {
               String  entry = keys.next();
               Integer count = stats.get(entry);
               Long time = BlockManager.time.get(entry);
               
               out.printf("\"%s\",\"%d\",\"%d\"\n", entry, count, time == null ? 0 : time);
            }
         
            
            out.printf("Opened transactions in total T: %d\n", Persistence.openTxCountT);
            out.printf("Committed transactions in total T: %d\n", Persistence.commitTxCountT);
            out.printf("Opened transactions in total P: %d\n", Persistence.openTxCountP);
            out.printf("Committed transactions in total P: %d\n", Persistence.commitTxCountP);
            out.printf("Open tx time: %d\n", Persistence.openTxTime);
            out.printf("Commit tx time: %d\n", Persistence.commitTxTime);

            out.printf("Opened sessions in total: %d\n", Persistence.sessOpenCount);
            out.printf("Closed sessions in total: %d\n", Persistence.sessCloseCount);

            out.printf("Pushed tx in total: %d\n", Persistence.pushTxCount);
            out.printf("Popped tx in total: %d\n", Persistence.popTxCount);
            

            out.printf("Flush count: %d\n", Persistence.flushTxCount);
            out.printf("Flush time: %d\n", Persistence.flushTxTime);
            out.printf("Save count: %d\n", Persistence.saveTxCount);
            out.printf("Save time: %d\n", Persistence.saveTxTime);
            out.printf("Delete count: %d\n", Persistence.deleteTxCount);
            out.printf("Delete time: %d\n", Persistence.deleteTxTime);
            out.printf("Delete or update count: %d\n", Persistence.deleteUTxCount);
            out.printf("Delete or update time: %d\n", Persistence.deleteUTxTime);

            out.printf("Load 1 count: %d\n", Persistence.load1Count);
            out.printf("Load 1 time: %d\n", Persistence.load1Time);
            out.printf("Load 2 count: %d\n", Persistence.load2Count);
            out.printf("Load 2 time: %d\n", Persistence.load2Time);
            out.printf("Scroll count: %d\n", Persistence.scrollCount);
            out.printf("Scroll time: %d\n", Persistence.scrollTime);
            out.printf("List count: %d\n", Persistence.listCount);
            out.printf("List time: %d\n", Persistence.listTime);
            
            out.printf("Detach count: %d\n", Persistence.detachCount);
            out.printf("Detach time: %d\n", Persistence.detachTime);
            out.printf("Attach count: %d\n", Persistence.attachCount);
            out.printf("Attach time: %d\n", Persistence.attachTime);
            out.printf("Get query count: %d\n", Persistence.getQueryCount);
            out.printf("Get query time: %d\n", Persistence.getQueryTime);
            out.printf("Maybe flush count: %d\n", Persistence.maybeFlushCount);
            out.printf("Maybe flush time: %d\n", Persistence.maybeFlushTime);
            out.printf("Rollback count: %d\n", Persistence.rollbackCount);
            out.printf("Rollback flush time: %d\n", Persistence.rollbackTime);

            out.flush();
            out.close();
         }
      }
   }
   
   /**
    * Stores global data relating to the state of the current context's
    * blocks.
    */
   private static class WorkArea
   {
      /** A return value for each function block on the call stack. */
      public Stack<BaseDataType> funcRet = new Stack<BaseDataType>();
         
      /** Tracks the type of each top-level block as a stack. */
      public ArrayDeque<BlockType> topType = new ArrayDeque<BlockType>();
      
      /** Common, long-lived instance to make leave processing faster. */
      public LeaveUnwindException leaveExc = new LeaveUnwindException();
      
      /** Common, long-lived instance to make next processing faster. */
      public NextUnwindException nextExc = new NextUnwindException();
      
      /** Common, long-lived instance to make return processing faster. */
      public ReturnUnwindException retExc = new ReturnUnwindException();
      
      /** Block name which should be left. */
      public String deferredLeave = null;
      
      /** Block name which should be iterated. */
      public String deferredNext = null;
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Obtains the context-local instance of the contained global work
       * area.
       *
       * @return   The work area associated with this context.
       */
      public WorkArea obtain() 
      {
         return this.get();
      }   
      
      /**
       * Initializes the work area, the first time it is requested within a
       * new context.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }   
   }
   
   /**
    * Server-wide hook that gets called at initialization and termination. We
    * use this to write out execution statistics when the statistics counters
    * are enabled.
    */
   private static class StatsHelper
   implements InitTermListener
   {
      /**
       * This method is invoked when server is about to be started.
       */
      public void initialize()
      {
         // do nothing
      }
   
      /**
       * This method is invoked when the server is terminating.
       * 
       * @param    t
       *           Error or exception which caused exit of the server. Can
       *           be <code>null</code> if there was a normal end.
       */
      public void terminate(Throwable t)
      {
         dumpExecutionStats();
      }
   }
}
